(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8 Util.
Require Import AdomLib.
Require Import ZCongruences_defs.
Require Import IdealExpr IdealIntervals.

Set Implicit Arguments.
Unset Elimination Schemes.

Section var.

Context {var:Type}.

Record query_chan : Type :=
  { get_itv: iexpr var -> IdealIntervals.abs+⊤+⊥;
    get_congr: iexpr var → zcongr+⊥;
    get_eq_expr: var -> option (ite_iexpr var) }.

Record query_chan_gamma (chan:query_chan) (ρ:var->ideal_num) :=
  { get_itv_correct:
      ∀ e, eval_iexpr ρ e ⊆ γ (chan.(get_itv) e);
    get_congr_correct:
      ∀ e, (eval_iexpr ρ e ∘ INz) ⊆ γ (chan.(get_congr) e);
    get_eq_expr_correct:
      ∀ x e, chan.(get_eq_expr) x = Some e ->
             eval_ite_iexpr ρ e (ρ x) }.
Global Instance query_chan_gamma' : gamma_op query_chan (var -> ideal_num)
  := query_chan_gamma.

Definition query_chan_top : query_chan :=
  {| get_itv := fun e => ⊤;
     get_congr := fun e => ⊤;
     get_eq_expr := fun _ => None |}.

Lemma query_chan_top_correct :
  ∀ ρ, ρ ∈ γ query_chan_top.
Proof.
  intros. constructor.
  - intros. simpl.
    pose proof (iexpr_ty_correct H). destruct iexpr_ty.
    inv H0. constructor.
    inv H0. constructor.
  - intros. apply top_correct.
  - discriminate.
Qed.

Inductive message : Type :=
| Fact_msg : iexpr var -> bool -> message
| Known_value_msg : var -> ideal_num -> message
| Congr_msg : var -> zcongr -> message
| Broadcasted_msg : message -> message.

Global Instance message_gamma : gamma_op message (var->ideal_num) :=
  fix message_gamma m :=
    match m with
      | Fact_msg c b =>
        fun ρ => (INz (if b:bool then 1 else 0)) ∈ eval_iexpr ρ c
      | Known_value_msg x v => fun ρ => ρ x = v
      | Congr_msg x zc => fun ρ =>
        match ρ x with
          | INz v => v ∈ γ zc
          | _ => False
        end
      | Broadcasted_msg m =>
        message_gamma m
    end.

Definition messages_chan := list message.
Global Instance messages_chan_gamma : gamma_op messages_chan (var -> ideal_num) :=
  fun ml ρ => List.Forall (fun m => ρ ∈ γ m) ml.

Context `{EqDec var}.

Class ab_ideal_env (t:Type) : Type := {
  id_gamma:> gamma_op t (var -> ideal_num);
  id_top: query_chan -> (t * messages_chan)+⊥;

  id_leb: t * query_chan -> t -> bool;
  id_join: t * query_chan -> t * query_chan -> query_chan -> (t * messages_chan)+⊥;
  id_widen: t * query_chan -> t * query_chan -> query_chan -> (t * messages_chan)+⊥;

  assign: var -> iexpr var -> t * query_chan -> query_chan -> (t * messages_chan)+⊥;
  forget: var -> t * query_chan -> query_chan -> (t *  messages_chan)+⊥;

  process_msg: message -> t * query_chan -> (t * messages_chan)+⊥;
  enrich_query_chan: t -> query_chan -> query_chan;

  id_top_correct: forall chan ρ, ρ ∈ γ chan -> ρ ∈ γ (id_top chan);
  id_leb_correct: forall ab1 ab2 ρ,
    id_leb ab1 ab2 = true ->
    ρ ∈ γ ab1 ->
    ρ ∈ γ ab2;
  id_join_correct: forall ab1 ab2 chan ρ,
    ρ ∈ γ ab1 \/ ρ ∈ γ ab2 -> ρ ∈ γ chan ->
    ρ ∈ γ (id_join ab1 ab2 chan);
  assign_correct: forall x e ρ n ab chan,
    ρ ∈ γ ab -> (ρ+[x => n]) ∈ γ chan ->
    n ∈ eval_iexpr ρ e ->
    (ρ+[x => n]) ∈ γ (assign x e ab chan);
  forget_correct: forall x ρ n ab chan,
    ρ ∈ γ ab -> (ρ+[x => n]) ∈ γ chan ->
    (ρ+[x => n]) ∈ γ (forget x ab chan);

  process_msg_correct: forall m ρ ab,
    ρ ∈ γ ab -> ρ ∈ γ m ->
    ρ ∈ γ (process_msg m ab);
  enrich_query_chan_correct: forall ab chan ρ,
    ρ ∈ γ ab -> ρ ∈ γ chan ->
    ρ ∈ γ (enrich_query_chan ab chan)
}.

End var.
