(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import AdomLib Util Coqlib.
Require Import AbIdealEnv IdealExpr.

Section S.

Context {var: Type} {varEq:EqDec var}.

Fixpoint process_msg_list {B:Type} `{@ab_ideal_env _ _ B}
         (mchan:messages_chan) (ab:B * query_chan) (accu_mchan:messages_chan) :=
  match mchan with
    | nil => NotBot (fst ab, accu_mchan)
    | cons m mchanq =>
      do_bot (ab', mchan') <- process_msg m ab;
      process_msg_list mchanq (ab', snd ab) (mchan'++accu_mchan)%list
  end.

Lemma process_msg_list_correct {A:Type} `{@ab_ideal_env _ _ A}:
  forall (mchan:messages_chan) (ab:A * query_chan) (accu_mchan:messages_chan) ρ,
    ρ ∈ γ mchan -> ρ ∈ γ ab -> ρ ∈ γ accu_mchan ->
    ρ ∈ γ (process_msg_list mchan ab accu_mchan).
Proof.
  induction mchan. intros. destruct H1. split; auto. intros. simpl. inv H0.
  eapply botbind_spec with ρ, process_msg_correct; auto. intros [? ?] [? ?].
  apply IHmchan. auto. destruct H1; split; auto.
  simpl in *. clear - H2 H3.
  induction l. auto. inv H3. constructor. auto. apply IHl. auto.
Qed.

Instance ab_ideal_env_product {A B:Type}
         `(@ab_ideal_env var varEq A) `(@ab_ideal_env var varEq B) : ab_ideal_env (A*B)%type :=
  {| id_top qchan :=
       do_bot (a, mchana) <- id_top qchan;
       let qchana := enrich_query_chan a qchan in
       do_bot (b, mchanb) <- id_top qchana;
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) mchanb;
       NotBot ((a, b), mchan)
   ; id_leb x y :=
       let '((xa, xb), chan) := x in
       let '(ya, yb) := y in
       (id_leb (xa, chan) ya && id_leb (xb, chan) yb)%bool
   ; id_join x y qchan :=
       let '((xa, xb), xqchan) := x in
       let '((ya, yb), yqchan) := y in
       do_bot (a, mchana) <- id_join (xa, xqchan) (ya, yqchan) qchan;
       let qchana := enrich_query_chan a qchan in
       do_bot (b, mchanb) <- id_join (xb, xqchan) (yb, yqchan) qchana;
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) mchanb;
       NotBot ((a, b), mchan)
   ; id_widen x y qchan :=
       let '((xa, xb), xqchan) := x in
       let '((ya, yb), yqchan) := y in
       do_bot (a, mchana) <- id_widen (xa, xqchan) (ya, yqchan) qchan;
       let qchana := enrich_query_chan a qchan in
       do_bot (b, mchanb) <- id_widen (xb, xqchan) (yb, yqchan) qchana;
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) mchanb;
       NotBot ((a, b), mchan)
   ; assign v e ab qchan :=
       let '((a, b), abqchan) := ab in
       do_bot (a, mchana) <- assign v e (a, abqchan) qchan;
       let qchana := enrich_query_chan a qchan in
       do_bot (b, mchanb) <- assign v e (b, abqchan) qchana;
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) mchanb;
       NotBot ((a, b), mchan)
   ; forget v ab qchan :=
       let '((a, b), abqchan) := ab in
       do_bot (a, mchana) <- forget v (a, abqchan) qchan;
       let qchana := enrich_query_chan a qchan in
       do_bot (b, mchanb) <- forget v (b, abqchan) qchana;
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) mchanb;
       NotBot ((a, b), mchan)
   ; enrich_query_chan ab qchan :=
       let '(a, b) := ab in
       let qchan := enrich_query_chan a qchan in
       enrich_query_chan b qchan
   ; process_msg msg ab :=
       let '((a, b), abqchan) := ab in
       do_bot (a, mchana) <- process_msg msg (a, abqchan);
       let qchana := enrich_query_chan a abqchan in
       do_bot (b, mchan) <- process_msg_list mchana (b, qchana) nil;
       NotBot ((a, b), mchan)
  |}.
Proof.
  - intros.
    eapply botbind_spec with ρ, @id_top_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with ρ, @id_top_correct, @enrich_query_chan_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with ρ, process_msg_list_correct; auto. intros [? ?] [? ?].
    split. split. auto. auto. auto.
    split. auto. apply enrich_query_chan_correct; auto.
  - intros. destruct ab1 as [[] ?], ab2.
    rewrite Bool.andb_true_iff in H1. destruct H1.
    destruct H2. destruct H2.
    split. eapply id_leb_correct in H1. eauto. split. auto. auto.
    eapply id_leb_correct in H3. eauto. split. auto. auto.
  - intros. destruct ab1 as [[] ?], ab2  as [[] ?].
    eapply botbind_spec with ρ, @id_join_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with ρ, @id_join_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with ρ, process_msg_list_correct; auto. intros [? ?] [? ?].
    split. split. auto. auto. auto.
    split. auto. apply enrich_query_chan_correct; auto.
    destruct H1 as [[[] ?]|[[] ?]]; [left|right]; split; auto.
    apply enrich_query_chan_correct; auto.
    destruct H1 as [[[] ?]|[[] ?]]; [left|right]; split; auto.
  - intros. destruct ab as [[] ?], H1 as [[] ?].
    eapply botbind_spec with (upd ρ x n), @assign_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with (upd ρ x n), @assign_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with (upd ρ x n), @process_msg_list_correct; auto. intros [? ?] [? ?].
    split. split. auto. auto. auto.
    split. auto. apply enrich_query_chan_correct; auto. split; auto.
    apply enrich_query_chan_correct; auto. split; auto.
  - intros. destruct ab as [[] ?], H1 as [[] ?].
    eapply botbind_spec with (upd ρ x n), @forget_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with (upd ρ x n), @forget_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with (upd ρ x n), @process_msg_list_correct; auto. intros [? ?] [? ?].
    split. split. auto. auto. auto.
    split. auto. apply enrich_query_chan_correct; auto. split; auto.
    apply enrich_query_chan_correct; auto. split; auto.
  - intros. destruct ab as [[a  b]  c], H1 as [[Ha Hb] Hc].
    eapply botbind_spec with ρ, @process_msg_correct; auto. intros [? ?] [? ?].
    eapply botbind_spec with ρ, process_msg_list_correct; auto. intros [? ?] [? ?].
    split. split. auto. auto. auto.
    split. auto. apply enrich_query_chan_correct; auto.
    constructor.
    split; auto.
  - intros. destruct ab, H1.
    apply enrich_query_chan_correct; auto.
    apply enrich_query_chan_correct; auto.
Defined.

Section Reduce.

Context (A:Type) `{@ab_ideal_env var varEq A}.

Definition reduced_product := { ab : A * query_chan | γ (fst ab) ⊆ γ (snd ab) }.

Instance reduced_gamma : gamma_op reduced_product (var -> ideal_num) :=
  fun ab => γ (proj1_sig ab).

Fixpoint filter_broadcasted (l:@messages_chan var) : messages_chan :=
  match l with
    | nil => nil
    | Broadcasted_msg m :: l => m :: filter_broadcasted l
    | _ :: l => filter_broadcasted l
  end.

Program Definition reduce (ab:(A*@messages_chan var)+⊥) :
  (reduced_product*@messages_chan var)+⊥ :=
  do_bot ab <- ab;
  let '(ab, mchanab) := ab return _ in
  let c := enrich_query_chan ab query_chan_top in
  do_bot ab <- process_msg_list (filter_broadcasted mchanab) (ab, c) nil;
  let '(ab, _) := ab return _ in
  NotBot ((ab, enrich_query_chan ab query_chan_top):reduced_product, nil).
Next Obligation. apply enrich_query_chan_correct. auto. apply query_chan_top_correct. Qed.

Lemma reduce_correct:
  forall (ab:(A*messages_chan)+⊥) (ρ:var->ideal_num), ρ ∈ γ ab -> ρ ∈ γ (reduce ab).
Proof.
  intros.
  apply botbind_spec with ρ; auto. intros. destruct a, H1.
  apply botbind_spec with ρ, process_msg_list_correct; auto. intros. destruct a0, H3.
  split. split. auto. apply enrich_query_chan_correct, query_chan_top_correct; auto. constructor.
  clear - H2. simpl in *. unfold messages_chan_gamma in *.
  induction m. auto. inv H2. destruct a0; auto. constructor; auto.
  split. auto. apply enrich_query_chan_correct, query_chan_top_correct; auto.
  constructor.
Qed.

Program Instance ab_ideal_env_reduce: ab_ideal_env reduced_product :=
  {| id_top := fun _ => reduce (id_top query_chan_top)
   ; id_leb x y :=
       let '(x, _) := x in
       let '(y, _) := y in
       id_leb x y
   ; id_join x y := fun _ =>
       let '(x, _) := x in
       let '(y, _) := y in
       reduce (id_join x y query_chan_top)
   ; id_widen x y qchan :=
       let '(x, _) := x in
       let '(y, _) := y in
       do_bot (ab, mchanab) <- id_widen x y query_chan_top;
       NotBot ((ab, enrich_query_chan ab query_chan_top):reduced_product, nil)
   ; assign v e ab qchan :=
       let '(ab, _) := ab in
       reduce (assign v e ab query_chan_top)
   ; forget v ab qchan :=
       let '(ab, _) := ab in
       reduce (forget v ab query_chan_top)
   ; enrich_query_chan ab := fun _ =>
       snd ab
   ; process_msg msg ab :=
       let '(ab, _) := ab in
       reduce (process_msg msg ab)
  |}.
Next Obligation. apply enrich_query_chan_correct, query_chan_top_correct. auto. Qed.
Next Obligation. apply reduce_correct, id_top_correct, query_chan_top_correct. Qed.
Next Obligation. destruct r, H1, ab2 as [[ab2 ab2c] Hab2]. simpl in H0.
                 eapply id_leb_correct in H0; eauto. split; eauto. Qed.
Next Obligation. apply reduce_correct, id_join_correct, query_chan_top_correct.
                 destruct H0 as [[]|[]]; auto. Qed.
Next Obligation. apply reduce_correct, assign_correct; auto. apply H0.
                 apply query_chan_top_correct. Qed.
Next Obligation. apply reduce_correct, forget_correct; auto. apply H0.
                 apply query_chan_top_correct. Qed.
Next Obligation. apply reduce_correct, process_msg_correct; auto. apply H0. Qed.
Next Obligation. apply H0. Qed.

Program Instance reduced_ToString {T:ToString.ToString A} : ToString.ToString reduced_product :=
  fun x => T (fst x).

End Reduce.

End S.

Existing Instances reduced_ToString ab_ideal_env_reduce ab_ideal_env_product.