(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import Util.
Require Import Coqlib.
Require Import Ctypes.
Require Import Values.
Require Import Integers.
Require Import Floats.
Require Import Cminor.
Require Import AdomLib.
Require Import IntSet.
Require Import FloatIntervals.
Require Import ZIntervals.

Set Implicit Arguments.
Section nexpr.
  Variable var : Type.

  Inductive num_val : Type :=
  | NVint : int -> num_val
  | NVint64 : int64 -> num_val
  | NVfloat : float -> num_val
  | NVsingle : float32 -> num_val.

  Inductive num_val_ty : Type :=
  | NVTint | NVTint64 | NVTfloat | NVTsingle.

  Inductive num_val_ty_gamma : gamma_op num_val_ty num_val :=
  | NVTSint : ∀ i, num_val_ty_gamma NVTint (NVint i)
  | NVTSint64 : ∀ i, num_val_ty_gamma NVTint64 (NVint64 i)
  | NVTSfloat : ∀ f, num_val_ty_gamma NVTfloat (NVfloat f)
  | NVTSsingle : ∀ f, num_val_ty_gamma NVTsingle (NVsingle f).
  Global Existing Instance num_val_ty_gamma.

  Hint Constructors num_val_ty_gamma.

  Definition compat_fun (v:val) : num_val :=
    match v with
    | Vptr _ i | Vint i => NVint i
    | Vlong l => NVint64 l
    | Vundef => NVint Int.zero
    | Vfloat f => NVfloat f
    | Vsingle f => NVsingle f
    end.

  Inductive num_val_itv : Type :=
  | NVIint : ZIntervals.itv+⊤ -> num_val_itv
  | NVIint64 : ZIntervals.itv+⊤ -> num_val_itv
  | NVIfloat : FloatIntervals.itv+⊤ -> num_val_itv
  | NVIsingle : FloatIntervals.itv+⊤ -> num_val_itv.

  Inductive num_val_itv_gamma : gamma_op num_val_itv num_val :=
  | NVISint : ∀ itv x, x ∈ γ itv -> num_val_itv_gamma (NVIint itv) (NVint (Int.repr x))
  | NVISint64 : ∀ itv x, x ∈ γ itv -> num_val_itv_gamma (NVIint64 itv) (NVint64 (Int64.repr x))
  | NVISfloat : ∀ itv f, f ∈ γ itv -> num_val_itv_gamma (NVIfloat itv) (NVfloat f)
  | NVISsingle : ∀ itv f, (Float.of_single f) ∈ γ itv -> num_val_itv_gamma (NVIsingle itv) (NVsingle f).
  Global Existing Instance num_val_itv_gamma.

  Inductive nexpr : Type :=
  | NEvar : num_val_ty -> var -> nexpr
  | NEconst : num_val_itv -> nexpr
  | NEunop : unary_operation -> nexpr -> nexpr
  | NEbinop : binary_operation -> nexpr -> nexpr -> nexpr.

  Fixpoint ne_free_var (ne: nexpr) : ℘(var) :=
    match ne with
    | NEvar _ v => eq v
    | NEconst _ => ∅
    | NEunop _ ne' => ne_free_var ne'
    | NEbinop _ ne₁ ne₂ => ne_free_var ne₁ ∪ ne_free_var ne₂
    end.

  Section eval_nexpr.

    Variable ρ: var -> num_val.

    Inductive is_bool : num_val -> Prop :=
    | is_bool_false: is_bool (NVint Int.zero)
    | is_bool_true: is_bool (NVint Int.one).

    Definition Ntrue: num_val := NVint Int.one.
    Definition Nfalse: num_val := NVint Int.zero.

    Definition of_bool (b: bool): num_val :=
      if b then Ntrue else Nfalse.

    Inductive eval_nunop: unary_operation -> num_val -> num_val -> Prop :=
    | eval_nunop_Ocast8unsigned: forall i,
      eval_nunop Ocast8unsigned  (NVint i)  (NVint (Int.zero_ext 8 i))
    | eval_nunop_Ocast8signed: forall i,
      eval_nunop Ocast8signed  (NVint i)  (NVint (Int.sign_ext 8 i))
    | eval_nunop_Ocast16unsigned: forall i,
      eval_nunop Ocast16unsigned  (NVint i)  (NVint (Int.zero_ext 16 i))
    | eval_nunop_Ocast16signed: forall i,
      eval_nunop Ocast16signed  (NVint i)  (NVint (Int.sign_ext 16 i))
    | eval_nunop_Onegint : forall i,
      eval_nunop Onegint  (NVint i)  (NVint (Int.neg i))
    | eval_nunop_Onotint : forall i,
      eval_nunop Onotint  (NVint i)  (NVint (Int.not i))
    | eval_nunop_Onegf : forall f,
      eval_nunop Onegf  (NVfloat f)  (NVfloat (Float.neg f))
    | eval_nunop_Onegfs : forall f,
      eval_nunop Onegfs (NVsingle f)  (NVsingle (Float32.neg f))
    | eval_nunop_Oabsf : forall f,
      eval_nunop Oabsf  (NVfloat f)  (NVfloat (Float.abs f))
    | eval_nunop_Oabsfs : forall f,
      eval_nunop Oabsfs  (NVsingle f)  (NVsingle (Float32.abs f))
    | eval_nunop_Osingleoffloat : forall f,
      eval_nunop Osingleoffloat  (NVfloat f)  (NVsingle (Float.to_single f))
    | eval_nunop_Ofloatofsingle : forall f,
      eval_nunop Ofloatofsingle  (NVsingle f)  (NVfloat (Float.of_single f))
    | eval_nunop_Ointoffloat : forall f i,
      Float.to_int f = Some i ->
      eval_nunop Ointoffloat  (NVfloat f)  (NVint i)
    | eval_nunop_Ointofsingle : forall f i,
      Float32.to_int f = Some i ->
      eval_nunop Ointofsingle  (NVsingle f)  (NVint i)
    | eval_nunop_Ointuoffloat : forall f i,
      Float.to_intu f = Some i ->
      eval_nunop Ointuoffloat  (NVfloat f)  (NVint i)
    | eval_nunop_Ointuofsingle : forall f i,
      Float32.to_intu f = Some i ->
      eval_nunop Ointuofsingle  (NVsingle f)  (NVint i)
    | eval_nunop_Ofloatofint : forall i,
      eval_nunop Ofloatofint  (NVint i)  (NVfloat (Float.of_int i))
    | eval_nunop_Osingleofint : forall i,
      eval_nunop Osingleofint  (NVint i)  (NVsingle (Float32.of_int i))
    | eval_nunop_Ofloatofintu : forall i,
      eval_nunop Ofloatofintu  (NVint i)  (NVfloat (Float.of_intu i))
    | eval_nunop_Osingleofintu : forall i,
      eval_nunop Osingleofintu  (NVint i)  (NVsingle (Float32.of_intu i))
    | eval_nunop_Onegl : forall i,
      eval_nunop Onegl  (NVint64 i)  (NVint64 (Int64.neg i))
    | eval_nunop_Onotl : forall i,
      eval_nunop Onotl  (NVint64 i)  (NVint64 (Int64.not i))
    | eval_nunop_Ointoflong : forall i,
      eval_nunop Ointoflong  (NVint64 i)  (NVint (Int64.loword i))
    | eval_nunop_Olongofint : forall i,
      eval_nunop Olongofint  (NVint i)  (NVint64 (Int64.repr (Int.signed i)))
    | eval_nunop_Olongofintu : forall i,
      eval_nunop Olongofintu  (NVint i)  (NVint64 (Int64.repr (Int.unsigned i)))
    | eval_nunop_Olongoffloat : forall f i,
      Float.to_long f = Some i ->
      eval_nunop Olongoffloat  (NVfloat f)  (NVint64 i)
    | eval_nunop_Olonguoffloat : forall f i,
      Float.to_longu f = Some i ->
      eval_nunop Olonguoffloat  (NVfloat f)  (NVint64 i)
    | eval_nunop_Olongofsingle : forall f i,
      Float32.to_long f = Some i ->
      eval_nunop Olongofsingle  (NVsingle f)  (NVint64 i)
    | eval_nunop_Olonguofsingle : forall f i,
      Float32.to_longu f = Some i ->
      eval_nunop Olonguofsingle  (NVsingle f)  (NVint64 i)
    | eval_nunop_Ofloatoflong : forall i,
      eval_nunop Ofloatoflong  (NVint64 i)  (NVfloat (Float.of_long i))
    | eval_nunop_Osingleoflong : forall i,
      eval_nunop Osingleoflong  (NVint64 i)  (NVsingle (Float32.of_long i))
    | eval_nunop_Ofloatoflongu : forall i,
      eval_nunop Ofloatoflongu  (NVint64 i)  (NVfloat (Float.of_longu i))
    | eval_nunop_Osingleoflongu : forall i,
      eval_nunop Osingleoflongu  (NVint64 i)  (NVsingle (Float32.of_longu i)).

    Inductive eval_nbinop : binary_operation -> num_val -> num_val -> num_val -> Prop :=
    | eval_nbinop_Oadd : forall i1 i2,
      eval_nbinop Oadd  (NVint i1)  (NVint i2)  (NVint (Int.add i1 i2))
    | eval_nbinop_Osub : forall i1 i2,
      eval_nbinop Osub  (NVint i1)  (NVint i2)  (NVint (Int.sub i1 i2))
    | eval_nbinop_Omul : forall i1 i2,
      eval_nbinop Omul  (NVint i1)  (NVint i2)  (NVint (Int.mul i1 i2))
    | eval_nbinop_Odiv : forall i1 i2,
      Int.eq i2 Int.zero = false ->
      (Int.eq i1 (Int.repr Int.min_signed) = false
        \/ Int.eq i2 Int.mone = false) ->
      eval_nbinop Odiv  (NVint i1)  (NVint i2)  (NVint (Int.divs i1 i2))
    | eval_nbinop_Odivu : forall i1 i2,
      Int.eq i2 Int.zero = false ->
      eval_nbinop Odivu  (NVint i1)  (NVint i2)  (NVint (Int.divu i1 i2))
    | eval_nbinop_Omod : forall i1 i2,
      Int.eq i2 Int.zero = false ->
      (Int.eq i1 (Int.repr Int.min_signed) = false \/
        Int.eq i2 Int.mone = false) ->
      eval_nbinop Omod  (NVint i1)  (NVint i2)  (NVint (Int.mods i1 i2))
    | eval_nbinop_Omodu : forall i1 i2,
      Int.eq i2 Int.zero = false ->
      eval_nbinop Omodu  (NVint i1)  (NVint i2)  (NVint (Int.modu i1 i2))
    | eval_nbinop_Oand : forall i1 i2,
      eval_nbinop Oand  (NVint i1)  (NVint i2)  (NVint (Int.and i1 i2))
    | eval_nbinop_Oor : forall i1 i2,
      eval_nbinop Oor  (NVint i1)  (NVint i2)  (NVint (Int.or i1 i2))
    | eval_nbinop_Oxor : forall i1 i2,
      eval_nbinop Oxor  (NVint i1)  (NVint i2)  (NVint (Int.xor i1 i2))
    | eval_nbinop_Oshl : forall i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshl  (NVint i1)  (NVint i2)  (NVint (Int.shl i1 i2))
    | eval_nbinop_Oshr : forall i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshr  (NVint i1)  (NVint i2)  (NVint (Int.shr i1 i2))
    | eval_nbinop_Oshru : forall i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshru  (NVint i1)  (NVint i2)  (NVint (Int.shru i1 i2))

    | eval_nbinop_Oaddf : forall f1 f2,
      eval_nbinop Oaddf (NVfloat f1) (NVfloat f2) (NVfloat (Float.add f1 f2))
    | eval_nbinop_Osubf : forall f1 f2,
      eval_nbinop Osubf (NVfloat f1) (NVfloat f2) (NVfloat (Float.sub f1 f2))
    | eval_nbinop_Omulf : forall f1 f2,
      eval_nbinop Omulf (NVfloat f1) (NVfloat f2) (NVfloat (Float.mul f1 f2))
    | eval_nbinop_Odivf : forall f1 f2,
      eval_nbinop Odivf (NVfloat f1) (NVfloat f2) (NVfloat (Float.div f1 f2))

    | eval_nbinop_Oaddfs : forall f1 f2,
      eval_nbinop Oaddfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.add f1 f2))
    | eval_nbinop_Osubfs : forall f1 f2,
      eval_nbinop Osubfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.sub f1 f2))
    | eval_nbinop_Omulfs : forall f1 f2,
      eval_nbinop Omulfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.mul f1 f2))
    | eval_nbinop_Odivfs : forall f1 f2,
      eval_nbinop Odivfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.div f1 f2))

    | eval_nbinop_Oaddl : forall i1 i2,
      eval_nbinop Oaddl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.add i1 i2))
    | eval_nbinop_Osubl : forall i1 i2,
      eval_nbinop Osubl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.sub i1 i2))
    | eval_nbinop_Omull : forall i1 i2,
      eval_nbinop Omull  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.mul i1 i2))
    | eval_nbinop_Odivl : forall i1 i2,
      Int64.eq i2 Int64.zero = false ->
      (Int64.eq i1 (Int64.repr Int64.min_signed) = false
        \/ Int64.eq i2 Int64.mone = false) ->
      eval_nbinop Odivl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.divs i1 i2))
    | eval_nbinop_Odivlu : forall i1 i2,
      Int64.eq i2 Int64.zero = false ->
      eval_nbinop Odivlu  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.divu i1 i2))
    | eval_nbinop_Omodl : forall i1 i2,
      Int64.eq i2 Int64.zero = false ->
      (Int64.eq i1 (Int64.repr Int64.min_signed) = false \/
        Int64.eq i2 Int64.mone = false) ->
      eval_nbinop Omodl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.mods i1 i2))
    | eval_nbinop_Omodlu : forall i1 i2,
      Int64.eq i2 Int64.zero = false ->
      eval_nbinop Omodlu  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.modu i1 i2))
    | eval_nbinop_Oandl : forall i1 i2,
      eval_nbinop Oandl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.and i1 i2))
    | eval_nbinop_Oorl : forall i1 i2,
      eval_nbinop Oorl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.or i1 i2))
    | eval_nbinop_Oxorl : forall i1 i2,
      eval_nbinop Oxorl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.xor i1 i2))
    | eval_nbinop_Oshll : forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshll  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shl' i1 i2))
    | eval_nbinop_Oshrl : forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshrl  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shr' i1 i2))
    | eval_nbinop_Oshrlu : forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshrlu  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shru' i1 i2))

    | eval_nbinop_Ocmp : forall c i1 i2,
      eval_nbinop (Ocmp c)  (NVint i1)  (NVint i2) (of_bool (Int.cmp c i1 i2))
    | eval_nbinop_Ocmpu : forall c i1 i2,
      eval_nbinop (Ocmpu c)  (NVint i1)  (NVint i2) (of_bool (Int.cmpu c i1 i2))
    | eval_nbinop_Ocmpl : forall c i1 i2,
      eval_nbinop (Ocmpl c)  (NVint64 i1)  (NVint64 i2) (of_bool (Int64.cmp c i1 i2))
    | eval_nbinop_Ocmplu : forall c i1 i2,
      eval_nbinop (Ocmplu c)  (NVint64 i1)  (NVint64 i2) (of_bool (Int64.cmpu c i1 i2))
    | eval_nbinop_Ocmpf : forall c f1 f2,
      eval_nbinop (Ocmpf c)  (NVfloat f1)  (NVfloat f2) (of_bool (Float.cmp c f1 f2))
    | eval_nbinop_Ocmpfs : forall c f1 f2,
      eval_nbinop (Ocmpfs c)  (NVsingle f1)  (NVsingle f2) (of_bool (Float32.cmp c f1 f2)).

    Inductive eval_nexpr: nexpr -> num_val -> Prop :=
    | eval_NEvar : forall id i ty,
      ρ id = i -> γ ty i ->
      eval_nexpr (NEvar ty id) i
    | eval_NEconst: forall x itv, x ∈ γ itv ->
      eval_nexpr (NEconst itv) x
    | eval_NEunop: forall op a1 n1 n,
      eval_nexpr a1 n1 ->
      eval_nunop op n1 n ->
      eval_nexpr (NEunop op a1) n
    | eval_Ebinop: forall op a1 a2 n1 n2 n,
      eval_nexpr a1 n1 ->
      eval_nexpr a2 n2 ->
      eval_nbinop op n1 n2 n ->
      eval_nexpr (NEbinop op a1 a2) n.

    Inductive block_nunop: unary_operation -> num_val -> Prop :=
    | block_nunop_Ointoffloat: forall f, Float.to_int f = None ->
      block_nunop Ointoffloat (NVfloat f)
    | block_nunop_Ointuoffloat: forall f, Float.to_intu f = None ->
      block_nunop Ointuoffloat (NVfloat f)
    | block_nunop_Olongoffloat: forall f, Float.to_long f = None ->
      block_nunop Olongoffloat (NVfloat f)
    | block_nunop_Olonguoffloat: forall f, Float.to_longu f = None ->
      block_nunop Olonguoffloat (NVfloat f)
    | block_nunop_Ointofsingle: forall f, Float32.to_int f = None ->
      block_nunop Ointofsingle (NVsingle f)
    | block_nunop_Ointuofsingle: forall f, Float32.to_intu f = None ->
      block_nunop Ointuofsingle (NVsingle f)
    | block_nunop_Olongofsingle: forall f, Float32.to_long f = None ->
      block_nunop Olongofsingle (NVsingle f)
    | block_nunop_Olonguofsingle: forall f, Float32.to_longu f = None ->
      block_nunop Olonguofsingle (NVsingle f).

    Inductive block_nbinop: binary_operation -> num_val -> num_val -> Prop :=
    | block_nbinop_Odiv_0: forall i,
      block_nbinop Odiv (NVint i) (NVint Int.zero)
    | block_nbinop_Odiv_minint:
      block_nbinop Odiv (NVint (Int.repr Int.min_signed)) (NVint Int.mone)
    | block_nbinop_Odivu_0: forall i,
      block_nbinop Odivu (NVint i) (NVint Int.zero)
    | block_nbinop_Omod_0: forall i,
      block_nbinop Omod (NVint i) (NVint Int.zero)
    | block_nbinop_Omod_minint:
      block_nbinop Omod (NVint (Int.repr Int.min_signed)) (NVint Int.mone)
    | block_nbinop_Omodu_0: forall i,
      block_nbinop Omodu (NVint i) (NVint Int.zero)
    | block_nbinop_Oshl: forall i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      block_nbinop Oshl (NVint i1) (NVint i2)
    | block_nbinop_Oshr: forall i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      block_nbinop Oshr (NVint i1) (NVint i2)
    | block_nbinop_Oshru: forall i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      block_nbinop Oshru (NVint i1) (NVint i2)
    | block_nbinop_Odivl_0: forall i,
      block_nbinop Odivl (NVint64 i) (NVint64 Int64.zero)
    | block_nbinop_Odivl_minint:
      block_nbinop Odivl (NVint64 (Int64.repr Int64.min_signed)) (NVint64 Int64.mone)
    | block_nbinop_Odivlu_0: forall i,
      block_nbinop Odivlu (NVint64 i) (NVint64 Int64.zero)
    | block_nbinop_Omodl_0: forall i,
      block_nbinop Omodl (NVint64 i) (NVint64 Int64.zero)
    | block_nbinop_Omodl_minint:
      block_nbinop Omodl (NVint64 (Int64.repr Int64.min_signed)) (NVint64 Int64.mone)
    | block_nbinop_Omodlu_0: forall i,
      block_nbinop Omodlu (NVint64 i) (NVint64 Int64.zero)
    | block_nbinop_Oshll: forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      block_nbinop Oshll (NVint64 i1) (NVint i2)
    | block_nbinop_Oshrl: forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      block_nbinop Oshrl (NVint64 i1) (NVint i2)
    | block_nbinop_Oshrlu: forall i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      block_nbinop Oshrlu (NVint64 i1) (NVint i2).

    Inductive block_nexpr: nexpr -> Prop :=
    | block_NEunop: forall op a1 n1,
      eval_nexpr a1 n1 ->
      block_nunop op n1 ->
      block_nexpr (NEunop op a1)
    | block_NEunop_rec: forall op a1,
      block_nexpr a1 ->
      block_nexpr (NEunop op a1)
    | block_NEbinop: forall op a1 n1 a2 n2,
      eval_nexpr a1 n1 ->
      eval_nexpr a2 n2 ->
      block_nbinop op n1 n2 ->
      block_nexpr (NEbinop op a1 a2)
    | block_NEbinop_rec1: forall op a1 a2,
      block_nexpr a1 ->
      block_nexpr (NEbinop op a1 a2)
    | block_NEbinop_rec2: forall op a1 a2,
      block_nexpr a2 ->
      block_nexpr (NEbinop op a1 a2).

  End eval_nexpr.

  Lemma eval_nexpr_fv ρ ρ' ne :
    (∀ x, x ∈ ne_free_var ne → ρ x = ρ' x) →
    eval_nexpr ρ ne ⊆ eval_nexpr ρ' ne.
  Proof.
    induction ne; simpl; intros H q Hq; inv Hq; econstructor (eauto).
    specialize (H _ eq_refl). easy.
  Qed.

  Definition nexpr_ty e : num_val_ty :=
    match e with
      | NEvar ty _ => ty
      | NEconst (NVIint _) => NVTint
      | NEconst (NVIint64 _) => NVTint64
      | NEconst (NVIfloat _) => NVTfloat
      | NEconst (NVIsingle _) => NVTsingle
      | NEunop (Ocast8unsigned | Ocast8signed | Ocast16unsigned | Ocast16signed |
                Onegint | Onotint | Ointoffloat | Ointuoffloat |
                Ointofsingle | Ointuofsingle | Ointoflong) _ =>
        NVTint
      | NEunop (Onegf | Oabsf | Ofloatofsingle | Ofloatofint | Ofloatofintu |
                Ofloatoflong | Ofloatoflongu) _ =>
        NVTfloat
      | NEunop (Onegfs | Oabsfs | Osingleoffloat | Osingleofint | Osingleofintu |
                Osingleoflong | Osingleoflongu) _ =>
        NVTsingle
      | NEunop (Onegl | Onotl | Olongofint | Olongofintu |
                Olongoffloat | Olonguoffloat | Olongofsingle | Olonguofsingle) _ =>
        NVTint64
      | NEbinop (Oadd | Osub | Omul | Odiv | Odivu | Omod | Omodu | Oand | Oor |
                 Oxor | Oshl | Oshr | Oshru |
                 Ocmp _ | Ocmpu _ | Ocmpf _ | Ocmpfs _ | Ocmpl _ | Ocmplu _ ) _ _ =>
        NVTint
      | NEbinop (Oaddf | Osubf | Omulf | Odivf) _ _ => NVTfloat
      | NEbinop (Oaddfs | Osubfs | Omulfs | Odivfs) _ _ => NVTsingle
      | NEbinop (Oaddl | Osubl | Omull | Odivl | Odivlu |
                 Omodl | Omodlu | Oandl | Oorl | Oxorl | Oshll | Oshrl | Oshrlu) _ _ =>
        NVTint64
    end.

  Lemma nexpr_ty_correct:
    ∀ n e ρ, eval_nexpr ρ e n -> γ (nexpr_ty e) n.
  Proof.
    destruct 1.
    - destruct H0; constructor.
    - destruct H; constructor.
    - destruct H0; constructor.
    - destruct H1; try constructor; simpl;
      match goal with |- context [of_bool ?b] => destruct b end; constructor.
  Qed.

  Definition ne_const_i (z: Z) : nexpr := NEconst (NVIint (Just (ZIntervals.ITV z z))).
  Definition ne_const_l (z: Z) : nexpr := NEconst (NVIint64 (Just (ZIntervals.ITV z z))).
  Definition ne_const_f (f: float) : nexpr := NEconst (NVIfloat (Just (FloatIntervals.ITV f f))).
  Definition ne_const_s (f: float32) : nexpr :=
    NEconst (NVIsingle (Just (FloatIntervals.ITV (Float.of_single f) (Float.of_single f)))).

  Lemma eval_nexpr_const_i ρ z n :
    eval_nexpr ρ (ne_const_i z) n →
    n = NVint (Int.repr z).
  Proof.
    unfold ne_const_i. intros H; inv H.
    repeat match goal with H : γ _ _ |- _ => inversion H; clear H; subst end.
    f_equal. f_equal.
    auto with zarith.
  Qed.

End nexpr.

Unset Elimination Schemes.
(** Domain of environments using machine numbers. *)

Class ab_machine_env (var t: Type) `{EqDec var} : Type :=
  { mach_gamma :> gamma_op t (var->num_val)

  ; mach_top :> top_op (t+⊥)
  ; mach_leb:> leb_op t
  ; mach_join:> join_op t (t+⊥)
  ; mach_widen:> widen_op t (t+⊥)

  ; concretize_int: N → nexpr var → t → int_set+⊤+⊥
  ; assign: var -> nexpr var -> t -> t+⊥
  ; forget: var -> t -> t+⊥
  ; assume: nexpr var -> bool -> t -> t+⊥
  ; nonblock: nexpr var -> t -> bool

  ; mach_top_correct:> top_op_correct (t+⊥) (var->num_val)
  ; mach_leb_correct:> leb_op_correct t (var->num_val)
  ; mach_join_correct:> join_op_correct t (t+⊥) (var->num_val)
  ; concretize_int_correct: ∀ n e ρ ab,
    ρ ∈ γ ab →
    (eval_nexpr ρ e ∘ NVint) ⊆ γ (concretize_int n e ab)
  ; assign_correct: forall x e ρ n ab,
    ρ ∈ γ ab ->
    n ∈ eval_nexpr ρ e ->
    (upd ρ x n) ∈ γ (assign x e ab)
  ; forget_correct: forall x ρ n ab,
    ρ ∈ γ ab ->
    (upd ρ x n) ∈ γ (forget x ab)
  ; assume_correct: forall e ρ ab b,
    ρ ∈ γ ab ->
    of_bool b ∈ eval_nexpr ρ e ->
    ρ ∈ γ (assume e b ab)
  ; nonblock_correct: forall e ρ ab,
    ρ ∈ γ ab ->
    nonblock e ab = true ->
    block_nexpr ρ e ->
    False
}.

Section MORE.

Context (var t : Type) `(EQ: EqDec var) `(D: @ab_machine_env var t EQ).

Lemma forget_correct':
  forall (x: var) (ρ ρ': var → num_val) (ab: t),
    ρ ∈ γ(ab) →
    (∀ y, y ≠ x → ρ'(y) = ρ(y)) →
    ρ' ∈ γ (forget x ab).
Proof.
  intros x ρ ρ' ab H3 H2.
  erewrite (Axioms.extensionality ρ').
  apply forget_correct; eauto.
  intros z. unfold upd.
  destruct @eq_dec as [ -> | NE ]. reflexivity.
  apply H2, NE.
Qed.

Lemma assume_bot:
  ∀ e b (ab: t),
    is_bot (assume e b ab) = true →
    ∀ ρ, ρ ∈ γ(ab) →
         ¬ of_bool(b) ∈ eval_nexpr ρ e.
Proof.
  intros e b ab B ρ G K.
  erewrite is_bot_spec in B. discriminate.
  apply assume_correct; eauto.
Qed.

End MORE.
