(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8
  Coqlib
  Globalenvs
  Csharpminor
  Behaviors
  String.

Require Import
  Utils
  ShareTree
  ToString
  AbMachineEnv
  IdealEnvToMachineEnv
  IdealIntervalsNonrel
  PolAdomZ
  BlockMon
  AbMemSignatureCsharpminor
  DebugAbMemCsharpminor
  CsharpminorIter
  CsharpminorIterproof
  Cells MemCsharpminor
  AbIdealEnvProduct
  AbVarExpEq
  Hash.

Require Import
  DebugAbMachineEnv
  ToString
  AdomLib
  ZCongruences
  CellEncoding.

Section S.

  Context (max_concretize : N)
          (trace verbose: bool)
          (unroll: nat).

  Context {tnum: Type}
          `{ToString tnum}
          (M: @ab_machine_env cell tnum _).

  Definition doit (p: program) : block_mon unit :=
    let ge := Genv.globalenv p in
    let mem_dom := mem_cshm_dom M ge max_concretize in
    iter_program _ unroll p (if trace then @debug_mem_dom _ mem_dom (TToString ge) verbose
                             else mem_dom).

End S.

Local Unset Elimination Schemes.
Inductive num_dom_kind :=
| ZIntervals
| Poly
.

Module IETME_AC := IdealEnvToMachineEnv.Make ACTree.
Module ZCIE_AC := ZCongruences.Make ACTree.
Module IIE_AC := IdealIntervalsNonrel.Make ACTree.
Module VEEQ_AC := AbVarExpEq.Make ACTree.
Instance VEEQ_AC_inst : AbIdealEnv.ab_ideal_env VEEQ_AC.t :=
  VEEQ_AC.abVarExpEq_env MIX_cell.

Module IETME_P := IdealEnvToMachineEnv.Make PShareTree.
Module ZCIE_P := ZCongruences.Make PShareTree.
Module IIE_P := IdealIntervalsNonrel.Make PShareTree.
Module VEEQ_P := AbVarExpEq.Make PShareTree.
Instance VEEQ_P_inst : AbIdealEnv.ab_ideal_env VEEQ_P.t :=
  VEEQ_P.abVarExpEq_env (fun h p => MIX_N h (Npos p)).

Module PTreeDom := WPFun PShareTree.

Existing Instance ACTreeDom.toString.
Existing Instance PTreeDom.toString.

Section VA.

  Context (kind: num_dom_kind)
          (max_concretize: N)
          (trace verbose: bool)
          (unroll: nat).

  Local Notation analysis :=
    (λ e u p,
     let ge := Genv.globalenv p in
     let CS : ToString _ := CellToString ge in
     doit max_concretize trace verbose unroll
          (if trace
           then debug_ab_machine_env verbose e
           else e)
          p).

  Existing Instance CellAsPos.
  Definition vanalysis (p: program) : block_mon unit :=
    match kind with
    | Poly =>
      let dom := IETME_P.idealenv_to_machineenv
        (reduced_product (VEEQ_P.t * (ZCIE_P.t * (IIE_P.t * PolAdomZ.A))))%type
      in
      analysis (wrap dom _) unroll p
    | ZIntervals =>
      let dom := IETME_AC.idealenv_to_machineenv
        (reduced_product (VEEQ_AC.t * (ZCIE_AC.t * IIE_AC.t)))%type
      in
      analysis dom unroll p
    end.

  Theorem vanalysis_correct :
    forall prog, vanalysis prog = (tt, nil) ->
    forall tr, program_behaves (semantics prog) (Goes_wrong tr) ->
    False.
  Proof.
    unfold vanalysis, doit. intros.
    destruct kind, trace, verbose; eapply iter_program_sound; eauto;
    match goal with
      | |- mem_dom_spec _ (debug_mem_dom _ _) _ => eapply debug_mem_dom_sound
      | _ => idtac
    end;
    eapply mem_cshm_dom_correct.
  Qed.

  Local Notation "⟦ p ⟧" := (program_behaves (semantics p)).
  Import Util.

  Corollary vanalysis_sound:
    ∀ p, vanalysis p = (tt, nil) →
         ⟦p⟧ ⊆ not_wrong.
  Proof.
    intros p H (); try easy.
    intros t H0. exact (vanalysis_correct p H _ H0).
  Qed.

End VA.
