(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib.
Require Import Maps.
Require Import Integers.
Require Import AST.
Require Import Globalenvs.
Require Import Csharpminor.
Require Import AdomLib.
Require Import AbMemSignatureCsharpminor.
Require Import String.
Require Import Debug.
Require Import AnyDefs.
Require Import BlockMon.

Open Scope string_scope.

Axiom fuel : nat. (* Instantiated with let rec fuel = S fuel *)

Section abmem.

Variable abstate: Type.

Variable (unroll: nat).
Variable (prog: program).
Let ge := Genv.globalenv prog.

Variable abmem : mem_dom abstate.

Existing Instance abmem.

Record outcome := ABS {
  onormal: abstate+⊥;
  oreturn: abstate+⊥;
  oexit: list (abstate+⊥);
  ogoto: PTree.t abstate
}.

Definition bot_goto := PTree.empty abstate.

Instance join_goto : join_op (PTree.t abstate) (PTree.t abstate) :=
  PTree.combine (fun a b =>
    match a, b with
      | Some a, Some b => match join a b with NotBot x => Some x | Bot => None end
      | Some x, None | None, Some x => Some x
      | None, None => None
    end).

Instance join_exit : join_op (list (abstate+⊥)) (list (abstate+⊥)) := fix join_exit l1 l2 :=
  match l1 with
    | nil => l2
    | t1::q1 =>
      match l2 with
        | nil => l1
        | t2 :: q2 =>
          let t :=
              match t1, t2 with
                | NotBot t1, NotBot t2 => join t1 t2
                | Bot, x | x, Bot => x
              end
          in
          t :: join_exit q1 q2
      end
  end.

Definition bind_normal_outcome (inormal:abstate+⊥) (f:abstate -> block_mon (abstate+⊥)) : block_mon outcome :=
  match inormal with
    | Bot => ret_bm (ABS Bot Bot nil bot_goto)
    | NotBot inormal =>
      do_bm onormal <- f inormal;
      ret_bm (ABS onormal Bot nil bot_goto)
  end.

Notation "'do_normal' X <- A ; B" := (bind_normal_outcome A (fun X => B))
 (at level 200, X ident, A at level 100, B at level 200).

Definition narrowing_steps := 1%nat.

Fixpoint narrow (f : abstate+⊥ -> block_mon outcome) (steps:nat) (cur:block_mon outcome) : block_mon outcome :=
  match steps with
    | S steps =>
      let curret := (fst cur).(onormal) in
      if is_bot curret then cur
      else
        let next := f curret in
        match snd cur with
          | nil =>
            match snd next with
              | nil => narrow f steps next
              | _ :: _ => cur
            end
          | _::_ =>
            if (fst next).(onormal) ⊑ curret then narrow f steps next
            else cur
        end
    | O => cur
  end.

Definition lfp (f : abstate+⊥ -> block_mon outcome) : block_mon outcome :=
  let fix rec fuel x :=
    match fuel with
      | S fuel =>
        let y := f x in
        let yret := (fst y).(onormal) in
        if yret ⊑ x then narrow f narrowing_steps y
        else rec fuel (widen x yret)
      | O => do_bm _ <- err_bm "Not enough fuel to reach a fixpoint"; f Bot
    end
  in
  rec fuel Bot.

Section stmt_iterator.

Definition ab_builtin (ef:external_function) (rettemp:option ident)
                      (args:list expr) (state:abstate) : block_mon (abstate+⊥) :=
  match ef with
    | EF_external id sig =>
      if ident_eq id any_int64_ident then
        if signature_eq sig any_int64_sig then
          match args with
            | nil =>
              match rettemp with
                | None => ret_bm (NotBot state)
                | Some id => assign_any id TypesDomain.AbTy.VL state
              end
            | _ :: _ =>
              do_bm _ <- err_bm "Parameter used with any_int64"; ret_bm Bot
          end
        else
          do_bm _ <- err_bm "Invalid signature for any_int64"; ret_bm Bot
      else if ident_eq id any_double_ident then
        if signature_eq sig any_double_sig then
         match args with
            | nil =>
              match rettemp with
                | None => ret_bm (NotBot state)
                | Some id => assign_any id TypesDomain.AbTy.VF state
              end
            | _ :: _ =>
              do_bm _ <- err_bm "Parameter used with any_double"; ret_bm Bot
          end
        else
          do_bm _ <- err_bm "Invalid signature for any_double"; ret_bm Bot
      else
        do_bm _ <- err_bm ("Call to external function " ++ string_of_ident id); ret_bm Bot
    | EF_builtin _ _ => do_bm _ <- err_bm "Builtin are not supported"; ret_bm Bot
    | EF_inline_asm _ => do_bm _ <- err_bm "Inline ASM is not supported"; ret_bm Bot
    | EF_annot _ _ | EF_annot_val _ _ =>
    (* TODO: we could use that to generate some certified bounds *)
    (* TODO: don't err if not necessary *)
      do_bm _ <- err_bm "Annotations are not supported";
      ret_bm (NotBot state)
    | EF_memcpy al sz =>
      match args with
        | dst::src::nil =>
          match rettemp with
            | Some _ => do_bm _ <- err_bm "Use of return value of memcpy";
                       ret_bm Bot
            | None => ab_memcpy al sz dst src state
          end
        | _ => do_bm _ <- err_bm "memcpy used without the right number of arguments";
               ret_bm Bot
      end
    | EF_vload _ | EF_vload_global _ _ _
    | EF_vstore _ | EF_vstore_global _ _ _  =>
      do_bm _ <- err_bm "Volatiles are not supported";
      ret_bm Bot
    | EF_malloc =>
      match args with
        | sz::nil => ab_malloc sz rettemp state
        | _ => do_bm _ <- err_bm "malloc used without the right number of arguments";
               ret_bm Bot
      end
    | EF_free =>
      match args with
        | ptr::nil =>
          match rettemp with
            | Some _ => do_bm _ <- err_bm "Use of return value of free";
                        ret_bm Bot
            | None => ab_free ptr state
          end
        | _ => do_bm _ <- err_bm "free used with at least one parameter";
               ret_bm Bot
      end
  end.

Variable iter_function:
  forall (sig:signature)
         (name: ident) (function:fundef) (rettemp:option ident)
         (args:list expr) (state:abstate), block_mon (abstate+⊥).

Variable rettemp: option ident.

Fixpoint default_abstate_switch
         (islong:bool)
         (st:abstate) (e:expr) (s:lbl_stmt) : block_mon (abstate+⊥) :=
  match s with
    | LSnil => ret_bm (NotBot st)
    | LScons None _ s => default_abstate_switch islong st e s
    | LScons (Some n) _ s =>
      if zlt n 0 ||
         zlt (if islong then Int64.max_unsigned else Int.max_unsigned) n then
        default_abstate_switch islong st e s
      else
        do_bm sts <-
          assume
            (if islong then Ebinop (Cminor.Ocmpl Ceq) e (Econst (Olongconst (Int64.repr n)))
             else Ebinop (Cminor.Ocmp Ceq) e (Econst (Ointconst (Int.repr n))))
            st;
        match sts with
          | (_, NotBot st) => default_abstate_switch islong st e s
          | _ => ret_bm Bot
        end
  end.

Definition verasco_unroll_ident := ident_of_string "verasco_unroll"%string.

Fixpoint unroll_call_lookup_aux (s:stmt) : Z :=
  match s with
    | Sseq s1 s2 => Z.max (unroll_call_lookup_aux s1) (unroll_call_lookup_aux s2)
    | Sblock s => unroll_call_lookup_aux s
    | Scall _ _ (Eaddrof id) (Econst (Ointconst i)::nil) =>
      if ident_eq id verasco_unroll_ident then  Int.unsigned i
      else 0
    | _ => 0
  end.

Definition unroll_call_lookup (s:stmt) (unroll:nat) : nat :=
  Z.abs_nat (Z.max (unroll_call_lookup_aux s) (Z.of_nat unroll)).

Fixpoint iter (inormal:abstate+⊥) (igoto:PTree.t abstate) (s: stmt) : block_mon outcome :=
  match s with
  | Sskip => ret_bm (ABS inormal Bot nil bot_goto)
  | Sstore chunk b c =>
    do_normal inormal <- inormal;
    store chunk b c inormal
  | Sset x b =>
    do_normal inormal <- inormal;
    assign x b inormal
  | Scall x sig b cl =>
    do_normal inormal <- inormal;
    do_bm funcs <- deref_fun b inormal;
    (* TODO: Small precision loss here: we forget the relationnal
       information between the function and the rest of the memory *)
    List.fold_left (fun prev fn =>
        do_bm onormal <- iter_function sig (fst fn) (snd fn) x cl inormal;
        do_bm prevnormal <- prev;
        ret_bm (onormal ⊔ prevnormal))
      funcs (ret_bm Bot)
  | Sbuiltin x ef bl =>
    do_normal inormal <- inormal;
    ab_builtin ef x bl inormal
  | Sseq s1 s2 =>
    do_bm o1 <- iter inormal igoto s1;
    do_bm o2 <- iter o1.(onormal) igoto s2;
    ret_bm (ABS o2.(onormal)
                (o1.(oreturn) ⊔ o2.(oreturn))
                (o1.(oexit) ⊔ o2.(oexit))
                (o1.(ogoto) ⊔ o2.(ogoto)))
  | Sifthenelse e s1 s2 =>
    do_bm inormals <-
      match inormal with
        | Bot => ret_bm (Bot, Bot)
        | NotBot inormal => assume e inormal
      end;
    let (inormaltrue, inormalfalse) := inormals in
    do_bm o1 <- iter inormaltrue igoto s1;
    do_bm o2 <- iter inormalfalse igoto s2;
    ret_bm (ABS (o1.(onormal) ⊔ o2.(onormal))
                (o1.(oreturn) ⊔ o2.(oreturn))
                (o1.(oexit) ⊔ o2.(oexit))
                (o1.(ogoto) ⊔ o2.(ogoto)))
  | Sloop s =>
    let unroll := unroll_call_lookup s unroll in
    let fix unroll_loop (n:nat) (inormal:abstate+⊥) (igoto:PTree.t abstate): block_mon outcome :=
        match n with
          | O =>
            do_bm fp <- lfp (fun loopnormal => iter (inormal ⊔ loopnormal) igoto s);
            ret_bm (ABS Bot fp.(oreturn) fp.(oexit) fp.(ogoto))
          | S n =>
            do_bm o1 <- iter inormal igoto s;
            match o1.(onormal) with
              | Bot =>
                ret_bm (ABS Bot o1.(oreturn) o1.(oexit) o1.(ogoto))
              | _ =>
                do_bm o2 <- unroll_loop n o1.(onormal) bot_goto;
                ret_bm (ABS Bot
                            (o1.(oreturn) ⊔ o2.(oreturn))
                            (o1.(oexit) ⊔ o2.(oexit))
                            (o1.(ogoto) ⊔ o2.(ogoto)))
            end
        end in
    unroll_loop unroll inormal igoto
  | Sblock s =>
    do_bm o <- iter inormal igoto s;
    ret_bm (ABS (o.(onormal) ⊔ List.hd Bot o.(oexit))
                o.(oreturn)
                (List.tl o.(oexit))
                o.(ogoto))
  | Sexit n =>
    ret_bm (ABS Bot Bot (nat_iter n (cons Bot) (cons inormal nil)) bot_goto)
  | Sreturn ret =>
    match inormal with
      | Bot => ret_bm (ABS Bot Bot nil bot_goto)
      | NotBot inormal =>
        do_bm oreturn <- pop_frame ret rettemp inormal;
        ret_bm (ABS Bot oreturn nil bot_goto)
    end
  | Sswitch islong a sl =>
    do_bm idefault <-
      match inormal with
        | Bot => ret_bm Bot
        | NotBot inormal =>
          do_bm _ <- nonblock
                (Eunop (if islong then Cminor.Onegl else Cminor.Onegint) a) inormal;
          default_abstate_switch islong inormal a sl
      end;
    iter_switch islong inormal idefault Bot igoto a sl
  | Slabel lbl s =>
    iter (inormal ⊔
            match PTree.get lbl igoto with
              | None => Bot
              | Some ab => NotBot ab
            end) igoto s
  | Sgoto lbl =>
    ret_bm (ABS Bot Bot nil
                (match inormal with
                   | Bot => bot_goto
                   | NotBot inormal => PTree.set lbl inormal bot_goto
                 end))
  end

with iter_switch (islong:bool) (inormal:abstate+⊥) (idefault iprev:abstate+⊥)
                 (igoto:PTree.t abstate) (e:expr) (sl: lbl_stmt)
     : block_mon outcome :=
  match sl with
  | LSnil => ret_bm (ABS (iprev ⊔ idefault) Bot nil bot_goto)
  | LScons o s sl' =>
    do_bm icaseidefault <-
      match o with
        | None => ret_bm (idefault, Bot)
        | Some n =>
          do_bm icase <-
            match inormal with
              | Bot => ret_bm Bot
              | NotBot inormal =>
                do_bm abs <- assume
                  (if islong then Ebinop (Cminor.Ocmpl Ceq) e (Econst (Olongconst (Int64.repr n)))
                   else Ebinop (Cminor.Ocmp Ceq) e (Econst (Ointconst (Int.repr n))))
                  inormal;
                ret_bm (fst abs)
            end;
          ret_bm (icase, idefault)
      end;
    let '(icase, idefault) := icaseidefault in
    do_bm o1 <- iter (iprev ⊔ icase) igoto s;
    do_bm o2 <- iter_switch islong inormal idefault o1.(onormal) igoto e sl';
    ret_bm (ABS o2.(onormal)
                (o1.(oreturn) ⊔ o2.(oreturn))
                (o1.(oexit) ⊔ o2.(oexit))
                (o1.(ogoto) ⊔ o2.(ogoto)))
  end.

End stmt_iterator.

Instance leb_goto : leb_op (PTree.t abstate) := fun a b =>
  let t := PTree.combine (fun na nb =>
    match na, nb with
      | None, _ => None
      | Some _, None => Some tt
      | Some a, Some b => if a ⊑ b then None else Some tt
    end)
    a b
  in
  PTree.bempty t.

Fixpoint narrow_goto (f:PTree.t abstate -> block_mon outcome) (steps:nat) (cur:block_mon outcome) : block_mon outcome :=
  match steps with
    | S steps =>
      let curret := (fst cur).(ogoto) in
      if leb_goto curret bot_goto then cur
      else
        let next := f curret in
        match snd cur with
          | nil =>
            match snd next with
              | nil => narrow_goto f steps next
              | _ :: _ => cur
            end
          | _ :: _ =>
            if leb_goto (fst next).(ogoto) curret then narrow_goto f steps next
            else cur
        end
    | O => cur
  end.

Instance widen_goto : widen_op (PTree.t abstate) (PTree.t abstate) :=
  PTree.combine
    (fun nx ny:option abstate =>
       match widen (match nx with Some x => NotBot x | None => Bot end)
                   (match ny with Some y => NotBot y | None => Bot end)
       with Bot => None | NotBot x => Some x end).

Definition lfp_goto (f : PTree.t abstate -> block_mon outcome) : block_mon outcome :=
  let fix rec fuel x :=
      match fuel with
        | S fuel =>
          let y := f x in
          let yret := (fst y).(ogoto) in
          if leb_goto yret x then narrow_goto f narrowing_steps y
          else rec fuel (widen_goto x yret)
        | O => do_bm _ <- err_bm "Not enough fuel to compute goto fixpoints";
               f bot_goto (* Dummy, TODO ? *)
      end
  in
  rec fuel bot_goto.

Fixpoint labels_stmt (s:stmt) : PTree.t unit :=
  match s with
    | Sskip | Sstore _ _ _ | Sset _ _ | Scall _ _ _ _ | Sbuiltin _ _ _
    | Sexit _ | Sreturn _ | Sgoto _ => PTree.empty unit
    | Sseq s1 s2 | Sifthenelse _ s1 s2 =>
      PTree.combine (fun x y => match x, y with None, None => None | _, _ => Some tt end)
        (labels_stmt s1) (labels_stmt s2)
    | Sloop s | Sblock s => labels_stmt s
    | Sswitch _ _ s => labels_sl s
    | Slabel lbl s => PTree.set lbl tt (labels_stmt s)
  end
with labels_sl (s:lbl_stmt) : PTree.t unit :=
  match s with
    | LSnil => PTree.empty unit
    | LScons _ s1 s2 =>
      PTree.combine (fun x y => match x, y with None, None => None | _, _ => Some tt end)
        (labels_stmt s1) (labels_sl s2)
  end.

Fixpoint iter_function (fuel:nat) (sig:signature) (name: ident) (fn:fundef) (rettemp:option ident)
                       (params:list expr) (istate:abstate) : block_mon (abstate+⊥) :=
  match fuel with
    | O => do_bm _ <- err_bm "Not enough fuel for call stack depth";
           ret_bm Bot
    | S fuel =>
      if signature_eq sig (funsig fn) then
        match fn with
          | Internal fn =>
              do_bm _ <-
                if list_norepet_dec Pos.eq_dec (List.map fst fn.(fn_vars)) then ret_bm tt
                else err_bm "Repeating fn_vars";
              do_bm _ <-
                if list_norepet_dec Pos.eq_dec (fn.(fn_params)) then ret_bm tt
                else err_bm "Repeating fn_params";
              do_bm _ <-
                if list_disjoint_dec Pos.eq_dec (fn.(fn_params)) (fn.(fn_temps)) then ret_bm tt
                else err_bm "Non disjoint fn_params and fn_temps";
              do_bm inormal <- push_frame name fn params istate;
              do_bm o <- lfp_goto
                (fun igoto => iter (iter_function fuel) rettemp inormal igoto fn.(fn_body));
              do_bm _ <-
                match o.(oexit) with
                  | nil => ret_bm tt
                  | _::_ => err_bm "Non-scoped exit"
                end;
              (* We check that all gotos are well defined *)
              do_bm _ <-
                if PTree.bempty
                     (PTree.combine (fun x y => match x, y with Some _, None => Some tt | _, _ => None end)
                                    o.(ogoto) (labels_stmt fn.(fn_body)))
                then ret_bm tt
                else err_bm "goto with unbound label";
                  do_bm end_return_state <-
                    match o.(onormal) with
                      | Bot => ret_bm Bot
                      | NotBot onormal => pop_frame None rettemp onormal
                    end;
                  ret_bm (o.(oreturn) ⊔ end_return_state)
          | External ext => ab_builtin ext rettemp params istate
        end
      else
        do_bm _ <- err_bm "Could not prove that the called function has the right signature.";
        ret_bm Bot
  end.

Definition iter_program : block_mon unit :=
  do_bm mem <- init_mem prog.(prog_defs);
  match mem with
    | Bot => ret_bm tt
    | NotBot mem =>
      match Genv.find_symbol ge prog.(prog_main) with
        | None => err_bm "Could not find main symbol"
        | Some b =>
          match Genv.find_funct_ptr ge b with
            | None => err_bm "Could not find the code of the main function"
            | Some fn =>
              do_bm end_mem <- iter_function fuel
                (mksignature nil (Some AST.Tint) cc_default)
                prog.(prog_main) fn (Some 1%positive) nil mem;
              ret_bm tt
          end
      end
  end.

End abmem.
