(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String ToString
  AdomLib
  AbMachineEnv
  Debug DebugCminor.

Section S.

  Local Open Scope string_scope.

  Context {var t : Type}
          {VS: ToString var}
          {TS: ToString t}
          (verbose: bool)
          `(D: ab_machine_env var t).

  Definition string_of_num_val_ty ty :=
    match ty with
      | NVTint => "%int"
      | NVTint64 => "%int64"
      | NVTfloat => "%float"
      | NVTsingle => "%single"
    end.

  Instance NVI_ToString : ToString num_val_itv := fun itv =>
    match itv with
      | NVIint itv => to_string itv ++ "%int"
      | NVIint64 itv => to_string itv ++ "%int64"
      | NVIfloat itv => to_string itv ++ "%float"
      | NVIsingle itv => to_string itv ++ "%single"
    end.

  Fixpoint string_of_nexpr (e: nexpr var) {struct e} : string :=
    match e with
    | NEvar ty v => to_string v ++ string_of_num_val_ty ty
    | NEconst itv => to_string itv
    | NEunop op e' => string_of_unop op ++ string_of_nexpr e'
    | NEbinop op x y => "(" ++ string_of_nexpr x ++ string_of_binop op ++ string_of_nexpr y ++ ")"
    end.

  Global Instance NExprToString : ToString (nexpr var) := string_of_nexpr.

  Definition trace_concretize_int e res :=
    "concretize_int " ++ e ++ " → " ++ res.

  Definition trace_assign x e (r:t+⊥) :=
    "assign " ++ x ++ " := " ++ e ++ if verbose then " → " ++ to_string r else "".

  Definition trace_forget x (r:t+⊥) :=
    "forget " ++ x ++ if verbose then " → " ++ to_string r else "".

  Definition trace_assume e b (r:t+⊥) :=
    "assume " ++ e ++ " " ++ b ++ if verbose then " → " ++ to_string r else "".

  Definition trace_nonblock e r :=
    "nonblock " ++ e ++ " → " ++ r.

  Program Definition debug_ab_machine_env : ab_machine_env t :=
    {| mach_leb := leb_print
    ; mach_join := join_print
    ; mach_widen := widen_print
    ; concretize_int n e ab :=
        let r := concretize_int n e ab in
        let s := trace_concretize_int (to_string e) (to_string r) in
        print s r
    ; assign x e ab :=
        let r := assign x e ab in
        let s := trace_assign (to_string x) (to_string e) r in
        print s r
    ; assume e b ab :=
        let r := assume e b ab in
        let s := trace_assume (to_string e) (to_string b) r in
        print s r
    ; forget b ab :=
        let r := forget b ab in
        let s := trace_forget (to_string b) r in
        print s r
    ; nonblock e ab :=
        let r := nonblock e ab in
        let s := trace_nonblock (string_of_nexpr e) (to_string r) in
        print s r
    |}.
  Next Obligation. rewrite print_id; eapply concretize_int_correct; eauto. Qed.
  Next Obligation. rewrite print_id; eapply assign_correct; eauto. Qed.
  Next Obligation. rewrite print_id; eapply forget_correct; eauto. Qed.
  Next Obligation. rewrite print_id; eapply assume_correct; eauto. Qed.
  Next Obligation. rewrite print_id in H1; eapply nonblock_correct; eauto. Qed.

End S.
