(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import ShareTree.
Require Import IdealExpr.
Require Import ZArith.
Require Import Coqlib ToString.
Require Import ZIntervals FloatIntervals.
Require Import AdomLib.

Inductive abs :=
| Az : ZIntervals.itv -> abs
| Af : FloatIntervals.itv -> abs.

Instance iditv_leb : leb_op abs := fun x y =>
  match x, y return _ with
    | Az x, Az y => x ⊑ y
    | Af x, Af y => x ⊑ y
    | Af _, Az _ | Az _, Af _ => false
  end.

Program Instance iditv_join : join_op abs (abs+⊤) :=  fun x y =>
  match x, y with
    | Az x', Az y' =>
      let res := x' ⊔ y' in
      Just (physEq x' res (fun _ => x) (fun _ =>
              physEq y' res (fun _ => y) (fun _ => Az res) _) _)
    | Af x', Af y' =>
      let res := x' ⊔ y' in
      Just (physEq x' res (fun _ => x) (fun _ =>
              physEq y' res (fun _ => y) (fun _ => Af res) _) _)
    | Af _, Az _ | Az _, Af _ => All
  end.
Next Obligation. congruence. Qed.
Next Obligation. unfold physEq. congruence. Qed.
Next Obligation. congruence. Qed.
Next Obligation. unfold physEq. congruence. Qed.

Program Instance iditv_widen : widen_op abs (abs+⊤) :=  fun x y =>
  match x, y with
    | Az x', Az y' =>
      do_top res <- widen x' y';
      Just (physEq x' res (fun _ => x) (fun _ =>
              physEq y' res (fun _ => y) (fun _ => Az res) _) _)
    | Af x', Af y' =>
      let res := widen x' y' in
      Just (physEq x' res (fun _ => x) (fun _ =>
              physEq y' res (fun _ => y) (fun _ => Af res) _) _)
    | Af _, Az _ | Az _, Af _ => All
  end.
Next Obligation. congruence. Qed.
Next Obligation. unfold physEq. congruence. Qed.

Instance itv_num_gamma : gamma_op abs ideal_num := fun v x =>
   match v, x with
     | Az v, INz x => γ v x
     | Af v, INf x => γ v x
     | _, _ => False
   end.

Instance iditv_leb_correct : leb_op_correct abs ideal_num.
Proof.
  repeat intro. destruct a1, a2, a; try easy; simpl in *.
  eapply leb_correct; eauto.
  eapply leb_correct; eauto.
Qed.

Instance iditv_join_correct : join_op_correct abs (abs+⊤) ideal_num.
Proof.
  repeat intro. destruct x, y, a; try easy; simpl in *; try now intuition; apply join_correct; auto.
Qed.

Definition in_z v : ZIntervals.itv+⊤+⊥ :=
  match v with
    | Just (Az v) => NotBot (Just v)
    | Just (Af v) => Bot
    | All => NotBot All
  end.

Lemma in_z_correct:
  ∀ ab x, γ ab (INz x) -> γ (in_z ab) x.
Proof.
  destruct ab as [|[]]; simpl; auto.
Qed.

Hint Resolve in_z_correct.

Definition in_f v : FloatIntervals.itv+⊤+⊥ :=
  match v with
    | Just (Af v) => NotBot (Just v)
    | Just (Az v) => Bot
    | All => NotBot All
  end.

Lemma in_f_correct:
  ∀ ab x, γ ab (INf x) -> γ (in_f ab) x.
Proof.
  destruct ab as [|[]]; simpl; auto.
Qed.

Hint Resolve in_f_correct.

Instance AbsToString : ToString abs := λ i,
  match i with
    | Az j => to_string j
    | Af j => to_string j
  end.
