(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8.
Require Import Maps ShareTree Util.
Require Import Integers.
Require Import Floats FloatLib.
Require Import Values.
Require Import AST.
Require Import Globalenvs.
Require Import Cminor.
Require Import Csharpminor.
Require Import Memory.
Require Import AdomLib.
Require Import Util.
Require Import PExpr.
Require Import TypesDomain.
Require Import AbMachineEnv.
Require Import String ToString.

Local Unset Elimination Schemes.

Local Open Scope string_scope.

Module Nconvert(T:SHARETREE).

  Module Eval := Cell2Val.Eval T.

  Module TDom := Eval.TDom.

  Instance top_tree : top_op (TDom.t AbTy.t) := TDom.top_tree _.
  Instance join_tree : join_op (TDom.t AbTy.t) (TDom.t AbTy.t) := TDom.join_tree _ _.
  Proof. abstract (destruct x; auto). Defined.
  Instance widen_tree : widen_op (TDom.t AbTy.t) (TDom.t AbTy.t) := TDom.widen_tree _ _.
  Proof. abstract (destruct x; auto). Defined.
  Instance leb_tree : leb_op (TDom.t AbTy.t) := TDom.leb_tree _ _.
  Proof. abstract (destruct x; auto). Defined.

  Instance tree_gamma : gamma_op (TDom.t AbTy.t) (T.elt -> val) :=
    TDom.gamma_tree _.

  Instance top_tree_correct : top_op_correct (TDom.t AbTy.t) (T.elt -> val) :=
    TDom.top_tree_correct _.
  Instance join_tree_correct : join_op_correct (TDom.t AbTy.t) (TDom.t AbTy.t) (T.elt -> val) :=
    TDom.join_tree_correct _ _.
  Instance leb_tree_correct : leb_op_correct (TDom.t AbTy.t)  (T.elt -> val) :=
    TDom.leb_tree_correct _.

  Section nconvert.
    Variable ge: genv.
    Variable ε: ident → option block.
    Variable var : Type.
    Variable m: Mem.mem.
    Variable ρC: T.elt -> val.
    Variable ρ: var -> num_val.
    Variable tp: TDom.t AbTy.t.
    Variable tp_correct: γ tp ρC.

    Variable elt2p: T.elt -> var.

    Definition convert_constant (cst: constant) : nexpr var :=
      match cst with
        | Ointconst n => NEconst _ (NVIint (Just (ZIntervals.ITV (Int.signed n) (Int.signed n))))
        | Olongconst n => NEconst _ (NVIint64 (Just (ZIntervals.ITV (Int64.signed n) (Int64.signed n))))
        | Ofloatconst f =>
          if Fappli_IEEE.is_nan _ _ f then
            NEconst _ (NVIfloat All)
          else
            NEconst _ (NVIfloat (Just (FloatIntervals.ITV f f)))
        | Osingleconst f =>
          if Fappli_IEEE.is_nan _ _ f then
            NEconst _ (NVIsingle All)
          else
            let f := Float.of_single f in
            NEconst _ (NVIsingle (Just (FloatIntervals.ITV f f)))
        | Oaddrsymbol _ ofs =>
          let n := Int.unsigned ofs in
          NEconst _ (NVIint (Just (ZIntervals.ITV n n)))
      end.

    Import AbTy.

    Let type_expr e := let x := Eval.eval_expr AbTy.adom tp e in x.

    Inductive nconvert_msg_kind : Set :=
    | MSG_TYPE
    | MSG_CONST_NONE
    | MSG_ILL_TYPED_UNOP
    | MSG_ILL_TYPED_BINOP
    .

    Definition nconvert_msg_type (k: nconvert_msg_kind) : Type :=
      match k with
      | MSG_TYPE
      | MSG_CONST_NONE
        => string
      | MSG_ILL_TYPED_UNOP => Cminor.unary_operation → AbTy.t+⊤ → string
      | MSG_ILL_TYPED_BINOP => Cminor.binary_operation → AbTy.t+⊤ → AbTy.t+⊤ → string
      end.

    Definition nconvert_msg (k: nconvert_msg_kind) : nconvert_msg_type k :=
      match k return nconvert_msg_type k with
      | MSG_TYPE => "var of unknown type"
      | MSG_CONST_NONE => "const None"
      | MSG_ILL_TYPED_UNOP => λ op ty, "ill-typed op: " ++ to_string op ++ to_string ty
      | MSG_ILL_TYPED_BINOP => λ op ty₁ ty₂, "ill-typed op: " ++ to_string ty₁ ++ to_string op ++ to_string ty₂
      end.
    Opaque nconvert_msg.

    Fixpoint nconvert (e:pexpr T.elt) : nexpr var + string :=
      match e with
        | PEvar c =>
          match TDom.get c tp with
            | Just (VI|VP|VIP) => inl (NEvar NVTint (elt2p c))
            | Just VL => inl (NEvar NVTint64 (elt2p c))
            | Just VF => inl (NEvar NVTfloat (elt2p c))
            | Just VS => inl (NEvar NVTsingle (elt2p c))
            | All => inr (nconvert_msg MSG_TYPE)
          end
        | PElocal _ => inl (NEconst _ (NVIint (Just (ZIntervals.ITV 0 0))))
        | PEconst (Some cst) => inl (convert_constant cst)
        | PEconst None => inr (nconvert_msg MSG_CONST_NONE)
        | PEunop op e =>
          match type_expr e, op with
            | Just VI,
              (Ocast8unsigned | Ocast8signed | Ocast16unsigned | Ocast16signed
               | Onegint | Onotint | Ofloatofint | Ofloatofintu
               | Osingleofint | Osingleofintu | Olongofint | Olongofintu )
            | Just VL,
              (Onegl | Onotl | Ointoflong | Ofloatoflong | Ofloatoflongu
               | Osingleoflong | Osingleoflongu)
            | Just VF,
              (Onegf | Oabsf | Osingleoffloat | Ointoffloat | Ointuoffloat
               | Olongoffloat | Olonguoffloat)
            | Just VS,
              (Onegfs | Oabsfs | Ofloatofsingle | Ointofsingle | Ointuofsingle
               | Olongofsingle | Olonguofsingle) =>
              match nconvert e with
                | inl e' => inl (NEunop op e')
                | q => q
              end
            | ty, op => inr (nconvert_msg MSG_ILL_TYPED_UNOP op ty)
          end
        | PEbinop op e1 e2 =>
          match type_expr e1, type_expr e2, op with
            | Just VI, Just VP, Oadd | Just VP, Just VI, Oadd
            (* | Just VP, Just VP, Osub *) (* TODO: subtraction of pointers *)
            | Just VP, Just VI, Osub
            | Just VI, Just VI, Ocmpu _
            | Just VP, Just VP, Ocmpu _
            | Just VI, Just VI,
              (Oadd | Osub | Omul | Odiv | Odivu | Omod | Omodu | Oand
               | Oor | Oxor | Oshl | Oshr | Oshru | Ocmp _)
            | Just VL, Just VL,
              (Oaddl | Osubl | Omull | Odivl | Odivlu | Omodl | Omodlu
               | Oandl | Oorl | Oxorl | Ocmpl _ | Ocmplu _)
            | Just VL, Just VI, (Oshll | Oshrl | Oshrlu)
            | Just VF, Just VF, (Oaddf | Osubf | Omulf | Odivf | Ocmpf _)
            | Just VS, Just VS, (Oaddfs | Osubfs | Omulfs | Odivfs | Ocmpfs _) =>
              match nconvert e1, nconvert e2 with
                | inl e1', inl e2' => inl (NEbinop op e1' e2')
                | inr q, _ => inr q
                | _, q => q
              end
            | ty₁, ty₂, op => inr (nconvert_msg MSG_ILL_TYPED_BINOP op ty₁ ty₂)
          end
      end.

    Lemma bind_inl {A B} (f: A → A + B) (ab: A + B) (a': A):
      match ab with inl a' => f a' | inr b => inr b end = inl a' →
      ∃ a : A, ab = inl a ∧ f a = inl a'.
    Proof.
      destruct ab as [ a | b ]. vauto. intro; eq_some_inv.
    Qed.

    Local Ltac ncinv :=
      try match goal with H : _ = Some _ |- _ => apply do_opt_some_inv in H; hsplit end;
      repeat match goal with H : _ = inl _ |- _ => apply bind_inl in H; hsplit end;
      eq_some_inv; subst; simpl;
      repeat match goal with H : _, K : _ → _ |- _ => specialize (K _ H) end.

    Lemma nconvert_fv_have_type (pe: pexpr _) :
      ∀ ne, nconvert pe = inl ne →
      ∀ c, c ∈ pe_free_var pe → TDom.get c tp ≠ All.
    Proof.
      induction pe; simpl; intros ne H c Hc K; auto.
      subst c. rewrite K in H. eq_some_inv.
      destruct (type_expr pe) as [|()]; eq_some_inv;
      destruct u; eq_some_inv;
      ncinv; auto.
      destruct (type_expr pe1) as [|()]; eq_some_inv;
      destruct (type_expr pe2) as [|()]; eq_some_inv;
      destruct b; eq_some_inv;
      ncinv; intuition; ncinv; auto.
    Qed.

    Lemma nconvert_fv pe :
      ∀ ne, nconvert pe = inl ne →
      pe_free_var pe ⊆ (ne_free_var ne ∘ elt2p).
    Proof.
      induction pe; simpl; intros ne Hne c Hc.
      subst c.
      destruct (TDom.get v tp) as [|()]; eq_some_inv; subst ne; reflexivity.
      destruct Hc. destruct Hc.
      destruct (type_expr pe) as [|()]; eq_some_inv;
      destruct u; eq_some_inv; ncinv; eq_some_inv; auto.
      destruct (type_expr pe1) as [|()]; eq_some_inv;
      destruct (type_expr pe2) as [|()]; eq_some_inv;
      destruct b; eq_some_inv;
      ncinv;
      intuition.
    Qed.

    Lemma nconvert_fv' pe :
      ∀ ne, nconvert pe = inl ne →
      ∀ x, x ∈ ne_free_var ne → ∃ y, y ∈ pe_free_var pe ∧ x = elt2p y.
    Proof.
      induction pe; simpl; intros ne Hne c Hc.
      destruct (TDom.get v tp) as [|()]; eq_some_inv; subst ne; simpl in Hc; subst c; vauto.
      eq_some_inv; subst ne. destruct Hc.
      destruct o as [()|]; eq_some_inv; subst ne; simpl in *; try (destruct Hc; fail);
      destruct Fappli_IEEE.is_nan, Hc.
      destruct (type_expr pe) as [|()]; eq_some_inv;
      destruct u; eq_some_inv; ncinv; eq_some_inv; auto.
      destruct (type_expr pe1) as [|()]; eq_some_inv;
      destruct (type_expr pe2) as [|()]; eq_some_inv;
      destruct b; eq_some_inv;
      ncinv; simpl in Hc;
      destruct Hc; ncinv; hsplit; vauto.
    Qed.

    Inductive ncompat : val -> num_val -> Prop :=
    | ncompat_int: forall i, ncompat (Vint i) (NVint i)
    | ncompat_ptr: forall b i, ncompat (Vptr b i) (NVint i)
    | ncompat_int64: forall i, ncompat (Vlong i) (NVint64 i)
    | ncompat_vundef: forall i, ncompat Vundef i
    | ncompat_float: forall f, ncompat (Vfloat f) (NVfloat f)
    | ncompat_single: forall f, ncompat (Vsingle f) (NVsingle f).
    Hint Constructors ncompat.

    Lemma ncompat_inv_int v i :
      ncompat v (NVint i) →
      v = Vint i ∨ (∃ b, v = Vptr b i) ∨ v = Vundef.
    Proof.
      intros H.
      set (d w k := match k with
                    | NVint j => w = Vint j ∨ (∃ b, w = Vptr b j) ∨ w = Vundef | _ => True end).
      change (d v (NVint i)). destruct H; simpl; eauto.
      destruct i0; simpl; eauto.
    Qed.

    Lemma ncompat_inv_int64 v i :
      ncompat v (NVint64 i) →
      v = Vlong i ∨ v = Vundef.
    Proof.
      intros H.
      set (d w k := match k with
                    | NVint64 j => w = Vlong j ∨ w = Vundef | _ => True end).
      change (d v (NVint64 i)). destruct H; simpl; eauto.
      destruct i0; simpl; eauto.
    Qed.

    Lemma ncompat_inv_float v f :
      ncompat v (NVfloat f) →
      v = Vfloat f ∨ v = Vundef.
    Proof.
      intros H.
      set (d w k := match k with
                    | NVfloat j => w = Vfloat j ∨ w = Vundef | _ => True end).
      change (d v (NVfloat f)). destruct H; simpl; eauto.
      destruct i; simpl; eauto.
    Qed.

    Lemma ncompat_inv_single v f :
      ncompat v (NVsingle f) →
      v = Vsingle f ∨ v = Vundef.
    Proof.
      intros H.
      set (d w k := match k with
                    | NVsingle j => w = Vsingle j ∨ w = Vundef | _ => True end).
      change (d v (NVsingle f)). destruct H; simpl; eauto.
      destruct i; simpl; eauto.
    Qed.

    Lemma ncompat_inv_int' f v :
      ncompat (Vint f) v →
      v = NVint f.
    Proof.
      intros H.
      set (d w k := match w with
                    | Vint j => k = NVint j | _ => True end).
      change (d (Vint f) v). destruct H; simpl; eauto.
    Qed.

    Lemma ncompat_inv_long' f v :
      ncompat (Vlong f) v →
      v = NVint64 f.
    Proof.
      intros H.
      set (d w k := match w with
                    | Vlong j => k = NVint64 j | _ => True end).
      change (d (Vlong f) v). destruct H; simpl; eauto.
    Qed.

    Lemma ncompat_inv_ptr' b i v :
      ncompat (Vptr b i) v →
      v = NVint i.
    Proof.
      intros H.
      set (d w k := match w with
                    | Vptr b j => k = NVint j | _ => True end).
      change (d (Vptr b i) v). destruct H; simpl; eauto.
    Qed.

    Lemma ncompat_inv_float' f v :
      ncompat (Vfloat f) v →
      v = NVfloat f.
    Proof.
      intros H.
      set (d w k := match w with
                    | Vfloat j => k = NVfloat j | _ => True end).
      change (d (Vfloat f) v). destruct H; simpl; eauto.
    Qed.

    Lemma ncompat_inv_single' f v :
      ncompat (Vsingle f) v →
      v = NVsingle f.
    Proof.
      intros H.
      set (d w k := match w with
                    | Vsingle j => k = NVsingle j | _ => True end).
      change (d (Vsingle f) v). destruct H; simpl; eauto.
    Qed.

    Ltac ncompat_inv H :=
      match type of H with
             | ncompat ?v (NVint ?i) =>
               destruct (ncompat_inv_int v i H) as [?|[(? & ?)|?]];
               clear H
             | ncompat ?v (NVint64 ?i) =>
               destruct (ncompat_inv_int64 v i H) as [?|?];
               clear H
             | ncompat ?v (NVfloat ?i) =>
               destruct (ncompat_inv_float v i H) as [?|?];
               clear H
             | ncompat ?v (NVsingle ?i) =>
               destruct (ncompat_inv_single v i H) as [?|?];
               clear H
             end.

    Ltac ncompat_inv' H :=
      match type of H with
      | ncompat (Vint ?i) ?v => apply (ncompat_inv_int' i v) in H
      | ncompat (Vlong ?i) ?v => apply (ncompat_inv_long' i v) in H
      | ncompat (Vptr ?b ?i) ?v => apply (ncompat_inv_ptr' b i v) in H
      | ncompat (Vfloat ?i) ?v => apply (ncompat_inv_float' i v) in H
      | ncompat (Vsingle ?i) ?v => apply (ncompat_inv_single' i v) in H
      end.

    Lemma ncompat_compat_fun v :
      ncompat v (compat_fun v).
    Proof.
      destruct v; simpl; constructor.
    Qed.

    Definition wtype_num (v:num_val) (tp:AbTy.t+⊤) : Prop :=
      match v, tp with
        | NVint _, Just (VI|VP|VIP)
        | NVint64 _, Just VL
        | NVfloat _, Just VF
        | NVsingle _, Just VS => True
        | _, All => True
        | _, _ => False
      end.

    Lemma wtype_num_top v : wtype_num v All.
    Proof. now destruct v. Qed.

    Definition compat (ρC:T.elt -> val) (ρ:var -> num_val) : Prop :=
      forall c, ncompat (ρC c) (ρ (elt2p c)) /\
                wtype_num (ρ (elt2p c)) (type_expr (PEvar c)).

    Variable Hcompat: compat ρC ρ.
    Hint Constructors is_bool.

    Lemma eval_pexpr_type_expr_ptr_VI: forall a b i,
      eval_pexpr ge m ρC ε a (Vptr b i) ->
      type_expr a = Just VI -> False.
    Proof.
      unfold type_expr. intros a b i H H0.
      generalize (@Eval.eval_expr_correct _ adom _ _ ge m ε tp ρC a _ tp_correct H).
      rewrite H0. refine (λ Hg, Hg _).
      discriminate.
    Qed.

    Lemma eval_pexpr_type_expr_int_VP: forall a i,
      eval_pexpr ge m ρC ε a (Vint i) ->
      type_expr a = Just VP -> False.
    Proof.
      unfold type_expr; intros a i H H0.
      generalize (@Eval.eval_expr_correct _ adom _ _ ge m ε tp ρC a _ tp_correct H).
      rewrite H0. refine (λ Hg, Hg _).
      discriminate.
    Qed.

    Lemma eval_pexpr_type_expr_float_VI: forall a f,
      eval_pexpr ge m ρC ε a (Vfloat f) ->
      type_expr a = Just VI -> False.
    Proof.
      unfold type_expr; intros a f H H0.
      generalize (@Eval.eval_expr_correct _ adom _ _ ge m ε tp ρC a _ tp_correct H).
      rewrite H0. refine (λ Hg, Hg _).
      discriminate.
    Qed.

    Lemma val_cmp_exists_int: forall c i1 i2,
      exists i,
        Val.cmp c (Vint i1) (Vint i2) = Vint i /\
        of_bool (Int.cmp c i1 i2) = (NVint i).
    Proof.
      unfold Val.cmp, of_bool, Val.cmp_bool; intros.
      destruct Int.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpu_exists_int: forall f c i1 i2,
      exists i,
        Val.cmpu f c (Vint i1) (Vint i2) = Vint i /\
        of_bool (Int.cmpu c i1 i2) =  NVint i.
    Proof.
      unfold Val.cmpu, of_bool, Val.cmpu_bool; intros.
      destruct Int.cmpu; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpf_exists_int: forall c f1 f2,
      exists i,
        Val.cmpf c (Vfloat f1) (Vfloat f2) = Vint i /\
        of_bool (Float.cmp c f1 f2) =  NVint i.
    Proof.
      unfold Val.cmpf, of_bool, Val.cmpf_bool; intros.
      destruct Float.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpfs_exists_int: forall c f1 f2,
      exists i,
        Val.cmpfs c (Vsingle f1) (Vsingle f2) = Vint i /\
        of_bool (Float32.cmp c f1 f2) =  NVint i.
    Proof.
      unfold Val.cmpfs, of_bool, Val.cmpfs_bool; intros.
      destruct Float32.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_of_bool_exists_int: forall b,
      exists i,
        Val.of_bool b = Vint i /\
        of_bool b = NVint i.
    Proof.
      unfold of_bool; intros.
      destruct b; repeat econstructor.
    Qed.

    Lemma ncompat_Vundef: forall n, ncompat Vundef n.
    Proof.
      destruct n; constructor.
    Qed.
    Hint Resolve ncompat_Vundef.

    Lemma ncompat_Vfloat: forall f, ncompat (Vfloat f) (NVfloat f).
    Proof.
      constructor.
    Qed.
    Hint Resolve ncompat_Vfloat.

    Lemma ncompat_Vsingle: forall f, ncompat (Vsingle f) (NVsingle f).
    Proof.
      constructor.
    Qed.
    Hint Resolve ncompat_Vsingle.

    Lemma ncompat_eval_nexpr: forall e v,
      match v with
        | Vint i => eval_nexpr ρ e (NVint i)
        | Vptr b i => eval_nexpr ρ e (NVint i)
        | Vlong i => eval_nexpr ρ e (NVint64 i)
        | Vundef => exists n, eval_nexpr ρ e n
        | Vfloat f => eval_nexpr ρ e (NVfloat f)
        | Vsingle f => eval_nexpr ρ e (NVsingle f)
      end ->
      exists n : num_val,
        ncompat v n /\  eval_nexpr ρ e n.
    Proof.
      destruct v; try (econstructor; split; eauto; fail).
      destruct 1; eauto.
    Qed.

    Lemma type_expr_sound : forall e v,
      eval_pexpr ge m ρC ε e v ->
      v ≠ Vundef →
      γ (type_expr e) v.
    Proof.
      intros e v H K.
      exact (@Eval.eval_expr_correct _ adom _ _ ge m ε tp ρC e _ tp_correct H K).
    Qed.

    Lemma exist_ncompat_Vundef: forall P,
      (exists n, P n) ->
      exists n : num_val, ncompat Vundef n /\ P n.
    Proof.
      intros P (n & H).
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vint: forall (P:num_val -> Prop) i,
      P (NVint i) ->
      exists n : num_val, ncompat (Vint i) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vptr: forall (P:num_val -> Prop) b i,
      P (NVint i) ->
      exists n : num_val, ncompat (Vptr b i) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vlong: forall (P:num_val -> Prop) l,
      P (NVint64 l) ->
      exists n : num_val, ncompat (Vlong l) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vfloat: forall (P:num_val -> Prop) f,
      P (NVfloat f) ->
      exists n : num_val, ncompat (Vfloat f) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vsingle: forall (P:num_val -> Prop) f,
      P (NVsingle f) ->
      exists n : num_val, ncompat (Vsingle f) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma of_bool_NVint : forall b,
      exists i, of_bool b = NVint i.
    Proof.
      unfold of_bool; destruct b; econstructor; reflexivity.
    Qed.

    Lemma nconvert_correct: forall e v,
      eval_pexpr ge m ρC ε e v ->
      forall e', nconvert e = inl e' ->
      v ≠ Vundef →
      exists n, ncompat v n /\
                eval_nexpr ρ e' n /\
                wtype_num n (type_expr e).
    Proof.
      induction 1; simpl.
      - destruct (Hcompat c); subst.
        exists (ρ (elt2p c)); repeat split; subst; auto.
        unfold type_expr, Eval.eval_expr in H1.
        change Eval.TDom.get with TDom.get in H1.
        destruct (TDom.get c tp) as [|[]]; eq_some_inv; subst;
        constructor; auto;
        destruct (ρ (elt2p c));
        try constructor;
        simpl in *; intuition.
      - intros e' E' NB; eq_some_inv. subst. exists (NVint Int.zero).
        repeat (split; auto). constructor. constructor. easy.
      - destruct cst; simpl in *; eq_some_inv; intros e' ?; eq_some_inv; subst;
        (econstructor; split; [auto|split;[repeat econstructor|simpl; auto]]);
        try apply Zle_refl.
        + rewrite <- (Int.repr_signed i) at 3. repeat constructor; omega.
        + destruct (Fappli_IEEE.is_nan _ _ f) eqn:?; repeat constructor; auto.
        + destruct (Fappli_IEEE.is_nan _ _ f) eqn:?; repeat constructor;
          auto using floatofsingle_nan.
        + rewrite <- (Int64.repr_signed i) at 3. repeat constructor; omega.
        + destruct Genv.find_symbol. rewrite Int.repr_unsigned. constructor.
          constructor.
      - intros; eq_some_inv.
      - intros e' He' HB. revert He'.
          assert (Hg:= type_expr_sound _ _ H). intro He'.
          destruct (type_expr a1) as [|[]] eqn:Ea1;
          destruct op; try discriminate;
          (destruct (nconvert a1) as [nconvert_a1|]; eq_some_inv; subst;
           specialize (IHeval_pexpr nconvert_a1 eq_refl));
          (destruct IHeval_pexpr as (n & E1 & E2 & E3); [
             destruct v1, v; simpl in *; congruence
            | destruct n, E3]);
          ncompat_inv E1; subst; simpl in *; eq_some_inv; subst;
          try (apply exist_ncompat_Vundef; repeat econstructor; simpl; eauto);
          try match goal with
            | H : option_map _ ?o = Some ?x |- _ => destruct o eqn:?; inv H
          end;
          repeat econstructor; simpl; eauto.
      - assert (Hg1:= type_expr_sound _ _ H). red in Hg1.
        assert (Hg2:= type_expr_sound _ _ H0). red in Hg2.
        intros e' He' HB.
        set (op':=op); unfold op';
        destruct (type_expr a1) as [|[| | | | |]] eqn:He_a1; simpl in Hg1;
        destruct (type_expr a2) as [|[| | | | |]] eqn:He_a2; simpl in Hg2;
        change (type_expr (PEbinop op a1 a2))
          with (eval_binop op (type_expr a1) (type_expr a2));
        try discriminate;
        destruct op; try discriminate;
        (destruct (nconvert a1) as [nconvert_a1|],
                  (nconvert a2) as [nconvert_a2|]; eq_some_inv; subst;
         specialize (IHeval_pexpr1 nconvert_a1 eq_refl);
         specialize (IHeval_pexpr2 nconvert_a2 eq_refl));
        (destruct IHeval_pexpr1 as (n1 & E1 & E2 & E3); [
           intros ->; apply HB; destruct v2; vm_compute in H1; congruence
          | destruct n1, E3]);
        (destruct IHeval_pexpr2 as (n2 & F1 & F2 & F3); [
           intros ->; apply HB; destruct v1; vm_compute in H1; congruence
          | destruct n2, F3]);
        ncompat_inv E1; ncompat_inv F1; subst; try (elim Hg1; discriminate); try (elim Hg2; discriminate);
        simpl in H1;
        try (unfold Val.cmpl in H1; simpl in H1);
        try (unfold Val.cmplu in H1; simpl in H1);
        eq_some_inv; subst;
        repeat match goal with
           | id: (if ?b then _ else _) = Some _ |- _ => destruct b eqn:?; inv id
           | id: (_ || _) = false |- _ => apply orb_false_elim in id; destruct id
           | id: (_ && _) = false |- _ => apply andb_false_iff in id
           | |- context [if ?b then _ else _] => destruct b eqn:?
         end;
        try match goal with
          | |- context[Oshl] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshr] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshru] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshll] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Oshrl] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Oshrlu] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Val.cmp ?c (Vint ?i1) (Vint ?i2)] =>
            destruct (val_cmp_exists_int c i1 i2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpu ?m ?c (Vint ?i1) (Vint ?i2)] =>
            destruct (val_cmpu_exists_int m c i1 i2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpf ?c (Vfloat ?f1) (Vfloat ?f2)] =>
            destruct (val_cmpf_exists_int c f1 f2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpfs ?c (Vsingle ?f1) (Vsingle ?f2)] =>
            destruct (val_cmpfs_exists_int c f1 f2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.of_bool ?b] =>
            destruct (val_of_bool_exists_int b) as (bi & B1 & B2); rewrite B1
        end;
        try (rewrite He_a1, He_a2);
        try (elim (HB eq_refl));
        try (apply exist_ncompat_Vundef; econstructor; split;
             [ econstructor; try eassumption; econstructor; eauto
             | try match goal with
                 | |- context[of_bool ?b] =>
                   destruct (of_bool_NVint b) as (nb & Hb); rewrite Hb
               end; simpl; auto]; fail);
        try (apply exist_ncompat_Vint;
             split; [econstructor; try eassumption; try rewrite <- B2;
                     econstructor; eauto |
                     simpl; auto]; fail);
        try (apply exist_ncompat_Vptr;
             split; [econstructor; try eassumption;
                     (econstructor || rewrite Int.add_commut; econstructor); eauto |
                     simpl; auto]; fail);
        try (now repeat (econstructor (eauto))).
    Qed.

    Fixpoint nconvert_def_pre (pe: pexpr T.elt) : Prop :=
      match pe with
      | PEvar v => ρC(v) ≠ Vundef
      | PElocal i => ε(i) ≠ None
      | PEconst (Some (Oaddrsymbol i _)) =>
        Genv.find_symbol ge i ≠ None
      | PEconst _ => True
      | PEunop u pe' =>
        nconvert_def_pre pe'
      | PEbinop b pe1 pe2 =>
        nconvert_def_pre pe1 ∧
        nconvert_def_pre pe2 ∧
        match b with
        | Ocmpu c =>
          ∀ b1 i1 b2 i2,
            Vptr b1 i1 ∈ eval_pexpr ge m ρC ε pe1 →
            Vptr b2 i2 ∈ eval_pexpr ge m ρC ε pe2 →
            b1 = b2 ∧ Mem.weak_valid_pointer m b1 (Int.unsigned i1) && Mem.weak_valid_pointer m b2 (Int.unsigned i2) = true
        | _ => True
        end
      end.

    Require Import NonBlock.

    Ltac Split :=
      repeat
      match goal with
      | H : ∃ _, _ |- _ => destruct H as (H & ?)
      | H : _ ∨ _ |- _ => destruct H as [H|H]
      | H : _ ∧ _ |- _ => destruct H as (H & ?)
      end.

    Lemma block_pexpr_unop_inv u q :
      block_pexpr ge m ρC ε (PEunop u q) →
      (∃ v, eval_pexpr ge m ρC ε q v ∧ block_unop u v) ∨ (block_pexpr ge m ρC ε q).
    Proof. inversion 1; subst; vauto. Qed.

    Lemma block_pexpr_binop_inv b q r :
      block_pexpr ge m ρC ε (PEbinop b q r) →
      (∃ v w, eval_pexpr ge m ρC ε q v ∧ eval_pexpr ge m ρC ε r w ∧ block_pbinop m b v w) ∨ (block_pexpr ge m ρC ε q) ∨ (block_pexpr ge m ρC ε r).
    Proof. inversion 1; subst; vauto. Qed.

    Lemma nconvert_def pe :
      nconvert_def_pre pe →
      ∀ ne, nconvert pe = inl ne →
      ¬ block_nexpr ρ ne →
      ¬ block_pexpr ge m ρC ε pe.
    Proof.
      induction pe; intros PRE ne.
    - simpl.
      destruct (Hcompat v) as (? & ?).
      intros NE _ B. inv B. apply PRE; assumption.
    - simpl. intros ?; eq_some_inv; subst ne.
      intros _. simpl in PRE. intros B; inv B. auto.
    - simpl. intros K; destruct o; eq_some_inv. subst ne.
      intros _. intros B; inv B; simpl in PRE; auto.

    - assert (nconvert_def_pre pe) as PRE'.
      { exact PRE. }
      specialize (IHpe PRE'). clear PRE'.
      pose proof (type_expr_sound pe) as TY.
      simpl.
      intros K NB.
      destruct (type_expr pe) as [|()]; eq_some_inv;
      destruct u;
      eq_some_inv;
      apply bind_inl in K;
      destruct K as (ne' & K & ?);
      eq_some_inv;
      subst ne;
      assert (¬ block_nexpr ρ ne') as NB' by (intro; elim NB; vauto);
      specialize (IHpe _ K NB');
      intros B; apply block_pexpr_unop_inv in B; destruct B as [ (? & EV & B) | B ];
      try (apply IHpe; assumption);
      match goal with
      | EV : eval_pexpr _ _ _ _ _ ?i,
        B : block_unop _ _ |- _ =>
        let DEF := fresh in
        assert (DEF := block_pexpr_def IHpe EV);
        specialize (TY _ EV DEF);
        destruct i; destruct TY;
        try (inversion B; intuition eauto; discriminate);
        let CPT := fresh in
        edestruct (nconvert_correct _ _ EV _ K) as (? & CPT & ? & _); eauto;
        ncompat_inv' CPT; subst; inversion B; intuition eauto;
        apply NB; vauto
      end.

    - assert (nconvert_def_pre pe1) as PRE1.
      { apply PRE. }
      specialize (IHpe1 PRE1). clear PRE1.
      assert (nconvert_def_pre pe2) as PRE2.
      { apply PRE. }
      specialize (IHpe2 PRE2). clear PRE2.
      pose proof (type_expr_sound pe1) as TY1.
      pose proof (type_expr_sound pe2) as TY2.
      simpl.
      intros K NB.
      destruct (type_expr pe1) as [|()]; eq_some_inv;
      destruct (type_expr pe2) as [|()]; eq_some_inv;
      destruct b; eq_some_inv;
      apply bind_inl in K; destruct K as (ne1 & NE1 & K);
      apply bind_inl in K; destruct K as (ne2 & NE2 & K);
      eq_some_inv; subst ne;
      assert (¬ block_nexpr ρ ne1) as NB1 by (intro; elim NB; vauto);
      specialize (IHpe1 _ NE1 NB1); clear NB1;
      assert (¬ block_nexpr ρ ne2) as NB2 by (intro; elim NB; vauto);
      specialize (IHpe2 _ NE2 NB2); clear NB2;

      intros B; apply block_pexpr_binop_inv in B; destruct B as [ ( ? & ? & EV1 & EV2 & B) | [ B | B ] ];
      try (apply IHpe1; assumption);
      try (apply IHpe2; assumption);

      match goal with
      | EV1: eval_pexpr _ _ _ _ pe1 ?i1
      , EV2: eval_pexpr _ _ _ _ pe2 ?i2
      , B: block_pbinop _ _ _ _ |- _ =>
        let DEF1 := fresh in
        assert (DEF1 := block_pexpr_def IHpe1 EV1);
        specialize (TY1 _ EV1 DEF1);
        let DEF2 := fresh in
        assert (DEF2 := block_pexpr_def IHpe2 EV2);
        specialize (TY2 _ EV2 DEF2);
        destruct i1; destruct TY1;
        destruct i2; destruct TY2;
        try (inversion B; Split; try discriminate; intuition eauto);
        let T := fresh in
        (destruct (nconvert_correct pe1 _ EV1 _ NE1) as (? & T & ? & _);[ easy | ncompat_inv' T; subst ]);
        (destruct (nconvert_correct pe2 _ EV2 _ NE2) as (? & T & ? & _);[ easy | ncompat_inv' T; subst ]);
        try (apply NB; vauto; fail)
      end;
      simpl in PRE; destruct PRE as (PRE1 & PRE2 & PRE);
      edestruct PRE; eauto.
      InvBooleans. congruence.
    Qed.

  End nconvert.

  Import Utf8.
  Hint Resolve wtype_num_top.

  Lemma compat_monotone var tp tp' :
    leb tp tp' = true →
    ∀ f ρC ρ,
      compat var tp f ρC ρ →
      compat var tp' f ρC ρ.
  Proof.
    intros H f ρC ρ K x. specialize (K x).
    destruct K as [K K']. intuition.
    simpl in *.
    unfold Eval.TDom.get in *.
    edestruct (fun H => TDom.leb_le _ H tp tp') as [X _].
    specialize (X H x).
    unfold flat_top_leb, TDom.get in X. simpl in X.
    destruct (T.get x tp) as [[| | | | |]|];
    destruct (T.get x tp') as [[| | | | |]|];
    auto;
    discriminate.
  Qed.

  Lemma compat_join var tp tp' f ρC ρ :
    compat var tp f ρC ρ ∨ compat var tp' f ρC ρ →
    compat var (tp ⊔ tp') f ρC ρ.
  Proof.
    intros [H|H] c; specialize (H c); destruct H as [H H']; intuition;
    simpl in *;
    unfold Eval.TDom.get, join, join_tree, TDom.join_tree, TDom.joinwiden in *;
    rewrite T.gshcombine; auto; destruct (T.get c tp) as [t|]; auto;
    destruct (T.get c tp') as [t'|]; auto; destruct t; auto; destruct t'; auto.
  Qed.

  Local Instance TeltEqDec : EqDec T.elt := T.elt_eq.
  Lemma compat_upd var `{EQ: EqDec var} tp f ρC ρ :
    (∀ x y, x ≠ y → f x ≠ f y) →
    compat var tp f ρC ρ →
    ∀ x ty v n ρC',
      ncompat v n →
      wtype_num n ty →
      ρC' ∈ upd_spec ρC x v →
      compat var (TDom.set tp x ty) f ρC' (upd ρ (f x) n).
  Proof.
    intros INJ H x ty v n ρC' V Y UPD c. rewrite (UPD c). clear UPD.
    unfold upd. simpl. rewrite Eval.TDom.gsspec. fold TeltEqDec. fold (eq_dec c x).
    destruct (eq_dec c x). subst. rewrite dec_eq_true. split; assumption.
    rewrite dec_eq_false; auto. apply H.
  Qed.

End Nconvert.
