(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Util
  Coqlib AST Values Integers Floats
  Memory
  Cminor Globalenvs
  ToString DebugCminor
  AdomLib Maps ShareTree OnMem.

Set Implicit Arguments.
Local Open Scope string_scope.

Section CST.
Local Unset Elimination Schemes.
Inductive constant :=
| Ointconst : int → constant
| Ofloatconst : float → constant
| Osingleconst : float32 → constant
| Olongconst : int64 → constant
| Oaddrsymbol : ident → int → constant (**r address of a global *)
.
End CST.

Inductive pexpr (var:Type) : Type :=
  | PEvar : var -> pexpr var
  | PElocal: ident → pexpr var (**r address of a local variable *)
  | PEconst : option constant → pexpr var (**r None is /any/ value *)
  | PEunop : unary_operation -> pexpr var -> pexpr var
  | PEbinop : binary_operation -> pexpr var -> pexpr var -> pexpr var.

Fixpoint pe_free_var {var} (pe: pexpr var) : ℘(var) :=
  match pe with
  | PEvar v => eq v
  | PElocal _
  | PEconst _ => ∅
  | PEunop _ pe' => pe_free_var pe'
  | PEbinop _ pe₁ pe₂ => pe_free_var pe₁ ∪ pe_free_var pe₂
  end.

Instance string_of_constant : ToString constant :=
  λ cst,
  match cst with
  | Ointconst i => to_string i
  | Ofloatconst f => to_string f
  | Osingleconst s => to_string s
  | Olongconst l => to_string l
  | Oaddrsymbol s ofs => to_string (s, ofs)
  end.

Fixpoint string_of_pexpr {A} `{ToString A} (e: pexpr A) {struct e} : string :=
  match e with
  | PEvar v => to_string v
  | PElocal i => "&" ++ to_string i
  | PEconst None => "?"
  | PEconst (Some c) => to_string c
  | PEunop op l => string_of_unop op ++ string_of_pexpr l
  | PEbinop op l r => "(" ++ string_of_pexpr l ++ string_of_binop op ++ string_of_pexpr r ++ ")"
  end.

Instance PExprToString A `{ToString A} : ToString (pexpr A) := string_of_pexpr.

Section EVAL.
  Variable var: Type.
  Variables F V: Type.
  Variable ge: Genv.t F V.
  Variable m: Mem.mem.
  Variable ρ: var → val.
  Variable e: ident → option block.

  Definition pexpr_eval_constant (cst: constant) : val :=
    match cst with
    | Ointconst n => Vint n
    | Ofloatconst n => Vfloat n
    | Osingleconst n => Vsingle n
    | Olongconst n => Vlong n
    | Oaddrsymbol s ofs =>
      match Genv.find_symbol ge s with
      | Some b => Vptr b ofs
      | None => Vundef
      end
    end.

  Definition pexpr_eval_binop (op: binary_operation) (x x': val) : option val :=
    match op, x, x' with
    | Ocmpu _, Vptr _ _, Vint _
    | Ocmpu _, Vint _, Vptr _ _ => None
    | Ocmpu c, Vptr b i, Vptr b' i' =>
      if eq_dec b b' && Mem.weak_valid_pointer m b (Int.unsigned i) && Mem.weak_valid_pointer m b' (Int.unsigned i')
      then Some (Val.of_bool (Int.cmpu c i i'))
      else None
    | _, _, _ => eval_binop op x x' m
    end.

  Inductive eval_pexpr: pexpr var -> val -> Prop :=
  | eval_PEvar: forall c v,
    ρ c = v ->
    eval_pexpr (PEvar c) v
  | eval_PElocal: forall i b,
    e(i) = Some b →
    eval_pexpr (PElocal _ i) (Vptr b Int.zero) 
  | eval_PEconst_some: forall cst v,
    pexpr_eval_constant cst = v ->
    eval_pexpr (PEconst _ (Some cst)) v
  | eval_PEconst_none: forall v,
    v ≠ Vundef →
    eval_pexpr (PEconst _ None) v
  | eval_PEunop: forall op a1 v1 v,
    eval_pexpr a1 v1 ->
    eval_unop op v1 = Some v ->
    eval_pexpr (PEunop op a1) v
  | eval_PEbinop: forall op a1 a2 v1 v2 v,
    eval_pexpr a1 v1 ->
    eval_pexpr a2 v2 ->
    pexpr_eval_binop op v1 v2 = Some v →
    eval_pexpr (PEbinop op a1 a2) v.

  Lemma eval_pexpr_var_inv c v :
    eval_pexpr (PEvar c) v →
    ρ c = v.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with PEvar c => ρ c = v | _ => True end).
    change (d (PEvar c) v).
    destruct H; easy.
  Qed.

  Lemma eval_pexpr_local_inv i v :
    eval_pexpr (PElocal _ i) v →
    ∃ b, e(i) = Some b ∧ v = Vptr b Int.zero.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with
                                 | PElocal i => ∃ b, e(i) = Some b ∧ v = Vptr b Int.zero
                  | _ => True
                  end).
    change (d (PElocal _ i) v).
    destruct H; simpl; eauto.
  Qed.

  Lemma eval_pexpr_const_Some_inv cst v :
    eval_pexpr (PEconst _ (Some cst)) v →
    v = pexpr_eval_constant cst.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with
                  | PEconst (Some cst') => v = pexpr_eval_constant cst'
                  | _ => True
                  end).
    change (d (PEconst _ (Some cst)) v).
    destruct H; easy.
  Qed.

  Lemma eval_pexpr_const_None_inv v :
    eval_pexpr (PEconst _ None) v →
    v ≠ Vundef.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with
                  | PEconst None => v ≠ Vundef
                  | _ => True
                  end).
    change (d (PEconst _ None) v).
    destruct H; easy.
  Qed.

  Lemma eval_pexpr_unop_inv op u v :
    eval_pexpr (PEunop op u) v →
    ∃ v', eval_pexpr u v' ∧ eval_unop op v' = Some v.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with
                                 | PEunop op u => ∃ v', eval_pexpr u v' ∧ eval_unop op v' = Some v
                                 | _ => True
                               end).
    change (d (PEunop op u) v).
    destruct H; simpl; eauto.
  Qed.

  Lemma eval_pexpr_unop_ptr op u b i :
    ¬ (eval_pexpr (PEunop op u) (Vptr b i)).
  Proof.
    remember (PEunop op u) as pe eqn: PE.
    remember (Vptr b i) as v.
    induction 1; try discriminate.
    inv PE.
    destruct op; simpl in *; eq_some_inv; destruct v1; simpl in *; eq_some_inv; try discriminate;
    match goal with
      | H : appcontext[Float.to_int ?f] |- _ => destruct (Float.to_int f)
      | H : appcontext[Float.to_intu ?f] |- _ => destruct (Float.to_intu f)
      | H : appcontext[Float.to_long ?f] |- _ => destruct (Float.to_long f)
      | H : appcontext[Float.to_longu ?f] |- _ => destruct (Float.to_longu f)
      | H : appcontext[Float32.to_int ?f] |- _ => destruct (Float32.to_int f)
      | H : appcontext[Float32.to_intu ?f] |- _ => destruct (Float32.to_intu f)
      | H : appcontext[Float32.to_long ?f] |- _ => destruct (Float32.to_long f)
      | H : appcontext[Float32.to_longu ?f] |- _ => destruct (Float32.to_longu f)
    end; simpl in *; eq_some_inv; discriminate.
  Qed.

  Lemma eval_pexpr_binop_inv op u₁ u₂ v :
    eval_pexpr (PEbinop op u₁ u₂) v →
    ∃ v₁ v₂, eval_pexpr u₁ v₁ ∧ eval_pexpr u₂ v₂ ∧ pexpr_eval_binop op v₁ v₂ = Some v.
  Proof.
    intros H.
    set (d (k: pexpr var) v := match k with
                                 | PEbinop op u₁ u₂ =>
                                           ∃ v₁ v₂, eval_pexpr u₁ v₁ ∧ eval_pexpr u₂ v₂ ∧ pexpr_eval_binop op v₁ v₂ = Some v
                                 | _ => True
                               end).
    change (d (PEbinop op u₁ u₂) v).
    destruct H; try exact I.
    repeat eexists; eauto.
  Qed.

End EVAL.

Ltac eval_pexpr_inv :=
  repeat match goal with
  | H : eval_pexpr _ _ _ _ (PEvar _) _ |- _ => apply eval_pexpr_var_inv in H
  | H : eval_pexpr _ _ _ _ (PElocal _ _) _ |- _ => apply eval_pexpr_local_inv in H; hsplit
  | H : eval_pexpr _ _ _ _ (PEconst _ None) _ |- _ => apply eval_pexpr_const_None_inv in H
  | H : eval_pexpr _ _ _ _ (PEconst _ (Some _)) _ |- _ => apply eval_pexpr_const_Some_inv in H
  | H : eval_pexpr _ _ _ _ (PEunop _ _) _ |- _ => apply eval_pexpr_unop_inv in H; hsplit
  | H : eval_pexpr _ _ _ _ (PEbinop _ _ _) _ |- _ => apply eval_pexpr_binop_inv in H; hsplit
  end.

Module Cell2Val.
Unset Elimination Schemes.
Class adom (t:Type) : Type := {
  c2v_leb :> leb_op t;
  c2v_join :> join_op t (t+⊤);
  c2v_widen :> widen_op t (t+⊤);
  eval_constant: constant → t+⊤;
  eval_local: ident -> t+⊤;
  eval_unop : unary_operation -> t+⊤ -> t+⊤;
  eval_binop : binary_operation -> t+⊤ -> t+⊤ -> t+⊤;
  eval_cond : t+⊤ -> t+⊤+⊥ -> t+⊤+⊥ -> t+⊤+⊥;

  c2v_gamma :> gamma_op t val;
  c2v_leb_correct :> leb_op_correct t val;
  c2v_join_correct :> join_op_correct t (t+⊤) val;
  eval_constant_correct: forall F V (ge:Genv.t F V) c v,
    pexpr_eval_constant ge c = v ->
    v ≠ Vundef →
    γ (eval_constant c) v;
  eval_local_correct: forall i b,
    (Vptr b Int.zero) ∈ γ(eval_local i);
  eval_unop_correct : forall op arg x v,
    Cminor.eval_unop op arg = Some v ->
    v ≠ Vundef →
    γ x arg ->
    γ (eval_unop op x) v;
  eval_binop_correct : forall op arg1 arg2 x1 x2 m v,
    pexpr_eval_binop m op arg1 arg2 = Some v ->
    v ≠ Vundef →
    γ x1 arg1 ->
    γ x2 arg2 ->
    γ (eval_binop op x1 x2) v;
  eval_cond_correct b l r B bv :
    γ b B ->
    Val.bool_of_val B bv ->
    forall K,
      γ (if bv then l else r) K ->
      γ (eval_cond b l r) K
}.

Module Eval (T:SHARETREE).

  Module TDom := WPFun T.

  Section val_adom.
    Context {t:Type} (a:adom t).
    Variable a_join_eq: ∀ x:t, join x x = Just x.
    Variable a_widen_eq: ∀ x, widen x x = Just x.
    Variable a_leb_eq: ∀ x, leb x x = true.

    Fixpoint eval_expr (ab: TDom.t t) (e:pexpr T.elt) : t+⊤ :=
      match e with
        | PEvar c => TDom.get c ab
        | PElocal i => eval_local i
        | PEconst (Some cst) => eval_constant cst
        | PEconst None => All
        | PEunop op e => eval_unop op (eval_expr ab e)
        | PEbinop op e1 e2 => eval_binop op (eval_expr ab e1) (eval_expr ab e2)
      end.

    Instance gamma : gamma_op (TDom.t t) (T.elt -> val) := γ.
    Instance join : join_op (TDom.t t) (TDom.t t) := TDom.join_tree _ a_join_eq.
    Instance widen : widen_op (TDom.t t) (TDom.t t) := TDom.widen_tree _ a_widen_eq.
    Instance leb : leb_op (TDom.t t) := TDom.leb_tree _ a_leb_eq.

    Lemma pexpr_eval_binop_vundef m op v v' v'' :
      v'' ≠ Vundef →
      pexpr_eval_binop m op v v' = Some v'' →
      v ≠ Vundef ∧ v' ≠ Vundef.
    Proof.
      Opaque Cminor.eval_binop.
      intros H.
      destruct op; simpl; intros K;
      split; intros ->; apply H;
      eauto using eval_binop_strict_l, eval_binop_strict_r.
      Transparent Cminor.eval_binop.
      simpl in *.
      destruct v; eq_some_inv; subst; reflexivity.
    Qed.

    Lemma eval_expr_correct {F V} : forall (ge: Genv.t F V) (m: Mem.mem) ε (ab: TDom.t t) ρ e v,
      γ ab ρ ->
      v ∈ eval_pexpr ge m ρ ε e →
      v ≠ Vundef →
      γ (eval_expr ab e) v.
    Proof.
      induction 2; simpl; intros K.
      subst. apply H.
      eapply eval_local_correct.
      eapply eval_constant_correct; eassumption.
      easy.
      eapply eval_unop_correct; eauto using eval_unop_vundef.
      destruct (pexpr_eval_binop_vundef _ _ _ _ K H0).
      eapply eval_binop_correct; eauto.
    Qed.
  End val_adom.
End Eval.

End Cell2Val.
