(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib Integers ArithLib AdomLib ToString.

Require Psatz.

Require Import
  ZIntervals
  AbIdealNonrel
  IdealExpr
  ZCongruences_defs
  AbIdealEnv IdealBoxDomain
  Util.

Local Coercion Z_of_N : N >-> Z.

Definition extern_top (ab:zcongr) : zcongr+⊤ :=
  match N.eq_dec 1%N ab.(zc_mod) with
    | left _ => All
    | right _ => Just ab
  end.

Definition intern_top (ab:zcongr+⊤) : zcongr :=
  match ab with
    | All => top
    | Just ab => ab
  end.

(** * Abstract Operators. *)
Program Definition zc_const (z: Z) : zcongr := ZC z 0 _.

Definition zc_true : zcongr := zc_const 1.
Definition zc_false: zcongr := zc_const 0.
Definition zc_bool (b: bool) : zcongr := if b then zc_true else zc_false.

Definition zc_itv (lo hi: Z) : zcongr :=
  if Z.eqb lo hi
  then zc_const lo
  else top.

Program Definition zc_add (plus: Z → Z → Z) (x y: zcongr) : zcongr :=
  let m := N.gcd x.(zc_mod) y.(zc_mod) in
  let r := plus x.(zc_rem) y.(zc_rem) in
  ZC (if m then r else r mod m) m _.
Next Obligation. destruct N.gcd. discriminate. apply Z_mod_lt. auto. Qed.

Program Definition zc_mul (x y: zcongr) : zcongr :=
  let g1 := Z.gcd x.(zc_rem) x.(zc_mod) in
  let g2 := Z.gcd (g1 * y.(zc_mod)) (y.(zc_rem) * x.(zc_mod)) in
  let m := Z.abs_N g2 in
  let r := x.(zc_rem) * y.(zc_rem) in
  ZC (if m then r else r mod m) m _.
Next Obligation. destruct Z.abs_N. discriminate. apply Z_mod_lt. auto. Qed.

Program Definition zc_quot_pos (x: zcongr) (y: positive) : zcongr :=
  match Zdivides_dec (Zpos y) x.(zc_rem), Ndivides_dec (Npos y) x.(zc_mod) return _ with
    | left DIV_rem, left DIV_mod =>
      {| zc_rem := x.(zc_rem) / Zpos y ; zc_mod := x.(zc_mod) / Npos y |}
    | _, _ => ⊤
  end.
Next Obligation.
  pose proof (zc_rem_itv x). destruct (zc_mod x). rewrite N.div_0_l in H; discriminate.
  specialize (H0 eq_refl). destruct H0.
  split.
  - apply Z_div_le with (c:=Z.pos y) in H0. rewrite Z.div_0_l in H0.
    auto. discriminate. reflexivity.
  - destruct DIV_mod, DIV_rem. rewrite N2Z.inj_gt, N2Z.inj_div, H2, H3 in *. clear H2 p H3 x.
    simpl in *. rewrite !Z.div_mul in * by discriminate.
    eapply Zmult_lt_reg_r; eauto. reflexivity.
Qed.

Program Definition zc_div_pos (x: zcongr) (y: positive) : zcongr :=
  match Ndivides_dec (Npos y) x.(zc_mod) return _ with
    | left DIV => {| zc_rem := x.(zc_rem) / Zpos y; zc_mod := x.(zc_mod) / Npos y |}
    | _ => top
  end.
Next Obligation.
  pose proof (zc_rem_itv x). destruct (zc_mod x). rewrite N.div_0_l in H; discriminate.
  specialize (H0 eq_refl). destruct H0.
  split.
  - apply Z_div_le with (c:=Z.pos y) in H0. rewrite Z.div_0_l in H0.
    auto. discriminate. reflexivity.
  - destruct DIV. rewrite N2Z.inj_gt, N2Z.inj_div, H2 in *. clear H2 p.
    simpl in *. rewrite !Z.div_mul in * by discriminate.
    eapply Zmult_lt_reg_r with (p:=Z.pos y). reflexivity.
    pose proof (Zmod_eq (zc_rem x) (Z.pos y) eq_refl).
    pose proof (Z_mod_lt (zc_rem x) (Z.pos y) eq_refl).
    omega.
Qed.

Definition zc_quot (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | Zpos y' => zc_quot_pos x y'
       | _ => top
       end
  else top.

Definition zc_shl (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | 0 => x
       | Zpos y' => zc_mul x (zc_const (2 ^ Zpos y'))
       | Zneg y' => zc_div_pos x (2 ^ y')
       end
  else top.

Definition zc_shr (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | 0 => x
       | Zpos y' => zc_div_pos x (2 ^ y')
       | Zneg y' => zc_mul x (zc_const (2 ^ Zpos y'))
       end
  else top.

Program Definition zc_remainder (x y: zcongr) : zcongr :=
  if N.eqb y.(zc_mod) 0 && Zdivides_dec y.(zc_rem) x.(zc_rem) &&
                           Zdivides_dec y.(zc_rem) x.(zc_mod)
  then zc_const 0
  else
    let m := Z.abs_N (Z.gcd x.(zc_mod) (Z.gcd y.(zc_mod) y.(zc_rem))) in
    ZC (if m then x.(zc_rem) else x.(zc_rem) mod m) m _.
Next Obligation. destruct Z.abs_N. discriminate. apply Z_mod_lt. auto. Qed.

Definition zc_cmp_eq (b: bool) (x y: zcongr) : zcongr :=
  match x.(zc_mod), y.(zc_mod) with
  | N0, N0 => zc_bool (eqb b (Z.eqb x.(zc_rem) y.(zc_rem)))
  | N0, Npos m
  | Npos m, N0 => if congr_dec (Zpos m) x.(zc_rem) y.(zc_rem)
                  then top
                  else zc_bool (negb b)
  | Npos x', Npos y' =>
    (* TODO *) top
  end.

Definition zc_cmp (cmp: comparison) (x y: zcongr) : zcongr :=
  match cmp with
  | Ceq => zc_cmp_eq true x y
  | Cne => zc_cmp_eq false x y
  | _ => (* TODO *)
    top
  end.

Program Definition zc_neg x :=
  Just (ZC (if x.(zc_mod)
            then - x.(zc_rem)
            else (- x.(zc_rem)) mod x.(zc_mod))
           x.(zc_mod) _).
Next Obligation. destruct (zc_mod x). discriminate. apply Z_mod_lt. auto. Qed.

Definition zc_forward_unop op (x:zcongr+⊤) : zcongr+⊤ :=
  match op with
    | IOneg => match x with All => All | Just x => zc_neg x end
    | IOnot => extern_top (zc_add Z.sub (zc_const (-1)) (intern_top x))
    | IOnegf
    | IOabsf
    | IOsingleoffloat
    | IOzoffloat
    | IOfloatofz
    | IOsingleofz
      => All
  end.

Definition zc_backward_unop op (ab:zcongr+⊤) : zcongr+⊤ :=
  match op with
    | IOneg => match ab with All => All | Just ab => zc_neg ab end
    | IOnot => extern_top (zc_add Z.sub (zc_const 1) (intern_top ab))
    | _ => top
  end.

Lemma lor_pos_neg:
  ∀ a, ∃ (k:N), Z.lor (Z.pos a) (Z.neg a) = -2^k.
Proof.
  simpl. destruct a; simpl.
  - replace (Pos.ldiff a a) with 0%N by (symmetry; apply (N.ldiff_diag (N.pos a))).
    exists 0%N. auto.
  - induction a; simpl.
    + simpl. replace (Pos.ldiff a a) with 0%N by (symmetry; apply (N.ldiff_diag (N.pos a))).
      exists 1%N. auto.
    + assert (∀ x, Z.neg (N.succ_pos (Pos.Nsucc_double x)) = -2*x-2)%Z.
      { intros []. auto. simpl. intro. rewrite Pos.add_1_r. auto. }
      rewrite H.
      destruct IHa as [k Hk]. exists (k+1)%N.
      rewrite N2Z.inj_add, Z.pow_add_r, Zopp_mult_distr_l, <- Hk by (zify; omega).
      destruct Pos.ldiff. auto. simpl. zify; omega.
    + exists 1%N. auto.
  - exists 0%N. auto.
Qed.

Lemma lor_opp:
  ∀ a, a = 0 \/ ∃ (k:N), Z.lor a (-a) = -2^k.
Proof.
  intros. destruct a. auto.
  right. apply lor_pos_neg.
  right. rewrite Z.lor_comm. apply lor_pos_neg.
Qed.

Program Definition zc_forward_and x y :=
  let maskx := Z.lor x.(zc_mod) (-x.(zc_mod)) in
  let masky := Z.lor y.(zc_mod) (-y.(zc_mod)) in
  let unknownx := Z.lor maskx x.(zc_rem) in
  let unknowny := Z.lor masky y.(zc_rem) in
  let unknowndirect := Z.lor maskx masky in
  let unknown := Z.land (Z.land unknownx unknowny) unknowndirect in
  let modu := -Z.lor unknown (-unknown) in
  let rem := Z.land (Z.land x.(zc_rem) y.(zc_rem)) (Z.pred modu) in
  ZC rem (Z.to_N modu) _.
Next Obligation.
  match goal with
    | |- context [Z.lor (?unknown0) (-?unknown0)] =>
      set (unknown:=unknown0) in *
  end.
  destruct (lor_opp unknown) as [|[k Hk]]. rewrite H0 in H. discriminate.
  rewrite Hk, Z.opp_involutive.
  set (rem := Z.land (Z.land (zc_rem x) (zc_rem y)) (Z.pred (2 ^ k))).
  assert (0 <= rem).
  { apply Z.land_nonneg. right. apply Zlt_0_le_0_pred, Z.pow_pos_nonneg; zify; omega. }
  split. auto.
  rewrite Z2N.id by (apply Z.pow_nonneg; omega).
  destruct (Z_lt_dec 0 rem). 2:replace rem with 0 by omega; apply Z.pow_pos_nonneg; zify; omega.
  apply Z.log2_lt_cancel. rewrite Z.log2_pow2 by (zify; omega).
  apply Z.bit_log2 in l.
  unfold rem at 1 in l. replace (Z.pred (2^k)) with (Z.lnot (-1*2^k)) in l by (unfold Z.lnot; omega).
  apply Z.nle_gt. intro.
  rewrite Z.land_spec, Bool.andb_true_iff, Z.lnot_spec, <- Z.shiftl_mul_pow2, Z.shiftl_spec,
          Int.Ztestbit_m1 in l
    by (zify; omega || apply Z.log2_nonneg).
  easy.
Qed.

Program Definition zc_forward_or x y :=
  let maskx := Z.lor x.(zc_mod) (-x.(zc_mod)) in
  let masky := Z.lor y.(zc_mod) (-y.(zc_mod)) in
  let unknownx := Z.lor maskx (Z.lnot x.(zc_rem)) in
  let unknowny := Z.lor masky (Z.lnot y.(zc_rem)) in
  let unknowndirect := Z.lor maskx masky in
  let unknown := Z.land (Z.land unknownx unknowny) unknowndirect in
  let modu := -Z.lor unknown (-unknown) in
  let rem := Z.land (Z.lor x.(zc_rem) y.(zc_rem)) (Z.pred modu) in
  ZC rem (Z.to_N modu) _.
Next Obligation.
  match goal with
    | |- context [Z.lor (?unknown0) (-?unknown0)] =>
      set (unknown:=unknown0) in *
  end.
  destruct (lor_opp unknown) as [|[k Hk]]. rewrite H0 in H. discriminate.
  rewrite Hk, Z.opp_involutive.
  set (rem := Z.land (Z.lor (zc_rem x) (zc_rem y)) (Z.pred (2 ^ k))).
  assert (0 <= rem).
  { apply Z.land_nonneg. right. apply Zlt_0_le_0_pred, Z.pow_pos_nonneg; zify; omega. }
  split. auto.
  rewrite Z2N.id by (apply Z.pow_nonneg; omega).
  destruct (Z_lt_dec 0 rem). 2:replace rem with 0 by omega; apply Z.pow_pos_nonneg; zify; omega.
  apply Z.log2_lt_cancel. rewrite Z.log2_pow2 by (zify; omega).
  apply Z.bit_log2 in l.
  unfold rem at 1 in l. replace (Z.pred (2^k)) with (Z.lnot (-1*2^k)) in l by (unfold Z.lnot; omega).
  apply Z.nle_gt. intro.
  rewrite Z.land_spec, Bool.andb_true_iff, Z.lnot_spec, <- Z.shiftl_mul_pow2, Z.shiftl_spec,
          Int.Ztestbit_m1 in l
    by (zify; omega || apply Z.log2_nonneg).
  easy.
Qed.

Program Definition zc_forward_xor x y :=
  let maskx := Z.lor x.(zc_mod) (-x.(zc_mod)) in
  let masky := Z.lor y.(zc_mod) (-y.(zc_mod)) in
  let unknown := Z.lor maskx masky in
  let modu := -Z.lor unknown (-unknown) in
  let rem := Z.land (Z.lxor x.(zc_rem) y.(zc_rem)) (Z.pred modu) in
  ZC rem (Z.to_N modu) _.
Next Obligation.
  match goal with
    | |- context [Z.lor (?unknown0) (-?unknown0)] =>
      set (unknown:=unknown0) in *
  end.
  destruct (lor_opp unknown) as [|[k Hk]]. rewrite H0 in H. discriminate.
  rewrite Hk, Z.opp_involutive.
  set (rem := Z.land (Z.lxor (zc_rem x) (zc_rem y)) (Z.pred (2 ^ k))).
  assert (0 <= rem).
  { apply Z.land_nonneg. right. apply Zlt_0_le_0_pred, Z.pow_pos_nonneg; zify; omega. }
  split. auto.
  rewrite Z2N.id by (apply Z.pow_nonneg; omega).
  destruct (Z_lt_dec 0 rem). 2:replace rem with 0 by omega; apply Z.pow_pos_nonneg; zify; omega.
  apply Z.log2_lt_cancel. rewrite Z.log2_pow2 by (zify; omega).
  apply Z.bit_log2 in l.
  unfold rem at 1 in l. replace (Z.pred (2^k)) with (Z.lnot (-1*2^k)) in l by (unfold Z.lnot; omega).
  apply Z.nle_gt. intro.
  rewrite Z.land_spec, Bool.andb_true_iff, Z.lnot_spec, <- Z.shiftl_mul_pow2, Z.shiftl_spec,
          Int.Ztestbit_m1 in l
    by (zify; omega || apply Z.log2_nonneg).
  easy.
Qed.

Definition zc_forward_binop op (x y:zcongr+⊤) : zcongr+⊤ :=
  match op with
    | IOadd | IOsub | IOmul | IOdiv | IOshl | IOshr | IOmod | IOcmp _
    | IOand | IOor | IOxor =>
      let x := intern_top x in
      let y := intern_top y in
      match op with
        | IOadd => extern_top (zc_add Z.add x y)
        | IOsub => extern_top (zc_add Z.sub x y)
        | IOmul => extern_top (zc_mul x y)
        | IOdiv => extern_top (zc_quot x y)
        | IOshl => extern_top (zc_shl x y)
        | IOshr => extern_top (zc_shr x y)
        | IOmod => extern_top (zc_remainder x y)
        | IOcmp cmp => extern_top (zc_cmp cmp x y)
        | IOand => extern_top (zc_forward_and x y)
        | IOor => extern_top (zc_forward_or x y)
        | IOxor => extern_top (zc_forward_xor x y)
        | _ => top
      end
    | IOcmpf _ | IOaddf | IOsubf | IOmulf | IOdivf => top
  end.

Definition zc_bw_cmp (cmp: comparison) (res: zcongr) (x y: zcongr) : (zcongr+⊤ * zcongr+⊤)+⊥ :=
  if res.(zc_mod)
  then let cmp := if Z_zerop res.(zc_rem) then negate_comparison cmp else cmp in
       match cmp with
       | Ceq =>
         do_bot z <- x ⊓ y; NotBot (extern_top z, extern_top z)
       | _ => NotBot (extern_top x, extern_top y)
       end
  else NotBot (extern_top x, extern_top y).

Definition zc_backward_add res l r :=
  NotBot (extern_top (zc_add Z.sub res r), extern_top (zc_add Z.sub res l)).

Definition zc_backward_sub res l r :=
  NotBot (extern_top (zc_add Z.add res r), extern_top (zc_add Z.sub l res)).

(* TODO : improve *)
Program Definition zc_backward_mod res (l r:zcongr) :=
  let m := Z.abs_N (Z.gcd res.(zc_mod) (Z.gcd r.(zc_mod) r.(zc_rem))) in
  NotBot (
      extern_top (ZC (if m then res.(zc_rem) else res.(zc_rem) mod m) m _),
      ⊤:zcongr+⊤).
Next Obligation. destruct Z.abs_N. discriminate. apply Z_mod_lt. auto. Qed.

Program Definition zc_backward_and_l res x y :=
  let maskres := Z.lor res.(zc_mod) (-res.(zc_mod)) in
  let maskx := Z.lor x.(zc_mod) (-x.(zc_mod)) in
  let masky := Z.lor y.(zc_mod) (-y.(zc_mod)) in
  let unknown0res :=
      Z.lor (Z.lor (res.(zc_rem)) (Z.lnot y.(zc_rem)))
            (Z.lor masky maskres)
  in
  let unknown0 := Z.land unknown0res (Z.lor maskx x.(zc_rem)) in
  let unknown1res := Z.lor (Z.lnot res.(zc_rem)) maskres in
  let unknown1 := Z.land unknown1res (Z.lor maskx (Z.lnot x.(zc_rem))) in
  if Z.eqb (Z.lor unknown0 unknown1) (-1) then
    let unknown := Z.land unknown0 unknown1 in
    let modu := -Z.lor unknown (-unknown) in
    let rem := Z.land (Z.ldiff unknown0 unknown1) (Z.pred modu) in
    NotBot (ZC rem (Z.to_N modu) _)
  else
    Bot.
Next Obligation.
  repeat match goal with
    | |- context [Z.lor (?x) (-?x)] =>
      try (unfold x; fail 1);
      let u := fresh "z" in
      set (u:=x) in *
  end.
  destruct (lor_opp z2) as [|[k Hk]]. rewrite H0 in H. discriminate.
  rewrite Hk, Z.opp_involutive.
  match goal with
    | |- context [0 <= ?rem'] => set (rem := rem')
  end.
  assert (0 <= rem).
  { apply Z.land_nonneg. right. apply Zlt_0_le_0_pred, Z.pow_pos_nonneg; zify; omega. }
  split. auto.
  rewrite Z2N.id by (apply Z.pow_nonneg; omega).
  destruct (Z_lt_dec 0 rem). 2:replace rem with 0 by omega; apply Z.pow_pos_nonneg; zify; omega.
  apply Z.log2_lt_cancel. rewrite Z.log2_pow2 by (zify; omega).
  apply Z.bit_log2 in l.
  unfold rem at 1 in l. replace (Z.pred (2^k)) with (Z.lnot (-1*2^k)) in l by (unfold Z.lnot; omega).
  apply Z.nle_gt. intro.
  rewrite Z.land_spec, Bool.andb_true_iff, Z.lnot_spec, <- Z.shiftl_mul_pow2, Z.shiftl_spec,
          Int.Ztestbit_m1 in l
    by (zify; omega || apply Z.log2_nonneg).
  easy.
Qed.

Program Definition zc_backward_or_l res x y :=
  let maskres := Z.lor res.(zc_mod) (-res.(zc_mod)) in
  let maskx := Z.lor x.(zc_mod) (-x.(zc_mod)) in
  let masky := Z.lor y.(zc_mod) (-y.(zc_mod)) in
  let unknown1res :=
      Z.lor (Z.lor (Z.lnot res.(zc_rem)) y.(zc_rem))
            (Z.lor masky maskres)
  in
  let unknown1 := Z.land unknown1res (Z.lor maskx (Z.lnot x.(zc_rem))) in
  let unknown0res := Z.lor res.(zc_rem) maskres in
  let unknown0 := Z.land unknown0res (Z.lor maskx x.(zc_rem)) in
  if Z.eqb (Z.lor unknown0 unknown1) (-1) then
    let unknown := Z.land unknown0 unknown1 in
    let modu := -Z.lor unknown (-unknown) in
    let rem := Z.land (Z.ldiff unknown0 unknown1) (Z.pred modu) in
    NotBot (ZC rem (Z.to_N modu) _)
  else
    Bot.
Next Obligation.
  repeat match goal with
    | |- context [Z.lor (?x) (-?x)] =>
      try (unfold x; fail 1);
      let u := fresh "z" in
      set (u:=x) in *
  end.
  destruct (lor_opp z2) as [|[k Hk]]. rewrite H0 in H. discriminate.
  rewrite Hk, Z.opp_involutive.
  match goal with
    | |- context [0 <= ?rem'] => set (rem := rem')
  end.
  assert (0 <= rem).
  { apply Z.land_nonneg. right. apply Zlt_0_le_0_pred, Z.pow_pos_nonneg; zify; omega. }
  split. auto.
  rewrite Z2N.id by (apply Z.pow_nonneg; omega).
  destruct (Z_lt_dec 0 rem). 2:replace rem with 0 by omega; apply Z.pow_pos_nonneg; zify; omega.
  apply Z.log2_lt_cancel. rewrite Z.log2_pow2 by (zify; omega).
  apply Z.bit_log2 in l.
  unfold rem at 1 in l. replace (Z.pred (2^k)) with (Z.lnot (-1*2^k)) in l by (unfold Z.lnot; omega).
  apply Z.nle_gt. intro.
  rewrite Z.land_spec, Bool.andb_true_iff, Z.lnot_spec, <- Z.shiftl_mul_pow2, Z.shiftl_spec,
          Int.Ztestbit_m1 in l
    by (zify; omega || apply Z.log2_nonneg).
  easy.
Qed.

Definition zc_backward_binop op res l r :=
  match op with
    | IOadd => zc_backward_add res (intern_top l) (intern_top r)
    | IOsub => zc_backward_sub res (intern_top l) (intern_top r)
    | IOmod => zc_backward_mod res (intern_top l) (intern_top r)
    | IOand =>
      let l := intern_top l in let r := intern_top r in
      do_bot l <- zc_backward_and_l res l r;
      do_bot r <- zc_backward_and_l res r l;
      NotBot (extern_top l, extern_top r)
    | IOor =>
      let l := intern_top l in let r := intern_top r in
      do_bot l <- zc_backward_or_l res l r;
      do_bot r <- zc_backward_or_l res r l;
      NotBot (extern_top l, extern_top r)
    | IOxor =>
      let l := intern_top l in let r := intern_top r in
      NotBot (extern_top (zc_forward_xor res r),
              extern_top (zc_forward_xor res l))
    | IOcmp cmp =>
      zc_bw_cmp cmp res (intern_top l) (intern_top r)
    | _ => NotBot (top, top) (* TODO *)
  end.

(** * Soundness. *)

Lemma zc_bool_sound b : (if b then 1 else 0) ∈ γ(zc_bool b).
Proof. destruct b; exists 0; reflexivity. Qed.

Lemma zc_add_sound (plus: Z → Z → Z) (x y: zcongr) (i j: Z) :
  (∀ n u u' v v', congr n u u' → congr n v v' → congr n (plus u v) (plus u' v')) →
  γ x i → γ y j → γ (zc_add plus x y) (plus i j).
Proof.
  intros COMPAT.
  unfold zc_add.
  destruct x as [xr xm ?]. destruct y as [yr ym ?].
  simpl. intros [p Hp] [q Hq]. simpl in Hp, Hq. unfold zc_gamma. simpl. rewrite Hp, Hq.
  case_eq (N.gcd xm ym). rewrite N.gcd_eq_0. intros [-> ->].
  repeat rewrite Z.mul_0_r, Z.add_0_r. auto with congr.
  intros gcd G.
  eapply congr_trans. 2: apply congr_mod_compat. 2: reflexivity.
  rewrite <- G. clear G gcd.
  rewrite Ngcd_is_Zgcd.
  apply COMPAT.
  rewrite <- (Z.add_0_r i) at 1. apply congr_plus_compat.
  auto with congr. eapply congr_divide. 2: apply Zgcd_divides_l.
  exists p; ring.
  rewrite <- (Z.add_0_r j) at 1. apply congr_plus_compat.
  auto with congr. eapply congr_divide. 2: apply Zgcd_divides_r.
  exists q; ring.
Qed.

Lemma zc_mul_sound x y i j :
  γ x i → γ y j → γ (zc_mul x y) (i * j).
Proof.
  unfold zc_mul.
  destruct x as [xr xm], y as [yr ym]. simpl. intros [qx Hx] [qy Hy]. simpl in *.
  set (m := Z.abs_N (Z.gcd (Z.gcd xr xm * ym) (yr * xm))).
  assert (m | (xr*yr - i*j)).
  { subst m xr yr. rewrite Zabs2N.id_abs. apply Zdivide_Zabs_inv_l.
    replace ((i + qx * xm) * (j + qy * ym) - i * j)
    with (((j+qy*ym)*xm)*qx + ((i+qx*xm)*qy-xm*qx*qy)*ym) by ring.
    apply Z.divide_add_r. apply Z.divide_mul_l, Zgcd_divides_r.
    eapply Zdivides_trans. apply Zgcd_divides_l.
    auto using Z.mul_divide_mono_r, Z.divide_sub_r, Z.divide_mul_l, Zgcd_divides_l, Zgcd_divides_r.
  }
  unfold zc_gamma. simpl. destruct m. apply Z.divide_0_l in H.
  exists 0. simpl. rewrite <- H. ring.
  eapply congr_trans, congr_mod_compat. 2:reflexivity.
  destruct H as [q H]. exists q. change (xr * yr = i * j + q * N.pos p). omega.
Qed.

Lemma zc_quot_pos_sound x y i :
  γ x i → γ (zc_quot_pos x y) (i ÷ Zpos y).
Proof.
  unfold zc_quot_pos.
  destruct x as [xr xm]. intros [q H]. simpl in H. subst.
  simpl. destruct Zdivides_dec as [[x Hx]|].
  destruct Ndivides_dec as [[d H]|].
  simpl.
  exists q. simpl.
  assert (Z.pos y ≠ 0) as YNZ by discriminate.
  rewrite Hx.
  rewrite N2Z.inj_div. rewrite H in *. simpl in *.
  rewrite !Z_div_mult_full; auto.
  assert (i = (x - q * d) * Z.pos y). Psatz.lia.
  subst. rewrite Z.quot_mul; auto. ring.
  apply top_correct. apply top_correct.
Qed.

Lemma zc_div_pos_sound x y i :
  γ x i → γ (zc_div_pos x y) (i / Zpos y).
Proof.
  unfold zc_div_pos.
  destruct x as [xr xm]. intros [q H]. simpl in *. subst.
  destruct Ndivides_dec as [[d H]|].
  exists q. simpl.
  rewrite N2Z.inj_div, H. simpl.
  assert (Z.pos y ≠ 0) as YNZ by discriminate.
  rewrite Z_div_mult_full; auto.
  replace (q * (d * Zpos y)) with ((q * d) * Zpos y) by ring.
  rewrite Z.div_add; auto.
  apply top_correct.
Qed.

Lemma zc_quot_sound x y i j :
  γ x i → γ y j → γ (zc_quot x y) (i ÷ j).
Proof.
  unfold zc_quot.
  destruct y as [yr ym].
  simpl. destruct ym.
  destruct yr as [|y'|]. intros; apply top_correct.
  intros K H. apply congr_0_eq in H. subst. simpl.
  apply zc_quot_pos_sound; auto.
  intros; apply top_correct. intros; apply top_correct.
Qed.

Lemma zc_shl_sound x y i j :
  γ x i → γ y j → γ (zc_shl x y) (Z.shiftl i j).
Proof.
  unfold zc_shl.
  destruct y as [yr ym]. destruct ym; auto; simpl.
  intros K H. apply congr_0_eq in H. subst. simpl.
  destruct yr as [|y'|y'].
  exact K.
  rewrite Z.shiftl_mul_pow2. apply zc_mul_sound; auto. exists 0. simpl. ring.
  discriminate.
  rewrite Z.shiftl_div_pow2.
  change (- Z.neg y') with (Z.pos y').
  rewrite <- Pos2Z.inj_pow.
  apply zc_div_pos_sound; auto.
  discriminate.
  intros; apply top_correct.
Qed.

Lemma zc_shr_sound x y i j :
  γ x i → γ y j → γ (zc_shr x y) (Z.shiftr i j).
Proof.
  unfold zc_shr.
  destruct y as [yr ym]. destruct ym; auto.
  intros K H. apply congr_0_eq in H. subst.
  unfold zc_mod, zc_rem.
  destruct yr as [|y'|y'].
  exact K.
  rewrite Z.shiftr_div_pow2.
  rewrite <- Pos2Z.inj_pow. apply zc_div_pos_sound; auto.
  discriminate.
  rewrite Z.shiftr_mul_pow2.
  change (- Z.neg y') with (Z.pos y').
  rewrite <- Pos2Z.inj_pow.
  apply zc_mul_sound; auto. exists 0. simpl. ring.
  discriminate.
  intros; apply top_correct.
Qed.

Lemma zc_remainder_sound x y i j :
  j <> 0 -> γ x i → γ y j → γ (zc_remainder x y) (Z.rem i j).
Proof.
  unfold zc_remainder.
  destruct y as [yr ym], x as [xr xm]. simpl. intros Hj [qx Hx] [qy Hy]. simpl in *.
  destruct ((ym =? 0)%N && Zdivides_dec yr xr && Zdivides_dec yr xm) eqn:?.
  { rewrite !Bool.andb_true_iff in Heqb.
    do 2 destruct Zdivides_dec; try easy. destruct ym; try easy. clear Heqb.
    exists 0. simpl. replace (Z.rem i j) with 0. auto. symmetry. apply Z.rem_divide. auto.
    simpl in Hy. assert (yr=j) by omega. clear Hy. subst.
    replace i with (i + qx * xm - qx*xm) by ring.
    apply Z.divide_sub_r. auto. apply Z.divide_mul_r. auto.
  }
  set (m:=Z.abs_N (Z.gcd xm (Z.gcd ym yr))).
  assert (m|xr - Z.rem i j).
  { subst m xr yr. rewrite Zabs2N.id_abs. apply Zdivide_Zabs_inv_l.
    rewrite Z.rem_eq by auto.
    replace (i + qx * xm - (i - j * (i ÷ j)))
    with (xm * qx + ((j + qy * ym) * (i ÷ j) - ym * (qy * (i ÷ j)))) by ring.
    apply Z.divide_add_r. apply Z.divide_mul_l, Zgcd_divides_l.
    eapply Zdivides_trans. apply Zgcd_divides_r.
    auto using Z.mul_divide_mono_r, Z.divide_sub_r, Z.divide_mul_l, Zgcd_divides_l, Zgcd_divides_r.
  }
  unfold zc_gamma. simpl. destruct m. apply Z.divide_0_l in H.
  exists 0. simpl. rewrite <- H. ring.
  eapply congr_trans, congr_mod_compat. 2:reflexivity.
  destruct H as [q H]. exists q. change (xr = Z.rem i j + q * N.pos p). omega.
Qed.

Lemma lor_div:
  ∀ a, (Z.lor a (-a)|a).
Proof.
intros. destruct (lor_opp a) as [|[k Hk]].
- exists 0. auto.
- simpl Z.opp. rewrite Hk.
  exists (-Z.shiftr a k).
  apply Z.bits_inj'. intros.
  rewrite Z.mul_opp_opp, <- Z.shiftl_mul_pow2, Z.shiftl_spec by (zify; omega).
  destruct (Z_lt_dec (n-k) 0). 2:rewrite Z.shiftr_spec; f_equal; omega.
  rewrite Z.testbit_neg_r with (n:=n-k) by omega.
  apply (f_equal (fun a => Z.testbit a n)) in Hk.
  replace (-2^k) with (-1*2^k) in Hk by ring.
  rewrite <- Z.shiftl_mul_pow2, Z.lor_spec, Z.shiftl_spec, Z.testbit_neg_r with (n:=n-k) in Hk by (zify; omega).
  destruct (Z.testbit a n); easy.
Qed.

Lemma congr_testbit :
  ∀ a b c, a ≡ b [c] ->
  ∀ i, Z.testbit (Z.lor c (-c)) i = false ->
    Z.testbit a i = Z.testbit b i.
Proof.
  intros a b c [q Hq] i Hi. subst b.
  destruct (lor_opp c) as [|[k Hk]]. subst. f_equal. ring.
  rewrite Hk in Hi.
  replace (-2^k) with (-1*2^k) in Hi by ring.
  rewrite <- Z.shiftl_mul_pow2 in Hi by (zify; omega).
  destruct (Z_lt_dec i k) as [Hi'|Hi'].
  2:rewrite Z.shiftl_spec_high, Int.Ztestbit_m1 in Hi by (zify; omega); discriminate.
  clear Hi.
  pose proof (lor_div c). rewrite Hk in H. destruct H.
  rewrite H, <- Z.mul_opp_comm, Z.mul_assoc, <- Z.shiftl_mul_pow2 by (zify; omega).
  pose proof Z.add_carry_bits a (Z.shiftl (q*-x) k) false.
  destruct H0 as (d & ? & ? & ?).
  simpl in H0. rewrite Z.add_0_r in H0. rewrite H0. clear H0.
  rewrite Z.lxor_assoc, Z.lxor_spec, <- (xorb_false_r (Z.testbit a i)) at 1. f_equal.
  rewrite <- Zdiv2_div in H1.
  clear Hk H. revert k i Hi' a d H1 H2.  
  induction k using N.peano_ind; intros.
  - rewrite Z.testbit_neg_r; auto.
  - refine (let H := IHk (Z.pred i) _ (Z.div2 a) (Z.div2 d) _ _ in _).
    + zify; omega.
    + rewrite H1 at 1.
      rewrite !Z.div2_spec, Z.shiftr_lor, !Z.shiftr_land, Z.shiftr_lor,
              Z.shiftr_shiftl_l, Z.sub_1_r, N2Z.inj_succ, Z.pred_succ by (zify; omega).
      auto.
    + rewrite H1, Z.lor_spec, !Z.land_spec, Z.lor_spec, Z.shiftl_spec_low, H2,
        andb_false_r, andb_false_l by (zify; omega). auto.
    + rewrite Z.lxor_spec, Z.shiftl_spec_low by auto.
      rewrite Z.lxor_spec, Z.shiftl_spec_low, Z.div2_spec in H by (zify; omega).
      destruct i.
      * rewrite H2. auto.
      * rewrite H, Z.shiftr_spec at 1 by (zify; omega). f_equal. f_equal. omega.
      * rewrite Z.testbit_neg_r; reflexivity.
Qed.

Lemma testbit_congr :
  ∀ a b c,
    (∀ i, 0 <= i -> Z.testbit (Z.lor c (-c)) i = false ->
          Z.testbit a i = Z.testbit b i) ->
    a ≡ b [-Z.lor c (-c)].
Proof.
  intros.
  destruct (lor_opp c) as [|[k Hk]].
  { subst. replace b with a. apply congr_refl.
    apply Z.bits_inj'. intros. apply H. auto. apply Z.bits_0. }
  assert (∀ i, i < k -> Z.testbit a i = Z.testbit b i).
  { intros. destruct (Z_le_dec 0 i).
    apply H. auto. rewrite Hk. replace (-2^k) with (-1*2^k) by ring.
    rewrite <- Z.shiftl_mul_pow2, Z.shiftl_spec_low by (zify; omega). auto.
    rewrite !Z.testbit_neg_r by omega; auto. }
  clear H.
  assert (∀ i, 0 <= i -> i < k -> Z.testbit (a-b) i = false).
  { intros i Hi Hi'. clear Hk.
    pose proof Z.add_carry_bits a (Z.lnot b) true.
    destruct H as (d & ? & ? & ?).
    rewrite <- Z.div2_div in H1.
    replace (a+Z.lnot b+Z.b2z true) with (a-b) in H by (unfold Z.lnot, Z.b2z; omega).
    rewrite H, !Z.lxor_spec, Z.lnot_spec by omega.
    rewrite H0, <- negb_xorb_r, xorb_nilpotent by auto.
    replace (Z.testbit d i) with true. auto. clear H.
    revert i Hi Hi' d a b H0 H1 H2. induction k using N.peano_ind.
    - intros. zify; omega.
    - intros. destruct (Z_lt_dec 0 i). 2:rewrite <- H2; f_equal; omega.
      refine (let IHks := IHk (Z.pred i) _ _ (Z.div2 d) (Z.div2 a) (Z.div2 b) _ _ _ in _).
      + omega.
      + zify; omega.
      + intros. destruct (Z_lt_dec i0 0). rewrite !Z.testbit_neg_r; now auto.
        rewrite !Z.div2_spec, !Z.shiftr_spec by omega. apply H0. zify; omega.
      + rewrite H1 at 1.
        rewrite !Z.div2_spec, Z.shiftr_lor, !Z.shiftr_land, Z.shiftr_lor, Z.lnot_shiftr by omega. auto.
      + rewrite H1, Z.lor_spec, !Z.land_spec, Z.lor_spec, Z.lnot_spec, H0, H2, orb_negb_r,
                orb_true_r by (zify; omega).
        auto.
      + rewrite Z.div2_spec, Z.shiftr_spec in IHks by omega.
        rewrite IHks. f_equal. omega. }
  assert (Z.shiftl (Z.shiftr (a-b) k) k = a-b).
  { apply Z.bits_inj'. intros.
    rewrite Z.shiftl_spec by auto.
    destruct (Z_lt_dec n k).
    - rewrite H, Z.testbit_neg_r by omega. auto.
    - rewrite Z.shiftr_spec by omega. f_equal. ring. }
  rewrite Z.shiftl_mul_pow2 in H1 by (zify; omega).
  exists (-Z.shiftr (a-b) k). rewrite Hk, Z.opp_involutive, <- Zopp_mult_distr_l, H1.
  omega.
Qed.

Lemma zc_forward_and_sound x y x0 y0 :
  γ x x0 → γ y y0 → γ (zc_forward_and x y) (Z.land x0 y0).
Proof.
  intros Hx0 Hy0.
  assert (Hx:=congr_testbit _ _ _ Hx0).
  assert (Hy:=congr_testbit _ _ _ Hy0).
  unfold zc_forward_and.
  set (maskx := Z.lor x.(zc_mod) (-x.(zc_mod))) in *.
  set (masky := Z.lor y.(zc_mod) (-y.(zc_mod))) in *.
  set (unknownx := Z.lor maskx x.(zc_rem)).
  set (unknowny := Z.lor masky y.(zc_rem)).
  set (unknowndirect := Z.lor maskx masky).
  set (unknown := Z.land (Z.land unknownx unknowny) unknowndirect).
  set (modu := -Z.lor unknown (-unknown)).
  assert (0 <= modu).
  { unfold modu. destruct (lor_opp unknown) as [|[k Hk]]. rewrite H. compute. discriminate.
    rewrite Hk. rewrite Z.opp_involutive. apply Z.pow_nonneg. omega. }
  red. red. simpl. rewrite Z2N.id by auto. apply testbit_congr.
  replace (Z.pred modu) with (Z.lnot (-modu)) by (unfold Z.lnot; rewrite Z.opp_involutive; auto).
  intros.
  unfold modu. rewrite Z.opp_involutive.
  rewrite !Z.land_spec, Z.lnot_spec, H1, Bool.andb_true_r by auto.
  rewrite Z.lor_spec, Bool.orb_false_iff in H1. destruct H1.
  unfold unknown in H1. rewrite !Z.land_spec, !Bool.andb_false_iff in H1.
  destruct H1 as [[]|].
  - unfold unknownx in H1. rewrite Z.lor_spec, Bool.orb_false_iff in H1.
    destruct H1. rewrite Hx, H3 by auto. auto.
  - unfold unknowny in H1. rewrite Z.lor_spec, Bool.orb_false_iff in H1.
    destruct H1. rewrite Hy, H3, !Bool.andb_false_r by auto. auto.
  - unfold unknowndirect in H1. rewrite Z.lor_spec, Bool.orb_false_iff in H1.
    destruct H1. rewrite Hx, Hy by auto. auto.
Qed.

Lemma zc_forward_or_sound x y x0 y0 :
  γ x x0 → γ y y0 → γ (zc_forward_or x y) (Z.lor x0 y0).
Proof.
  intros Hx0 Hy0.
  assert (Hx:=congr_testbit _ _ _ Hx0).
  assert (Hy:=congr_testbit _ _ _ Hy0).
  unfold zc_forward_or.
  set (maskx := Z.lor x.(zc_mod) (-x.(zc_mod))) in *.
  set (masky := Z.lor y.(zc_mod) (-y.(zc_mod))) in *.
  set (unknownx := Z.lor maskx (Z.lnot x.(zc_rem))).
  set (unknowny := Z.lor masky (Z.lnot y.(zc_rem))).
  set (unknowndirect := Z.lor maskx masky).
  set (unknown := Z.land (Z.land unknownx unknowny) unknowndirect).
  set (modu := -Z.lor unknown (-unknown)).
  assert (0 <= modu).
  { unfold modu. destruct (lor_opp unknown) as [|[k Hk]]. rewrite H. compute. discriminate.
    rewrite Hk. rewrite Z.opp_involutive. apply Z.pow_nonneg. omega. }
  red. red. simpl. rewrite Z2N.id by auto. apply testbit_congr.
  replace (Z.pred modu) with (Z.lnot (-modu)) by (unfold Z.lnot; rewrite Z.opp_involutive; auto).
  intros.
  unfold modu. rewrite Z.opp_involutive.
  rewrite Z.land_spec, !Z.lor_spec, Z.lnot_spec, H1, Bool.andb_true_r by auto.
  rewrite Z.lor_spec, Bool.orb_false_iff in H1. destruct H1.
  unfold unknown in H1. rewrite !Z.land_spec, !Bool.andb_false_iff in H1.
  destruct H1 as [[]|].
  - unfold unknownx in H1. rewrite Z.lor_spec, Z.lnot_spec, Bool.orb_false_iff, negb_false_iff in H1 by auto.
    destruct H1. rewrite Hx, H3 by auto. auto.
  - unfold unknowny in H1. rewrite Z.lor_spec, Z.lnot_spec, Bool.orb_false_iff, negb_false_iff in H1 by auto.
    destruct H1. rewrite Hy, H3, !Bool.orb_true_r by auto. auto.
  - unfold unknowndirect in H1. rewrite Z.lor_spec, Bool.orb_false_iff in H1.
    destruct H1. rewrite Hx, Hy by auto. auto.
Qed.

Lemma zc_forward_xor_sound x y x0 y0 :
  γ x x0 → γ y y0 → γ (zc_forward_xor x y) (Z.lxor x0 y0).
Proof.
  intros Hx0 Hy0.
  assert (Hx:=congr_testbit _ _ _ Hx0).
  assert (Hy:=congr_testbit _ _ _ Hy0).
  unfold zc_forward_xor.
  set (maskx := Z.lor x.(zc_mod) (-x.(zc_mod))) in *.
  set (masky := Z.lor y.(zc_mod) (-y.(zc_mod))) in *.
  set (unknown := Z.lor maskx masky).
  set (modu := -Z.lor unknown (-unknown)).
  assert (0 <= modu).
  { unfold modu. destruct (lor_opp unknown) as [|[k Hk]]. rewrite H. compute. discriminate.
    rewrite Hk. rewrite Z.opp_involutive. apply Z.pow_nonneg. omega. }
  red. red. simpl. rewrite Z2N.id by auto. apply testbit_congr.
  replace (Z.pred modu) with (Z.lnot (-modu)) by (unfold Z.lnot; rewrite Z.opp_involutive; auto).
  intros.
  unfold modu. rewrite Z.opp_involutive.
  destruct (Z_le_dec 0 i). 2:rewrite !Z.testbit_neg_r by omega; auto.
  rewrite Z.land_spec, !Z.lxor_spec, Z.lnot_spec, H1, Bool.andb_true_r by auto.
  rewrite Z.lor_spec, Bool.orb_false_iff in H1. destruct H1.
  unfold unknown in H1. rewrite Z.lor_spec, Bool.orb_false_iff in H1.
  destruct H1. rewrite Hx, Hy by auto. auto.
Qed.

Lemma zcmp_eq_spec i j : Zcmp Ceq i j = eqb true (i =? j).
Proof.
  simpl. rewrite Z.eqb_compare.
  generalize (Zcompare_elim (i = j) (i < j) (i > j) i j id id id).
  destruct (i ?= j); intuition.
Qed.

Lemma zcmp_ne_spec i j : Zcmp Cne i j = eqb false (i =? j).
Proof.
  simpl. rewrite Z.eqb_compare.
  generalize (Zcompare_elim (i = j) (i < j) (i > j) i j id id id).
  destruct (i ?= j); intuition.
Qed.

Lemma zc_cmp_eq_sound b x y i j :
  i ∈ γ(x) → j ∈ γ(y) → (if eqb b (i =? j) then 1 else 0) ∈ γ(zc_cmp_eq b x y).
Proof.
  unfold zc_cmp_eq.
  intros I J.
  destruct x as [xr [|x']];
  destruct y as [yr [|y']];
  try apply congr_0_eq in I;
  try apply congr_0_eq in J;
  simpl in *; subst.
- generalize (Z.eqb_spec xr yr). inversion 1; apply zc_bool_sound.
- destruct congr_dec as [|K]. apply top_correct.
  generalize (Z.eqb_spec xr j); inversion 1.
  elim K. destruct J as [q J]. simpl in J.
  exists q. congruence.
  destruct b; exists 0; reflexivity.
- destruct congr_dec as [|K]. apply top_correct.
  generalize (Z.eqb_spec i yr); inversion 1.
  elim K. destruct I as [q I]. simpl in I.
  exists (-q). subst. ring.
  destruct b; exists 0; reflexivity.
- apply top_correct.
Qed.

Lemma zc_cmp_sound c x y i j :
  γ x i → γ y j → γ (zc_cmp c x y) (if Zcmp c i j then 1 else 0).
Proof.
  intros I J.
  destruct c; auto.
  rewrite zcmp_eq_spec. apply zc_cmp_eq_sound; auto.
  rewrite zcmp_ne_spec. apply zc_cmp_eq_sound; auto.
  apply top_correct. apply top_correct. apply top_correct. apply top_correct.
Qed.

Instance zc_meet_sound: meet_op_correct zcongr (zcongr+⊥) Z.
Proof.
  intros [xr xm] [yr ym] a.
  intros [X Y]. unfold meet, zc_meet in *. simpl in *. unfold zc_gamma, zc_rem, zc_mod in X, Y.
  destruct congr_dec as [C|NC].
  destruct rem_eqn_particular as (p & P).
  destruct xm as [|xm]. easy.
  destruct ym as [|ym]. easy.
  unfold zc_gamma, gamma_bot, γ, zc_rem, zc_mod.
  case_eq (N.gcd (Npos xm) (Npos ym)). easy.
  intros gcd GCD. rewrite <- GCD.
  eapply congr_trans. 2: eapply congr_mod_compat.
  generalize (rem_eqn_all_solutions xr yr (Npos xm) (Npos ym)).
  rewrite <- Nlcm_is_Zlcm. unfold N.lcm.
  intros H. eapply H; try now zify.
  split; auto with congr.
  fold (N.lcm (Npos xm) (Npos ym)).
  generalize (N.lcm_eq_0 (Npos xm) (Npos ym)). destruct N.lcm; Psatz.lia.
  apply NC; clear NC. rewrite Ngcd_is_Zgcd.
  apply congr_trans with a; eapply congr_divide; eauto.
  apply congr_sym; eauto.
  apply Z.gcd_divide_l.
  apply Z.gcd_divide_r.
Qed.

Instance zc_ideal_gamma : gamma_op zcongr ideal_num := fun ab x =>
  match x with INz z => z ∈ γ ab | INf _ => False end.

Lemma extern_top_correct_id:
  ∀ (ab:zcongr) (x:Z), x ∈ γ ab -> INz x ∈ γ (extern_top ab).
Proof. intros. unfold extern_top. destruct N.eq_dec; easy. Qed.

Lemma intern_top_correct_id:
  ∀ (ab:zcongr+⊤) (x:Z), INz x ∈ γ ab -> x ∈ γ (intern_top ab).
Proof. intros. destruct ab. apply top_correct. auto. Qed.

Lemma zc_forward_unop_sound op i a :
  γ a i → ∀ n, eval_iunop op i n → γ (zc_forward_unop op a) n.
Proof.
  destruct op; intros H n K; simpl; auto; inv K.
  - (* neg *)
    destruct a. auto.
    simpl. destruct H as [q Q].
    unfold zc_gamma. simpl zc_rem; simpl zc_mod. rewrite Q.
    destruct (z.(zc_mod)). ring_simplify. auto with congr.
    eapply congr_trans. 2:apply congr_mod_compat.
    apply congr_neg_compat.
    exists q. ring.
    reflexivity.
  - (* not *)
    replace (Z.lnot i0) with (- 1 - i0).
    apply extern_top_correct_id, zc_add_sound, intern_top_correct_id, H.
    apply congr_minus_compat. exists 0. auto.
    pose proof (Z.add_lnot_diag i0). omega.
Qed.

Lemma zc_forward_binop_sound op :
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ n, eval_ibinop op i j n →
       γ (zc_forward_binop op a b) n.
Proof.
  destruct op; intros i x X j y Y a H; unfold γ; simpl; auto; try apply top_correct;
  inv H; apply extern_top_correct_id;
  apply intern_top_correct_id in X; apply intern_top_correct_id in Y.
  apply zc_add_sound; auto. apply congr_plus_compat.
  apply zc_add_sound; auto. apply congr_minus_compat.
  apply zc_mul_sound; auto.
  apply zc_quot_sound; auto.
  apply zc_remainder_sound; auto.
  apply zc_forward_and_sound; auto.
  apply zc_forward_or_sound; auto.
  apply zc_forward_xor_sound; auto.
  apply zc_shl_sound; auto.
  apply zc_shr_sound; auto.
  apply zc_cmp_sound; auto.
Qed.

Lemma zc_backward_add_sound:
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ res, γ res (i+j) ->
  γ (zc_backward_add res a b) (INz i, INz j).
Proof.
  intros. split; simpl.
  replace i with ((i+j)-j) by ring. apply extern_top_correct_id, zc_add_sound; auto. apply congr_minus_compat.
  replace j with ((i+j)-i) by ring. apply extern_top_correct_id, zc_add_sound; auto. apply congr_minus_compat.
Qed.

Lemma zc_backward_sub_sound:
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ res, γ res (i-j) ->
  γ (zc_backward_sub res a b) (INz i, INz j).
Proof.
  intros. split; simpl.
  replace i with ((i-j)+j) by ring. apply extern_top_correct_id, zc_add_sound; auto. apply congr_plus_compat.
  replace j with (i-(i-j)) by ring. apply extern_top_correct_id, zc_add_sound; auto. apply congr_minus_compat.
Qed.

Lemma zc_backward_mod_sound:
  ∀ i a, γ a i →
  ∀ j b, j <> 0 -> γ b j →
  ∀ res, γ res (Z.rem i j) ->
  γ (zc_backward_mod res a b) (INz i, INz j).
Proof.
  intros. split. 2:easy. simpl. apply extern_top_correct_id.
  set (m:=Z.abs_N (Z.gcd (zc_mod res) (Z.gcd (zc_mod b) (zc_rem b)))).
  destruct a as [ar am], b as [br bm], res as [resr resm], H, H1, H2. simpl in *.
  assert (m | i - Z.rem i j).
  { subst m br ar. rewrite Zabs2N.id_abs. apply Zdivide_Zabs_inv_l.
    rewrite Z.rem_eq by auto.
    rewrite Z.sub_sub_distr, Z.sub_diag, Z.add_0_l.
    eapply Z.divide_trans, Z.divide_mul_l. apply Zgcd_divides_r.
    replace j with (j+x0*bm - x0*bm) at 2 by ring.
    apply Z.divide_sub_r. apply Zgcd_divides_r. apply Z.divide_mul_r, Zgcd_divides_l. }
  assert (m | Z.rem i j - resr).
  { subst m resr. rewrite Zabs2N.id_abs. apply Zdivide_Zabs_inv_l.
    rewrite Z.sub_add_distr, Z.sub_diag, Z.sub_0_l.
    apply Zdivide_opp_r, Z.divide_mul_r, Zgcd_divides_l. }
  eapply Z.divide_add_r in H4. 2:apply H3.
  replace (i - Z.rem i j + (Z.rem i j - resr))
  with (i - resr) in H4 by ring.
  destruct H4.
  unfold zc_gamma, zc_rem, zc_mod. destruct m.
  - simpl in H4. assert (i = resr) by omega. subst i. exists 0. simpl. omega.
  - rewrite Z.mod_eq by easy.
    exists (- x2 - (resr/N.pos p)).
    ring_simplify. rewrite (Z.mul_comm (N.pos p) x2), <- H4. ring.
Qed.

Lemma zc_bw_cmp_sound c:
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ res, γ res (if Zcmp c i j then 1 else 0) ->
  γ (zc_bw_cmp c res a b) (INz i, INz j).
Proof.
  unfold zc_bw_cmp. destruct res as [res_rem []]; simpl. 2:split; apply extern_top_correct_id; auto.
  intro RES.
  assert (Zcmp c i j = negb (Z_zerop res_rem)).
  { destruct RES. simpl in H1. destruct Z_zerop, Zcmp; subst; auto; omega. } clear RES.
  destruct c, Z_zerop; (try now (split; apply extern_top_correct_id; auto));
  simpl in H1; destruct (Z.compare_spec i j); try discriminate; subst;
  (apply botbind_spec with j; [split; apply extern_top_correct_id; eauto|apply meet_correct; auto]).
Qed.

Lemma zc_backward_and_l_sound:
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ res, γ res (Z.land i j) ->
  γ (zc_backward_and_l res a b) i.
Proof.
  intros i x Hx0 j y Hy0 res Hres0.
  assert (Hx:=congr_testbit _ _ _ Hx0).
  assert (Hy:=congr_testbit _ _ _ Hy0).
  assert (Hres:=congr_testbit _ _ _ Hres0).
  unfold zc_backward_and_l.
  set (maskx := Z.lor x.(zc_mod) (-x.(zc_mod))) in *.
  set (masky := Z.lor y.(zc_mod) (-y.(zc_mod))) in *.
  set (maskres := Z.lor res.(zc_mod) (-res.(zc_mod))) in *.
  set (unknown0res := Z.lor (Z.lor (zc_rem res) (Z.lnot (zc_rem y))) (Z.lor masky maskres)).
  set (unknown0 := Z.land unknown0res (Z.lor maskx (zc_rem x))).
  set (unknown1res := Z.lor (Z.lnot (zc_rem res)) maskres).
  set (unknown1 := Z.land unknown1res (Z.lor maskx (Z.lnot (zc_rem x)))).
  set (unknown := Z.land unknown0 unknown1).
  assert (∀ n, 0 <= n -> Z.testbit unknown0 n = false -> Z.testbit i n = false).
  { intros. subst unknown0 unknown0res.
    rewrite Z.land_spec, !Z.lor_spec, Z.lnot_spec, Bool.andb_false_iff, !Bool.orb_false_iff in H0 by omega.
    destruct H0 as [[[] []]|[]].
    - rewrite <- Hres, Z.land_spec, Bool.andb_false_iff, Hy in H0 by auto.
      destruct H0. auto. rewrite H0 in H1. discriminate.
    - rewrite <- Hx in H1 by auto. auto. }
  assert (∀ n, 0 <= n -> Z.testbit unknown1 n = false -> Z.testbit i n = true).
  { intros. subst unknown1 unknown1res.
    rewrite Z.land_spec, !Z.lor_spec, !Z.lnot_spec, Bool.andb_false_iff, !Bool.orb_false_iff in H1 by omega.
    destruct H1 as [[]|[]].
    - rewrite <- Hres, Z.land_spec, Bool.negb_false_iff in H1 by auto.
      destruct (Z.testbit i n). auto. discriminate.
    - rewrite <- Hx in H2 by auto.
      destruct (Z.testbit i n). auto. discriminate. }
  assert (Z.lor unknown0 unknown1 = -1).
  { apply Z.bits_inj'. intros.
    rewrite Z.lor_spec, Int.Ztestbit_m1 by auto.
    apply not_false_is_true. rewrite Bool.orb_false_iff. intro. destruct H2.
    apply H in H2. apply H0 in H3. congruence. auto. auto. }
  rewrite H1, Z.eqb_refl.
  set (modu := -Z.lor unknown (-unknown)).
  assert (0 <= modu).
  { unfold modu. destruct (lor_opp unknown) as [|[k Hk]]. rewrite H2. compute. discriminate.
    rewrite Hk. rewrite Z.opp_involutive. apply Z.pow_nonneg. omega. }
  red. red. red. red. simpl. rewrite Z2N.id by auto. apply testbit_congr.
  replace (Z.pred modu) with (Z.lnot (-modu)) by (unfold Z.lnot; rewrite Z.opp_involutive; auto).
  intros.
  unfold modu. rewrite Z.opp_involutive.
  rewrite !Z.land_spec, Z.lnot_spec, Z.ldiff_spec, H4 by auto.
  rewrite Z.lor_spec, Bool.orb_false_iff in H4.
  unfold unknown in H4. destruct H4.
  rewrite Z.land_spec, Bool.andb_false_iff in H4. destruct H4.
  - rewrite H, H4 by auto. auto.
  - rewrite H0, H4 by auto.
    apply (f_equal (fun n => Z.testbit n i0)) in H1.
    rewrite Z.lor_spec, Int.Ztestbit_m1, Bool.orb_true_iff in H1 by auto.
    destruct H1. rewrite H1. auto. congruence.
Qed.

Lemma zc_backward_or_l_sound:
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ res, γ res (Z.lor i j) ->
  γ (zc_backward_or_l res a b) i.
Proof.
  intros i x Hx0 j y Hy0 res Hres0.
  assert (Hx:=congr_testbit _ _ _ Hx0).
  assert (Hy:=congr_testbit _ _ _ Hy0).
  assert (Hres:=congr_testbit _ _ _ Hres0).
  unfold zc_backward_or_l.
  set (maskx := Z.lor x.(zc_mod) (-x.(zc_mod))) in *.
  set (masky := Z.lor y.(zc_mod) (-y.(zc_mod))) in *.
  set (maskres := Z.lor res.(zc_mod) (-res.(zc_mod))) in *.
  set (unknown1res := Z.lor (Z.lor (Z.lnot (zc_rem res)) (zc_rem y)) (Z.lor masky maskres)).
  set (unknown1 := Z.land unknown1res (Z.lor maskx (Z.lnot (zc_rem x)))).
  set (unknown0res := Z.lor (zc_rem res) maskres).
  set (unknown0 := Z.land unknown0res (Z.lor maskx (zc_rem x))).
  set (unknown := Z.land unknown0 unknown1).
  assert (∀ n, 0 <= n -> Z.testbit unknown1 n = false -> Z.testbit i n = true).
  { intros. subst unknown1 unknown1res.
    rewrite Z.land_spec, !Z.lor_spec, !Z.lnot_spec, Bool.andb_false_iff, !Bool.orb_false_iff in H0 by omega.
    destruct H0 as [[[] []]|[]].
    - rewrite <- Hres, Z.lor_spec, negb_false_iff, Bool.orb_true_iff, Hy in H0 by auto.
      destruct H0. auto. rewrite H0 in H1. discriminate.
    - rewrite <- Hx in H1 by auto. apply negb_false_iff. auto. }
  assert (∀ n, 0 <= n -> Z.testbit unknown0 n = false -> Z.testbit i n = false).
  { intros. subst unknown0 unknown0res.
    rewrite Z.land_spec, !Z.lor_spec, Bool.andb_false_iff, !Bool.orb_false_iff in H1 by omega.
    destruct H1 as [[]|[]].
    - rewrite <- Hres, Z.lor_spec in H1 by auto.
      destruct (Z.testbit i n). discriminate. auto.
    - rewrite <- Hx in H2 by auto. auto. }
  assert (Z.lor unknown0 unknown1 = -1).
  { apply Z.bits_inj'. intros.
    rewrite Z.lor_spec, Int.Ztestbit_m1 by auto.
    apply not_false_is_true. rewrite Bool.orb_false_iff. intro. destruct H2.
    apply H0 in H2. apply H in H3. congruence. auto. auto. }
  rewrite H1, Z.eqb_refl.
  set (modu := -Z.lor unknown (-unknown)).
  assert (0 <= modu).
  { unfold modu. destruct (lor_opp unknown) as [|[k Hk]]. rewrite H2. compute. discriminate.
    rewrite Hk. rewrite Z.opp_involutive. apply Z.pow_nonneg. omega. }
  red. red. red. red. simpl. rewrite Z2N.id by auto. apply testbit_congr.
  replace (Z.pred modu) with (Z.lnot (-modu)) by (unfold Z.lnot; rewrite Z.opp_involutive; auto).
  intros.
  unfold modu. rewrite Z.opp_involutive.
  rewrite !Z.land_spec, Z.lnot_spec, Z.ldiff_spec, H4 by auto.
  rewrite Z.lor_spec, Bool.orb_false_iff in H4.
  unfold unknown in H4. destruct H4.
  rewrite Z.land_spec, Bool.andb_false_iff in H4. destruct H4.
  - rewrite H0, H4 by auto. auto.
  - rewrite H, H4 by auto.
    apply (f_equal (fun n => Z.testbit n i0)) in H1.
    rewrite Z.lor_spec, Int.Ztestbit_m1, Bool.orb_true_iff in H1 by auto.
    destruct H1. rewrite H1. auto. congruence.
Qed.

Lemma zc_backward_binop_sound op :
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ n res, γ res n ->
  eval_ibinop op i j n →
  γ (zc_backward_binop op res a b) (i, j).
Proof.
  intros i x X j y Y n res RES H. inv H; (try (split; simpl; now auto)).
  - eapply zc_backward_add_sound; auto using intern_top_correct_id.
  - apply zc_backward_sub_sound; auto using intern_top_correct_id.
  - apply zc_backward_mod_sound; auto using intern_top_correct_id.
  - eapply botbind_spec with i1, zc_backward_and_l_sound; eauto using intern_top_correct_id.
    intros. eapply botbind_spec with i2, zc_backward_and_l_sound; eauto using intern_top_correct_id.
    intros. split; apply extern_top_correct_id; auto.
    rewrite Z.land_comm. auto.
  - eapply botbind_spec with i1, zc_backward_or_l_sound; eauto using intern_top_correct_id.
    intros. eapply botbind_spec with i2, zc_backward_or_l_sound; eauto using intern_top_correct_id.
    intros. split; apply extern_top_correct_id; auto.
    rewrite Z.lor_comm. auto.
  - split.
    + replace i1 with (Z.lxor (Z.lxor i1 i2) i2). apply extern_top_correct_id, zc_forward_xor_sound; auto using intern_top_correct_id.
      rewrite Z.lxor_assoc, Z.lxor_nilpotent. apply Z.lxor_0_r.
    + replace i2 with (Z.lxor (Z.lxor i1 i2) i1). apply extern_top_correct_id, zc_forward_xor_sound; auto using intern_top_correct_id.
      rewrite Z.lxor_comm, <- Z.lxor_assoc, Z.lxor_nilpotent. apply Z.lxor_0_l.
  - apply zc_bw_cmp_sound; auto using intern_top_correct_id.
Qed.

Definition enrich_query_chan {var} fev (chan:@query_chan var) : query_chan :=
  {| get_itv := chan.(get_itv);
     get_congr expr :=
       do_bot ab <- fev expr;
       meet (NotBot (intern_top ab)) (chan.(get_congr) expr);
     get_eq_expr := chan.(get_eq_expr) |}.

(* After extraction, this definition needs to be generalized on var: we use the
   unit parameter so that the value restriction applies. *)
Instance zc_ideal_non_rel var (_:unit) : ab_ideal_nonrel var zcongr :=
  { as_widen := join
  ; z_const cst := let 'ITV lo hi := cst in extern_top (zc_itv lo hi)
  ; float_const cst := All
  ; forward_unop op ab := NotBot (zc_forward_unop op ab)
  ; forward_binop op l r := NotBot (zc_forward_binop op l r)
  ; backward_unop op res ab := NotBot ab
  ; backward_binop op res l r :=
      match res with
        | All => top
        | Just res => zc_backward_binop op res l r
      end
  ; forward_var fev x ab c :=
      match ab with
        | All =>
          match c.(get_eq_expr) x with
            | Some (Leaf e) => fev e
            | _ => top
          end
        | Just _ => NotBot ab
      end
  ; backward_var fev AB bev x ab abenv := NotBot (fst abenv)
  ; enrich_query_chan := enrich_query_chan
  ; process_msg fev TAB bev msg ab := NotBot (fst ab, msg::nil)
  ; assign_msgs x zc :=
      if zc.(zc_mod) then Known_value_msg x (INz zc.(zc_rem))::nil
      else if N.eqb zc.(zc_mod) 1 then nil
           else Congr_msg x zc::nil
  ; assume_msgs x zc_old zc :=
      if zc.(zc_mod) then Known_value_msg x (INz zc.(zc_rem))::nil
      else if N.eqb zc.(zc_mod) 1 then nil
           else Congr_msg x zc::nil
  }.

Instance zc_ideal_non_rel_correct var : ab_ideal_nonrel_correct (zc_ideal_non_rel var tt) _.
Proof.
  split.
- simpl. intro. unfold join, join_top_res. rewrite zc_join_eq. auto.
- simpl. intro. unfold widen, join, join_top_res. rewrite zc_join_eq. auto.
- simpl. intro. unfold leb, zc_leb, zc_LE_dec.
  destruct Ndivides_dec. destruct congr_dec. auto.
  destruct n. apply congr_refl.
  destruct n. apply Z.divide_refl.
- simpl. intros x y LE [f|z] G. exact G.
  revert G. simpl. apply leb_correct; auto.
- intros x y [f|z]. destruct (x ⊔ y); simpl; now intuition.
  simpl. destruct (x ⊔ y) eqn:?; simpl; auto. intro. apply join_correct in H.
  rewrite Heqy0 in H. auto.
- simpl. intros x y [f|z]. destruct (x ⊓ y); simpl; now intuition.
  simpl. intros. apply meet_correct in H. destruct meet; auto.
- intros [hi lo] a H.
  simpl. unfold zc_itv.
  generalize (Z.eqb_eq hi lo). destruct (hi =? lo).
  intuition.
  assert (a = lo). hnf in H. intuition. subst.
  exists 0. simpl. ring.
  intros. apply top_correct.
- easy.
- intros. eapply zc_forward_unop_sound; eauto.
- intros. eapply zc_forward_binop_sound; eauto.
- easy.
- intros. destruct res_ab. split; apply top_correct.
  simpl backward_binop. eapply zc_backward_binop_sound, H2; auto.
- intros. simpl forward_var. destruct ab. 2:auto.
  pose proof (get_eq_expr_correct H1 x). destruct @get_eq_expr as [[]|]; auto.
  apply H. specialize (H2 _ eq_refl). inv H2. auto.
- intros. apply H2.
- split; intros; try (apply H0; now auto).
  eapply botbind_spec, H; eauto. intros. apply meet_correct. split.
  apply intern_top_correct_id. auto.
  eapply get_congr_correct, H1. auto.
- split. apply H2. repeat constructor. auto.
- intros. unfold assign_msgs, zc_ideal_non_rel.
  destruct v as [rem [|[]]]; repeat constructor; auto.
  simpl. destruct (ρ x), H. simpl in H. f_equal. omega.
- intros. unfold assume_msgs, zc_ideal_non_rel.
  destruct vNew as [rem [|[]]]; repeat constructor; auto.
  simpl. destruct (ρ x), H0. simpl in H0. f_equal. omega.
Qed.

Require Import ShareTree.

Module Make(T:SHARETREE).

  Module IBD := IdealBoxDomain.Make(T).

  Definition var := IBD.var.
  Definition t := @IBD.t zcongr.
  Existing Instance IBD.var_eq.

  Global Instance zcongr_ideal_env : ab_ideal_env (var:=var) t :=
    IBD.ideal_box_domain _ (zc_ideal_non_rel_correct _).

End Make.
