(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib Util ShareTree Integers ArithLib AdomLib ToString Axioms IdealExpr.

Require Psatz.

Local Unset Elimination Schemes.
Local Coercion Z_of_N : N >-> Z.

(** * Data type. *)
Record zcongr := ZC { zc_rem : Z;
                      zc_mod : N;
                      zc_rem_itv : (zc_mod > 0)%N -> 0 <= zc_rem < zc_mod }.

Instance zc_to_string : ToString zcongr := λ zc,
  ("≡ " ++ to_string zc.(zc_rem) ++
   " [" ++ to_string zc.(zc_mod) ++ "]")%string.

(** * Partial order. *)
Definition zc_LE (x y: zcongr) : Prop :=
    ( y.(zc_mod) | x.(zc_mod) )
  ∧ ( x.(zc_rem) ≡ y.(zc_rem) [y.(zc_mod)] ).

Definition zc_LE_dec (x y: zcongr) : { zc_LE x y } + { ¬ zc_LE x y } :=
  match Ndivides_dec y.(zc_mod) x.(zc_mod) with
  | left D => match congr_dec y.(zc_mod) x.(zc_rem) y.(zc_rem) with
              | left C => left (conj D C)
              | right C => right (fun X => C (proj2 X))
              end
  | right D => right (fun X => D (proj1 X))
  end.

Program Instance zc_top : top_op zcongr :=
  let t := {| zc_rem := 0; zc_mod := 1; zc_rem_itv := _ |} in t.
Next Obligation. omega. Qed.

Instance zc_leb : leb_op zcongr := zc_LE_dec.

Definition zc_share a a' :=
  if Z.eqb a.(zc_rem) a'.(zc_rem) &&
     N.eqb a.(zc_mod) a'.(zc_mod) then a' else a.

Lemma zc_share_eq:
  forall a a', zc_share a a' = a.
Proof.
  unfold zc_share. intros.
  destruct (Z.eqb_spec a.(zc_rem) a'.(zc_rem)),
           (N.eqb_spec a.(zc_mod) a'.(zc_mod)); auto.
  destruct a, a'; simpl in e, e0; simpl. subst. f_equal; apply Axioms.proof_irr.
Qed.

(** * Lattice Structure. *)
Program Instance zc_join : join_op zcongr (zcongr+⊤) := fun x y =>
  if Z.eqb x.(zc_rem) y.(zc_rem) && N.eqb x.(zc_mod) y.(zc_mod) then Just x
  else
    let m : N := N.gcd (N.gcd x.(zc_mod) y.(zc_mod))
                       (Nabs_diff x.(zc_rem) y.(zc_rem)) in
    match N.eq_dec m 1%N return _ with
      | left _ => All
      | right _ =>
        let res :=
            {| zc_rem :=
                 match m return _ with
                   | N0 => x.(zc_rem)
                   | _ => x.(zc_rem) mod m
                 end
               ; zc_mod := m |}
        in
        Just (zc_share (zc_share res x) y)
    end.
Next Obligation.
  destruct (N.gcd (N.gcd (zc_mod x) (zc_mod y)) (Nabs_diff (zc_rem x) (zc_rem y))) eqn:EQmod.
  discriminate. apply Z_mod_lt. auto.
Qed.

Lemma zc_join_eq:
  forall x:zcongr, x ⊔ x = Just x.
Proof.
  intros. unfold join, zc_join. rewrite Z.eqb_refl, N.eqb_refl. auto.
Qed.

Program Instance zc_meet : meet_op zcongr (zcongr+⊥) := fun x y =>
  let gcd : N := N.gcd x.(zc_mod) y.(zc_mod) in
  match congr_dec gcd x.(zc_rem) y.(zc_rem) return _ with
  | left HH =>
    if x.(zc_mod) then NotBot x else
    if y.(zc_mod) then NotBot y else
    let m : N := (x.(zc_mod) * (y.(zc_mod) / gcd))%N in
    let (n, K) :=
        rem_eqn_particular x.(zc_rem) y.(zc_rem)
          x.(zc_mod) y.(zc_mod)
                         (N2Z.is_nonneg _)
                         (N2Z.is_nonneg _)
                         (eq_ind _ (fun n : Z => _ ≡ _ [n]) HH _ (Ngcd_is_Zgcd _ _))
    in NotBot {| zc_rem := n mod m; zc_mod := m |}
  | right _ => Bot
  end.
Next Obligation. apply Z_mod_lt. zify; omega. Qed.
  
Instance zc_gamma : gamma_op zcongr Z :=
  fun x z => z ≡ x.(zc_rem) [x.(zc_mod)].

Instance zc_top_correct : top_op_correct zcongr Z.
Proof. intros i. exists (-i). simpl. ring. Qed.

Instance zc_leb_correct : leb_op_correct zcongr Z.
Proof.
  intros x y. simpl. unfold leb, zc_leb. destruct zc_LE_dec as [((a & A),(b & B))|].
  intros _ i.
  intros (q & Q). exists (q * a + b).
  rewrite B, Q, A. ring.
  discriminate.
Qed.

Instance zc_join_correct : join_op_correct zcongr (zcongr+⊤) Z.
Proof.
  intros [xr xm] [yr ym] i.
  simpl. unfold join, zc_join. simpl.
  destruct ((xr =? yr) && (xm =? ym)%N) eqn:?.
  rewrite Bool.andb_true_iff, Z.eqb_eq, N.eqb_eq in Heqb. destruct Heqb as [-> ->].
  now intuition.
  destruct N.eq_dec. easy.
  rewrite !zc_share_eq.
  unfold γ, zc_gamma. simpl.
  case_eq (N.gcd (N.gcd xm ym) (Nabs_diff xr yr)).
  repeat rewrite N.gcd_eq_0.
  intros ((-> & ->) & XY).
  assert (xr = yr).
  destruct (Nabs_diff_case xr yr) as [(A,B)|(A,B)];
    rewrite B in XY; clear B.
  destruct (Z_dec (yr - xr) 0) as [[|]|];
    try Psatz.lia. destruct (yr - xr); simpl in *; discriminate.
  destruct (Z_dec (xr - yr) 0) as [[|]|];
    try Psatz.lia. destruct (xr - yr); simpl in *; discriminate.
  subst. intuition.
  intros p H; rewrite <- H.
  intros [X|Y].
  eapply congr_trans.
  2: eapply congr_mod_compat.
  eapply congr_divide. eassumption.
  repeat rewrite Ngcd_is_Zgcd.
  eapply Zdivides_trans; apply Zgcd_divides_l.
  rewrite H; reflexivity.
  eapply congr_trans.
  2: eapply congr_mod_compat.
  2: rewrite H; reflexivity.
  repeat rewrite Ngcd_is_Zgcd.
  apply congr_trans with (xr + (yr - xr)).
  replace (xr + (yr - xr)) with yr by ring.
  eapply congr_divide. eassumption.
  eapply Zdivides_trans. apply Zgcd_divides_l. apply Zgcd_divides_r.
  eapply congr_divide.
  2: apply Zgcd_divides_r.
  destruct (Nabs_diff_case xr yr) as
      [(A,B)|(A,B)]; rewrite B; rewrite Z2N.id.
  exists (-1). ring. Psatz.lia.
  exists (1). ring. Psatz.lia.
Qed.
