(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8.
Require Import ZArith.
Require Import Integers.
Require Import Reals Fourier.
Require Import Fappli_IEEE.
Require Import Fappli_IEEE_bits.
Require Import Fappli_IEEE_extra.
Require Import Fappli_double_round.
Require Import Fcore_FLT.
Require Import Fcore_float_prop.
Require Import Fcore_defs.
Require Import Fcore_Raux.
Require Import Fcore_Zaux.
Require Import Fcore_generic_fmt.
Require Import Fcore_rnd_ne.
Require Import Fcore_digits.
Require Import Floats.

Ltac Rlt_bool_case H :=
  match type of H with
    | context [Rlt_bool ?x ?y] =>
      destruct (Rlt_bool_spec x y)
  end.

Ltac Rle_bool_case H :=
  match type of H with
    | context [Rle_bool ?x ?y] =>
      destruct (Rle_bool_spec x y)
  end.

Ltac check_nodup H :=
  let typ := type of H in
  match goal with
    | [H':typ |- _] =>
      match H' with
        | H => fail 1
        | _ => fail 2
      end
    | _ => idtac
  end.

Ltac use_float_spec :=
  let H := fresh in
  match goal with
    | [_ : context [binary_normalize ?prec ?emax ?Hprec ?Hemax ?mode ?m ?e ?s] |- _] =>
      assert (H:=binary_normalize_correct prec emax Hprec Hemax mode m e s);
      check_nodup H
    | [|- context [binary_normalize ?prec ?emax ?Hprec ?Hemax ?mode ?m ?e ?s]] =>
      assert (H:=binary_normalize_correct prec emax Hprec Hemax mode m e s);
      check_nodup H
    | [_ : context [Bconv ?prec1 ?emax1 ?prec2 ?emax2 ?Hprec2 ?Hemax2 ?pl ?mode ?x] |- _] =>
      assert (H:=Bconv_correct prec1 emax1 prec2 emax2 Hprec2 Hemax2 pl mode x);
      check_nodup H
    | [|- context [Bconv ?prec1 ?emax1 ?prec2 ?emax2 ?Hprec2 ?Hemax2 ?pl ?mode ?x]] =>
      assert (H:=Bconv_correct prec1 emax1 prec2 emax2 Hprec2 Hemax2 pl mode x);
      check_nodup H
    | [_ : context [BofZ ?prec ?emax ?Hprec ?Hemax ?x] |- _] =>
      assert (H:=BofZ_correct prec emax Hprec Hemax x);
      check_nodup H
    | [|- context [BofZ ?prec ?emax ?Hprec ?Hemax ?x]] =>
      assert (H:=BofZ_correct prec emax Hprec Hemax x);
      check_nodup H
    | [_ : context [Bplus ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y] |- _] =>
      assert (H:=Bplus_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [|-context [Bplus ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y]] =>
      assert (H:=Bplus_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [_ : context [Bminus ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y] |- _] =>
      assert (H:=Bminus_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [|-context [Bminus ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y]] =>
      assert (H:=Bminus_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [_ : context [Bmult ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y] |- _] =>
      assert (H:=Bmult_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [|-context [Bmult ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y]] =>
      assert (H:=Bmult_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [_ : context [Bdiv ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y] |- _] =>
      assert (H:=Bdiv_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
    | [|-context [Bdiv ?prec ?emax ?Hprec ?Hemax ?pl ?mode ?x ?y]] =>
      assert (H:=Bdiv_correct prec emax Hprec Hemax pl mode x y);
      check_nodup H
  end.

Ltac use_float_specs := repeat use_float_spec.

Hint Resolve fexp_correct valid_rnd_round_mode.

Definition Fleb {prec emax} f1 f2 :=
  match Bcompare prec emax f1 f2 with
    | Some (Eq|Lt) => true
    | Some Gt | None => false
  end.

Lemma Fleb_Rle:
  ∀ prec emax f1 f2,
    is_finite _ _ f1 = true -> is_finite _ _ f2 = true ->
    @Fleb prec emax f1 f2 = Rle_bool (B2R _ _ f1) (B2R _ _ f2).
Proof.
  intros. unfold Fleb.
  rewrite Bcompare_correct by auto. auto.
Qed.

Lemma Bcompare_nan:
  ∀ prec emax f1 f2,
    is_nan _ _ f1 = false -> is_nan _ _ f2 = false ->
    Bcompare prec emax f1 f2 <> None.
Proof.
  intros. intro.
  destruct f1 as [|[]| |[]], f2 as [|[]| |[]]; try discriminate.
  - simpl in H1. destruct Zcompare; discriminate.
  - simpl in H1. destruct Zcompare; discriminate.
Qed.

Lemma Fleb_total:
  ∀ prec emax f1 f2,
    is_nan _ _ f1 = false -> is_nan _ _ f2 = false ->
    @Fleb prec emax f1 f2 = true \/
     Fleb           f2 f1 = true.
Proof.
  unfold Fleb.
  intros. rewrite Bcompare_swap. pose proof Bcompare_nan _ _ _ _ H0 H.
  destruct Bcompare as [[]|]; auto.
Qed.

Lemma Bcompare_refl:
  ∀ prec emax f,
    is_nan _ _ f = false ->
    Bcompare prec emax f f = Some Eq.
Proof.
  intros.
  pose proof (Bcompare_nan _ _ _ _ H H).
  pose proof (Bcompare_swap _ _ f f).
  destruct Bcompare as [[]|]; simpl in H1; congruence.
Qed.

Lemma Fleb_refl:
  ∀ prec emax f,
    is_nan _ _ f = false ->
    @Fleb prec emax f f = true.
Proof.
  intros. unfold Fleb. rewrite Bcompare_refl; auto.
Qed.

Hint Resolve Fleb_refl.

Lemma Bcompare_trans:
  ∀ prec emax f1 f2 f3 c,
    is_nan _ _ f2 = false ->
    Bcompare prec emax f1 f2 = c ->
    Bcompare prec emax f2 f3 = c ->
    Bcompare prec emax f1 f3 = c.
Proof.
  unfold Bcompare. intros.
  destruct f1 as [|[]| |[]], f2 as [|[]| |[]], f3 as [|[]| |[]];
    simpl in H; try congruence;
  destruct (Z.compare_spec e e1), (Z.compare_spec e1 e3), (Z.compare_spec e e3);
  subst; try omega; try congruence;
  rewrite !Pos.compare_cont_spec in *;
  destruct (Pos.compare_spec m0 m1), (Pos.compare_spec m m0), (Pos.compare_spec m m1);
  simpl in H1; try congruence; Psatz.lia.
Qed.

Lemma Bcompare_Eq_trans:
  ∀ prec emax f1 f2 f3 c,
    Bcompare prec emax f1 f2 = Some Eq ->
    Bcompare prec emax f2 f3 = c ->
    Bcompare prec emax f1 f3 = c.
Proof.
  unfold Bcompare. intros.
  destruct f1 as [|[]| |[]], f2 as [|[]| |[]], f3 as [|[]| |[]];
    try congruence;
  destruct (Z.compare_spec e e1), (Z.compare_spec e1 e3), (Z.compare_spec e e3);
  subst; try omega; try congruence;
  rewrite !Pos.compare_cont_spec in *;
  destruct (Pos.compare_spec m0 m1), (Pos.compare_spec m m0), (Pos.compare_spec m m1);
  simpl in H; try congruence; Psatz.lia.
Qed.

Lemma Bcompare_trans_Eq:
  ∀ prec emax f1 f2 f3 c,
    Bcompare prec emax f1 f2 = c ->
    Bcompare prec emax f2 f3 = Some Eq ->
    Bcompare prec emax f1 f3 = c.
Proof.
  unfold Bcompare. intros.
  destruct f1 as [|[]| |[]], f2 as [|[]| |[]], f3 as [|[]| |[]];
    try congruence;
  destruct (Z.compare_spec e e1), (Z.compare_spec e1 e3), (Z.compare_spec e e3);
  subst; try omega; try congruence;
  rewrite !Pos.compare_cont_spec in *;
  destruct (Pos.compare_spec m0 m1), (Pos.compare_spec m m0), (Pos.compare_spec m m1);
  simpl in H0; try congruence; Psatz.lia.
Qed.

Lemma Fleb_nan_l:
  ∀ prec emax f1 f2,
    @Fleb prec emax f1 f2 = true -> is_nan _ _ f1 = false.
Proof.
  intros. unfold Fleb in H. destruct f1, f2; simpl in *; congruence.
Qed.

Hint Resolve Fleb_nan_l.

Lemma Fleb_nan_r:
  ∀ prec emax f1 f2,
    @Fleb prec emax f1 f2 = true -> is_nan _ _ f2 = false.
Proof.
  intros. unfold Fleb in H. destruct f1 as [|[]| |[]], f2; simpl in *; congruence.
Qed.

Hint Resolve Fleb_nan_r.

Lemma Fleb_trans:
  ∀ prec emax f1 f2 f3,
    Fleb            f1 f2 = true ->
    Fleb            f2 f3 = true ->
    @Fleb prec emax f1 f3 = true.
Proof.
  intros. assert (is_nan _ _ f2 = false) by eauto. unfold Fleb in *.
  pose proof (Bcompare_trans _ _ f1 f2 f3).
  pose proof (Bcompare_trans_Eq _ _ f1 f2 f3).
  pose proof (Bcompare_Eq_trans _ _ f1 f2 f3).
  destruct Bcompare as [[]|], Bcompare  as [[]|], Bcompare  as [[]|]; try reflexivity;
  (specialize (H2 _ H1 eq_refl eq_refl) || clear H2);
  (specialize (H3 _ eq_refl eq_refl) || clear H3);
  (specialize (H4 _ eq_refl eq_refl) || clear H4);
  discriminate.
Qed.

Lemma binary_normalize_is_nan:
  forall prec emax H1 H2 m e s mode,
    is_nan _ _ (binary_normalize prec emax H1 H2 mode m e s) = false.
Proof.
  intros. use_float_spec.
  destruct Rlt_bool.
  destruct H as [? []]. destruct binary_normalize; try discriminate; reflexivity.
  destruct mode; destruct binary_normalize; try reflexivity; destruct n; try discriminate;
    destruct Rlt_bool; discriminate.
Qed.

Lemma Bconv_incr:
  ∀ prec emax prec' emax' Hprec' Hmax' pl f1 f2,
    Fleb f1 f2 = true ->
    Fleb (Bconv prec emax prec' emax' Hprec' Hmax' pl mode_NE f1)
         (Bconv prec emax prec' emax' Hprec' Hmax' pl mode_NE f2) = true.
Proof.
  intros.
  assert (is_nan _ _ f1 = false) by eauto.
  assert (is_nan _ _ f2 = false) by eauto.
  use_float_specs.
  destruct (is_finite _ _ f1) eqn:FIN1;
  [specialize (H2 eq_refl)|clear H2; destruct f1 as [|[]| |]; try discriminate];
  (destruct (is_finite _ _ f2) eqn:FIN2;
  [specialize (H3 eq_refl)|clear H3; destruct f2 as [|[]| |]; try discriminate]);
  try reflexivity;
  try (pose proof H; rewrite Fleb_Rle in H by auto; Rle_bool_case H; [|discriminate];
       apply (round_le radix2 (FLT_exp (3-emax'-prec') prec') (round_mode mode_NE)) in H5);
  repeat match goal with H : context [Rlt_bool ?a ?b] |- _ =>
    Rlt_bool_case H;
    [destruct H as [H []]|
     unfold binary_overflow in H; simpl in H;
     apply B2FF_inj with (y:=B754_infinity _ _ _) in H]
  end.
  - rewrite Fleb_Rle by auto. apply Rle_bool_true. rewrite H2, H3. auto.
  - rewrite H2. destruct (Bsign _ _ f1) eqn:SIG.
    destruct (Bconv _ _ _ _ Hprec' Hmax' pl mode_NE f2) as [|[]| |];
      try reflexivity; discriminate.
    exfalso. destruct f1 as [| | |[]]; try discriminate.
    apply Rabs_lt_inv in H6. destruct H6.
    apply Rabs_ge_inv in H9. destruct H9. 2:fourier.
    apply Rle_lt_trans with (r3:=0%R), Rlt_not_le in H9.
    2:apply Ropp_lt_gt_0_contravar, bpow_gt_0.
    erewrite <- round_0 in H9 by auto. apply H9, round_le; eauto.
    apply F2R_ge_0_compat. compute. discriminate.
  - rewrite H3. destruct (Bsign _ _ f2) eqn:SIG.
    2:destruct (Bconv _ _ _ _ Hprec' Hmax' pl mode_NE f1) as [|[]| |[]];
      try reflexivity; discriminate.
    exfalso. destruct f2 as [| | |[]]; try discriminate.
    apply Rabs_lt_inv in H7. destruct H7.
    apply Rabs_ge_inv in H6. destruct H6. fourier.
    apply Rlt_le_trans with (r1:=0%R), Rlt_not_le in H6.
    2:apply bpow_gt_0.
    erewrite <- round_0 in H6 by auto. apply H6, round_le; eauto.
    apply F2R_le_0_compat. compute. discriminate.
  - rewrite H2, H3.
    destruct f1 as [| | |[]], f2 as [| | |[]]; try discriminate; reflexivity.
  - destruct f1 as [| | |[]]; discriminate.
  - destruct f1 as [| | |[]]; discriminate.
  - destruct Bconv as [| | |[]]; try reflexivity; discriminate.
  - destruct Bconv as [|[]| |]; try reflexivity; discriminate.
  - destruct Bconv as [| | |[]]; try reflexivity; discriminate.
  - destruct Bconv as [|[]| |]; try reflexivity; discriminate.
  - destruct f2 as [| | |[]]; discriminate.
  - destruct f2 as [| | |[]]; discriminate.
Qed.

Transparent Float.of_single Float.to_single.

Lemma singleoffloat_is_nan:
  forall f,
    is_nan _ _ f = false ->
    is_nan _ _ (Float.of_single (Float.to_single f)) = false.
Proof.
  intros. destruct (Float.to_single f) eqn:?; try reflexivity.
  destruct f; try discriminate.
  simpl in Heqf0.
  pose proof binary_normalize_is_nan 24 128 (@eq_refl _ Lt) (@eq_refl _ Lt) (cond_Zopp b0 (Z.pos m)) e b0 mode_NE.
  rewrite Heqf0 in H0. discriminate.
  apply binary_normalize_is_nan.
Qed.

Lemma floatofsingle_exact:
  forall f, B2R _ _ (Float.of_single f) = B2R _ _ f
            /\ (is_finite _ _ f = true -> is_finite _ _ (Float.of_single f) = true).
Proof.
  destruct f; try (split; try discriminate; reflexivity). unfold Float.of_single.
  use_float_specs.
  rewrite round_generic in H. 2:auto.
  rewrite Rlt_bool_true in H. now intuition.
  eapply Rlt_le_trans, bpow_le. apply abs_B2R_lt_emax. omega.
  eapply generic_inclusion_ln_beta, generic_format_B2R with (x:=B754_finite _ _ _ _ _ e0).
  unfold FLT_exp. zify; omega.
Qed.

Lemma floatofsingle_nan:
  ∀ f, is_nan _ _ f = false ->
       is_nan _ _ (Float.of_single f) = false.
Proof.
  destruct f; try reflexivity; try discriminate.
  intros. apply binary_normalize_is_nan.
Qed.

Lemma tooftosingle:
  ∀ f,
  Float.to_single (Float.of_single (Float.to_single f)) = Float.to_single f.
Proof.
  destruct f; try reflexivity.
Transparent Float.to_single Float.of_single Float.reduce_pl Float.expand_pl Float.transform_quiet_pl.
  - assert (∀ p,
      Pos.shiftr_nat (Pos.lor (Pos.shiftl_nat
        (Pos.shiftr_nat (Pos.lor p (nat_iter 51 xO 1%positive)) 29)
          29) (nat_iter 51 xO 1%positive)) 29 =
      Pos.shiftr_nat
        (Pos.lor p (nat_iter 51 xO 1%positive)) 29).
    { intro. assert (∀ x y, N.pos x = N.pos y -> x = y) by congruence.
      apply H. clear H.
      change (nat_iter 51 xO xH) with (Pos.shiftl_nat xH 51).
      assert (∀ x y, N.pos (Pos.lor x y) = N.lor (N.pos x) (N.pos y)) by auto.
      assert (∀ n p, N.pos (Pos.shiftl_nat p n) = N.shiftl_nat (N.pos p) n).
      { clear. induction n. auto.
        intros. unfold Pos.shiftl_nat, N.shiftl_nat in *. simpl.
        rewrite <- IHn. auto. }
      assert (∀ n p, N.shiftr_nat (N.pos p) n <> N0 ->
                     N.pos (Pos.shiftr_nat p n) = N.shiftr_nat (N.pos p) n).
      { clear. induction n. auto.
        intros. destruct p.
        - unfold Pos.shiftr_nat, N.shiftr_nat in *. rewrite !nat_iter_succ_r in *.
          apply IHn. auto.
        - unfold Pos.shiftr_nat, N.shiftr_nat in *. rewrite !nat_iter_succ_r in *.
          apply IHn. auto.
        - destruct H. unfold N.shiftr_nat. clear.
          induction n. reflexivity. simpl in *. rewrite IHn. auto. }
      rewrite !H1, !H, !H0, !H1, !H, H0.
      - apply Ndigits.Nbit_faithful. intro.
        rewrite !Ndigits.Nshiftr_nat_spec, !Ndigits.Nor_semantics,
                Ndigits.Nshiftl_nat_spec_high,
                Ndigits.Nshiftr_nat_spec, !Ndigits.Nor_semantics
          by omega.
        replace (n0+29-29+29)%nat with (n0+29)%nat by omega.
        destruct N.testbit_nat, N.testbit_nat; reflexivity.
      - rewrite H, H0.
        intro. apply (f_equal (fun n => N.testbit_nat n 22)) in H2.
        rewrite Ndigits.Nshiftr_nat_spec, Ndigits.Nor_semantics,
                Ndigits.Nshiftl_nat_spec_high, orb_true_r in H2
             by omega.
        discriminate.
      - rewrite H, H0.
        intro. apply (f_equal (fun n => N.testbit_nat n 22)) in H2.
        rewrite Ndigits.Nshiftr_nat_spec, Ndigits.Nor_semantics,
                Ndigits.Nshiftl_nat_spec_high, orb_true_r in H2
             by omega.
        discriminate.
      - rewrite H, H0.
        intro. apply (f_equal (fun n => N.testbit_nat n 22)) in H2.
        rewrite Ndigits.Nshiftr_nat_spec, Ndigits.Nor_semantics,
                Ndigits.Nshiftl_nat_spec_high, orb_true_r in H2
             by omega.
        discriminate. }
    simpl. f_equal.
    unfold Float.reduce_pl at 1 4.
    match goal with
      |- exist ?P ?x ?Hx = exist ?P ?y ?Hy =>
      assert (EQ:x = y);
      [|generalize (Hx:P x); rewrite EQ; generalize (Hy:P y); clear]
    end.
    destruct Archi.float_of_single_preserves_sNaN;
      try rewrite Float.transform_quiet_pl_idempotent;
      apply H.
    assert (∀ (b:bool) (e:b=b), e = eq_refl).
    { destruct b; intros;
      exact (match e with eq_refl => eq_refl end). }
    intros. f_equal. revert e0. destruct e. auto.
Opaque Float.reduce_pl Float.expand_pl Float.transform_quiet_pl.
  - simpl.
    apply Bconv_narrow_widen. omega. omega. apply binary_normalize_is_nan.
Opaque Float.to_single Float.of_single.
Qed.

Lemma BofZ_incr:
  ∀ z1 z2 prec emax Hprec Hmax,
    z1 <= z2 ->
    Fleb (BofZ prec emax Hprec Hmax z1)
         (BofZ _ _       Hprec Hmax z2) = true.
Proof.
  intros. use_float_specs.
  repeat match goal with H : context [Rlt_bool ?a ?b] |- _ =>
    Rlt_bool_case H;
    [destruct H as [H []]|
     unfold binary_overflow in H; simpl in H;
     apply B2FF_inj with (y:=B754_infinity _ _ _) in H]
  end.
  - rewrite Fleb_Rle, H0, H1 by auto. apply Rle_bool_true.
    apply round_le; auto. apply Z2R_le. auto.
  - rewrite H0. destruct (Zlt_bool_spec z1 0).
    destruct (BofZ _ _ Hprec Hmax z2); try discriminate; reflexivity.
    exfalso. eapply Rlt_le_trans, Rlt_not_le in H5. 2:apply H2.
    apply H5. rewrite <- !round_NE_abs, <- !Z2R_abs, !Z.abs_eq by (auto; omega).
    apply round_le, Z2R_le, H; auto. apply (valid_rnd_round_mode mode_NE).
  - rewrite H1. destruct (Zlt_bool_spec z2 0).
    2:destruct (BofZ _ _ Hprec Hmax z1) as [| | |[]]; try discriminate; reflexivity.
    exfalso. eapply Rlt_le_trans, Rlt_not_le in H3. 2:apply H2.
    apply H3. rewrite <- !round_NE_abs, <- !Z2R_abs, !Z.abs_neq by (auto; omega).
    apply round_le, Z2R_le; auto. apply (valid_rnd_round_mode mode_NE). omega.
  - rewrite H0, H1. destruct (Zlt_bool_spec z1 0), (Zlt_bool_spec z2 0); try reflexivity. omega.
Qed.

Lemma zoffloat_incr:
  forall prec emax f1 f2 z1 z2,
    @Fleb prec emax f1 f2 = true ->
    ZofB _ _ f1 = Some z1 ->
    ZofB _ _ f2 = Some z2 ->
    z1 <= z2.
Proof.
  intros.
  rewrite ZofB_correct in *.
  destruct (is_finite _ _ f1) eqn:FIN1; inv H0.
  destruct (is_finite _ _ f2) eqn:FIN2; inv H1.
  rewrite Fleb_Rle in H by auto. Rle_bool_case H; [|discriminate].
  apply Ztrunc_le, H0.
Qed.

Definition next_ulp_ff prec emax (f:full_float) : full_float :=
  match f return _ with
    | F754_nan _ _ => f
    | F754_infinity true => binary_overflow prec emax mode_UP true
    | F754_finite true m e =>
      let m' := Pos.pred m in
      if Z.eqb e (3-emax-prec) then
        if Pos.eqb m 1 then F754_zero false
        else F754_finite true m' e
      else
        if Z.eqb (Z.pos (digits2_pos m')) prec then
          F754_finite true m' e
        else
          F754_finite true (match 2 ^ prec - 1 with Zpos m => m | _ => xH end)%Z (Z.pred e)
    | F754_zero _ => F754_finite false 1 (3-emax-prec)
    | F754_finite false m e =>
      let m' := Pos.succ m in
      if Z.leb (Z.pos (digits2_pos m')) prec then
        F754_finite false m' e
      else
        if Z.eqb e (emax-prec) then F754_infinity false
        else F754_finite false (match 2 ^ (prec-1) with Zpos m => m | _ => xH end)%Z (Z.succ e)
    | F754_infinity false => f
  end.

Lemma next_ulp_ff_is_nan:
  forall prec emax f,
    is_nan_FF f = false ->
    is_nan_FF (next_ulp_ff prec emax f) = false.
Proof.
  destruct f as [|[]| |]; try reflexivity; try discriminate.
  intro. unfold next_ulp_ff.
  repeat match goal with
           | |- context [if ?b then _ else _] => destruct b
         end; reflexivity.
Qed.

Lemma pow2m1_pos :
  forall prec,
    0 < prec ->
    exists n, (2 ^ prec - 1)%Z = Zpos n /\
              Z.pos (digits2_pos n) = prec.
Proof.
  destruct prec; try discriminate. intro. clear H.
  induction p using Pos.peano_ind. repeat econstructor.
  rewrite Pos2Z.inj_succ, Z.pow_succ_r by (zify; omega).
  replace (2 * 2^Z.pos p - 1)%Z with (2 * (2^Z.pos p - 1)+1)%Z by ring.
  destruct IHp as [n []].
  rewrite H. eexists. split. reflexivity.
  simpl in *. zify; omega.
Qed.

Lemma pow2_pos :
  forall prec,
    0 < prec ->
    exists n, (2 ^ (prec-1))%Z = Zpos n /\
              Z.pos (digits2_pos n) = prec.
Proof.
  destruct prec; try discriminate. intro. clear H.
  induction p using Pos.peano_ind. repeat econstructor.
  replace (Z.pos (Pos.succ p) - 1) with (Z.succ (Zpos p - 1)) by (zify; omega).
  rewrite Pos2Z.inj_succ, Z.pow_succ_r by (zify; omega).
  destruct IHp as [n []].
  rewrite H. eexists. split. reflexivity.
  simpl in *. zify; omega.
Qed.

Definition RFF_lt (x:R) (y:full_float) :=
  if is_finite_FF y then (x < FF2R radix2 y)%R
  else y = F754_infinity false.

Definition RF_lt prec emax (x:R) (y:binary_float prec emax) :=
  RFF_lt x (B2FF _ _ y).

Lemma next_ulp_ff_valid_binary:
  forall prec emax
         (prec_gt_0_ : 0 < prec) (Hmax : prec < emax)
         (f:binary_float prec emax),
    valid_binary prec emax (next_ulp_ff prec emax (B2FF _ _ f)) = true.
Proof.
  intros.
  destruct f as [|[]| |[]]; try easy;
    unfold next_ulp_ff, B2FF, valid_binary, bounded, canonic_mantissa, FLT_exp,
           binary_overflow, overflow_to_inf, negb in *;
    rewrite ?andb_true_iff, ?Z.leb_le, <- ?Zeq_is_eq_bool in *.
  - simpl Z.pos. zify; omega.
  - destruct (pow2m1_pos prec) as [? []]. auto.
    rewrite H. zify; omega.
  - destruct n. apply e.
  - destruct e0.
    destruct (Z.eqb_spec e (3-emax-prec)).
    + subst. destruct (Pos.eqb_spec m 1). auto.
      rewrite andb_true_iff, Z.leb_le. split; [|auto].
      rewrite <- Zeq_is_eq_bool, Z.max_r_iff, Zpos_digits2_pos in *.
      etransitivity; [|now eauto].
      apply Z.sub_le_mono_r, Z.add_le_mono_r, Zdigits_le.
      discriminate.
      rewrite <- (Pos.succ_pred m) at 2 by easy. zify; omega.
    + destruct (Z.eqb_spec (Z.pos (digits2_pos (Pos.pred m))) prec);
        rewrite andb_true_iff, Z.leb_le, <- Zeq_is_eq_bool; split; try (zify; omega).
      destruct (pow2m1_pos prec) as [? []]. auto. rewrite H1, H2.
      zify; omega.
  - destruct (Z.leb_spec (Z.pos (digits2_pos (Pos.succ m))) prec);
      [|destruct (Z.eqb_spec e (emax-prec))]; unfold B2FF; try easy;
    rewrite andb_true_iff, Z.leb_le, <- Zeq_is_eq_bool in *; split; try (zify; omega).
    + destruct e0. revert H0. apply Zmax_case_strong; intros.
      assert (prec-1 < Z.pos (digits2_pos (Pos.succ m))).
      { rewrite Zpos_digits2_pos in *.
        assert (Z.pos m <= Z.pos (Pos.succ m)) by (zify; omega).
        apply (Zdigits_le radix2) in H3. omega. discriminate. }
      zify; omega.
      zify; omega.
    + destruct (pow2_pos prec) as [? []]. auto. rewrite H0, H1.
      zify; omega.
Qed.

Definition next_ulp prec emax prec_gt_0_ Hmax
           (f:binary_float prec emax) : binary_float prec emax :=
  FF2B prec emax _ (next_ulp_ff_valid_binary _ _ prec_gt_0_ Hmax f).

Lemma next_ulp_is_nan:
  forall prec emax prec_gt_0_ Hmax f,
    is_nan _ _ f = false ->
    is_nan _ _ (next_ulp prec emax prec_gt_0_ Hmax f) = false.
Proof.
  intros. unfold next_ulp. rewrite is_nan_FF2B.
  apply next_ulp_ff_is_nan. rewrite is_nan_FF_B2FF. auto.
Qed.

Lemma next_ulp_round :
  forall prec emax prec_gt_0_ Hmax, 1 < prec ->
  forall f, is_finite prec emax f = true ->
  forall x mode,
    (round
      radix2 (FLT_exp (3 - emax - prec) prec)
      (round_mode mode) x <= B2R _ _ f)%R ->
    RF_lt _ _ x (next_ulp _ _ prec_gt_0_ Hmax f).
Proof.
  intros prec emax prec_gt_0_ Hmax prec_gt_1 f FINf x mode LE.
  unfold RF_lt, RFF_lt in *.
  rewrite FF2R_B2FF.
  rewrite is_finite_FF_B2FF.
  destruct (is_finite _ _ (next_ulp prec emax prec_gt_0_ Hmax f)) eqn : FINnext.
  Focus 2.
  destruct f; try discriminate.
  unfold next_ulp, next_ulp_ff in *. simpl in *.
  rewrite is_finite_FF2B in FINnext. rewrite B2FF_FF2B.
  repeat match type of FINnext with context [if ?b then _ else _] => destruct b eqn:? end;
    try discriminate. auto.
  destruct f as [|[]| |[]]; try discriminate.
  - apply Rnot_le_lt. intro. eapply round_ge_generic in H.
    pose proof (Rle_trans _ _ _ H LE).
    simpl in H0. apply F2R_le_0_reg in H0. omega. auto. auto.
    apply generic_format_B2R.
  - apply Rnot_le_lt. intro.
    eapply round_ge_generic in H; eauto. 2:apply generic_format_B2R.
    pose proof (Rle_trans _ _ _ H LE). clear H LE FINf.
    unfold next_ulp, next_ulp_ff in *. simpl in *.
    rewrite B2R_FF2B in H0. rewrite is_finite_FF2B in FINnext.
    repeat match type of FINnext with context [if ?b then _ else _] => destruct b eqn:? end;
      try discriminate.
    + apply F2R_ge_0_reg in H0. zify; omega.
    + apply Rmult_le_reg_r, le_Z2R in H0. apply H0. simpl. apply CompOpp_iff. apply Ppred_Plt.
      destruct m; discriminate.
      apply bpow_gt_0.
    + apply Rmult_le_reg_r, le_Z2R in H0. apply H0. simpl. apply CompOpp_iff. apply Ppred_Plt.
      intro. subst. destruct prec as [|[]|]; discriminate.
      apply bpow_gt_0.
    + unfold FF2R, F2R in H0. simpl in H0.
      assert (e <> 3-emax-prec) by (apply Z.eqb_neq; auto). clear Heqb.
      assert (prec = Z.pos (digits2_pos m)).
      { apply andb_true_iff in e0. destruct e0.
        unfold canonic_mantissa, FLT_exp in H1. apply Zeq_bool_eq in H1. zify; omega. }
      assert (Z.pos (digits2_pos m) <> Z.pos (digits2_pos (Pos.pred m))).
      { destruct prec; try discriminate.
        rewrite Pos.eqb_neq in Heqb0. zify; omega. }
      clear Heqb0 FINnext.
      assert (m <> xH).
      { intro. subst. discriminate. }
      assert (Z.pos (digits2_pos (Pos.pred m)) <= Z.pos (digits2_pos m)).
      { rewrite !Zpos_digits2_pos, Pos2Z.inj_pred by auto.
        apply Zdigits_le. zify; omega. zify; omega. }
      rewrite !Zpos_digits2_pos in *.
      assert (Zdigits radix2 (Z.pos (Pos.pred m)) <= Zdigits radix2 (Z.pos m) - 1) by omega.
      clear H2 H4.
      apply Zpower_gt_Zdigits in H5.
      destruct (Zdigits_correct radix2 (Zpos m)).
      rewrite Pos2Z.inj_pred in H5 by auto.
      assert (Z.pos m = radix2 ^ (Zdigits radix2 (Z.pos m) - 1)) by (zify; omega).
      rewrite <- H1 in H6.
      clear H5 H2 H4 H1.
      destruct (pow2m1_pos prec) as [? []]; auto.
      rewrite H1 in H0.
      change (- Z2R (Z.pos x0) * bpow radix2 (Z.pred e) <= - Z2R (Z.pos m) * bpow radix2 e)%R in H0.
      rewrite <- H1, H6 in H0. change 2 with (radix_val radix2) in H0.
      rewrite Z2R_minus, !Z2R_Zpower, !Ropp_mult_distr_l_reverse in H0 by omega.
      apply Ropp_le_cancel in H0. rewrite <- bpow_plus in H0.
      replace (prec - 1 + e) with (prec + Zpred e) in H0 by (zify; omega).
      rewrite bpow_plus in H0.
      apply Rmult_le_reg_r in H0; [|apply bpow_gt_0].
      simpl in H0. fourier.
  - apply Rnot_le_lt. intro.
    eapply round_ge_generic in H; eauto. 2:apply generic_format_B2R.
    pose proof (Rle_trans _ _ _ H LE). clear H LE FINf.
    unfold next_ulp, next_ulp_ff in *. simpl in *.
    rewrite B2R_FF2B in H0. rewrite is_finite_FF2B in FINnext.
    repeat match type of FINnext with context [if ?b then _ else _] => destruct b eqn:? end;
      try discriminate.
    + apply Rmult_le_reg_r, le_Z2R in H0. simpl in H0. zify; omega.
      apply bpow_gt_0.
    + rewrite Z.leb_gt in Heqb.
      assert (Z.pos (digits2_pos m) <= prec).
      { apply andb_true_iff in e0. destruct e0.
        unfold canonic_mantissa, FLT_exp in H. apply Zeq_bool_eq in H. zify; omega. }
      rewrite Zpos_digits2_pos in *.
      apply Zpower_le_Zdigits in Heqb.
      apply Zpower_gt_Zdigits in H.
      assert (radix2^prec-1 = Z.pos m)%Z by (zify; omega).
      clear H Heqb Heqb0 FINnext.
      destruct (pow2_pos prec) as [? []]; auto.
      simpl in H0. rewrite H, <- H1, <- H in H0.
      unfold F2R, Fexp, Fnum in H0.
      rewrite <- Z.add_1_l, bpow_plus, <- Rmult_assoc in H0.
      apply Rmult_le_reg_r in H0; [|apply bpow_gt_0].
      rewrite Z2R_minus, !Z2R_Zpower with (r:=radix2), <- bpow_plus in H0 by omega.
      replace (prec-1+1) with prec in H0 by ring.
      simpl in H0. fourier.
Qed.

Lemma next_ulp_round_overflow :
    forall prec emax prec_gt_0_ Hmax x mode f, 1 < prec ->
    (if Rlt_bool
          (Rabs (round radix2 (FLT_exp (3 - emax - prec) prec)
                       (round_mode mode) x))
          (bpow radix2 emax)
     then B2R _ _ f = round radix2 (FLT_exp (3 - emax - prec) prec)
                            (round_mode mode) x /\
          is_finite _ _ f = true
     else B2FF _ _ f = binary_overflow prec emax mode (Rlt_bool x 0)) ->
  forall fM, Fleb f fM = true ->
  RF_lt _ _ x (next_ulp prec emax prec_gt_0_ Hmax fM).
Proof.
  intros.
  Rlt_bool_case H0; destruct (is_finite _ _ fM) eqn:?.
  - destruct H0.
    apply next_ulp_round with mode. auto. auto.
    rewrite <- H0. rewrite Fleb_Rle in H1 by auto.
    Rle_bool_case H1. auto. discriminate.
  - destruct H0.
    assert (is_nan _ _ fM = false) by eauto.
    destruct fM as [|[]| |]; try discriminate.
    destruct f as [| | |[]]; discriminate.
    reflexivity.
  - Rlt_bool_case H0.
    + rewrite Rabs_left1 in H2.
      2:erewrite <- round_0 by eauto; apply round_le; eauto; fourier.
      unfold RF_lt, RFF_lt. rewrite is_finite_FF_B2FF, FF2R_B2FF.
      destruct (is_finite _ _ (next_ulp prec emax prec_gt_0_ Hmax fM)) eqn:?.
      * apply Rnot_le_lt. intro.
        apply Ropp_le_contravar in H2. erewrite Ropp_involutive in H2.
        eapply Rle_trans in H2. 2:eapply round_le, H4; eauto.
        rewrite round_generic in H2. 2:auto. 2:apply generic_format_B2R.
        pose proof abs_B2R_lt_emax _ _ (next_ulp prec emax prec_gt_0_ Hmax fM).
        edestruct Rabs_lt_inv; eauto. fourier.
      * revert Heqb0. unfold next_ulp. rewrite is_finite_FF2B, B2FF_FF2B.
        destruct fM as [| | |[]]; try discriminate; unfold B2FF, next_ulp_ff.
        destruct Z.eqb. destruct Pos.eqb; discriminate. destruct Z.eqb; discriminate.
        destruct Z.eqb. destruct Z.leb. discriminate. auto.
        destruct Z.leb; discriminate.
    + unfold binary_overflow in H0.
      destruct (overflow_to_inf mode false);
        destruct f as [| |? []|]; inv H0;
        destruct fM as [|[]| |]; inv H1;
        try reflexivity.
      unfold Fleb, Bcompare in H4. destruct b. discriminate.
      unfold RF_lt, RFF_lt, next_ulp. rewrite B2FF_FF2B. simpl.
      clear Heqb. unfold bounded in e1. rewrite Bool.andb_true_iff in e1. destruct e1.
      apply Zle_is_le_bool in H1.
      destruct (Z.compare_spec (emax - prec) e). 2:omega. 2:discriminate.
      subst e. rewrite Z.eqb_refl.
      edestruct pow2m1_pos as (? & ? & ?). eauto. rewrite H5 in *.
      change (Pos.compare_cont x0 m Eq) with (Pos.compare x0 m) in H4.
      destruct (Pos.compare_spec x0 m); try discriminate.
      subst x0.
      rewrite Zpos_digits2_pos, Pos2Z.inj_succ, <- H5, Z.sub_1_r,
        Z.succ_pred, Zdigits_Zpower, Zle_bool_false by omega.
      reflexivity.
      apply canonic_canonic_mantissa with (sx:=false) in H0.
      unfold canonic, F2R, Fnum, Fexp, cond_Zopp, canonic_exp, FLT_exp in H0.
      rewrite ln_beta_mult_bpow in H0.
      assert (2^prec <= Z.pos m) by (zify; omega).
      apply Z2R_le in H8.
      rewrite Z2R_Zpower with (r:=radix2) in H8 by omega.
      apply ln_beta_le with (r:=radix2) in H8. 2:apply bpow_gt_0.
      rewrite ln_beta_bpow in H8. zify; omega.
      change R0 with (Z2R 0). apply Z2R_neq. discriminate.
  - assert (is_nan _ _ fM = false) by eauto.
    destruct fM as [|[]| |]; try discriminate. 2:reflexivity.
    destruct f as [|[]| |[]]; try discriminate.
    unfold binary_overflow in H0.
    destruct overflow_to_inf; inv H0.
    Rlt_bool_case H5. 2:discriminate.
    rewrite Rabs_left1 in H2.
    2:erewrite <- round_0 by eauto; apply round_le; eauto; fourier.
    unfold RF_lt, RFF_lt, next_ulp, next_ulp_ff. simpl is_finite_FF.
    rewrite FF2R_B2FF. unfold B2FF.
    apply Rnot_le_lt. intro.
    apply Ropp_le_contravar in H2. erewrite Ropp_involutive in H2.
    eapply Rle_trans in H2. 2:eapply round_le, H4; eauto.
    rewrite round_generic in H2. 2:auto. 2:apply generic_format_B2R.
    pose proof abs_B2R_lt_emax _ _
         (FF2B prec emax (binary_overflow prec emax mode_UP true)
               (next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                                         (B754_infinity prec emax true))).
    edestruct Rabs_lt_inv; eauto. fourier.
Qed.

Lemma next_ulp_lub:
  forall prec emax prec_gt_0_ Hmax f1 f2, 1 < prec ->
    is_nan _ _ f1 = false ->
    is_nan prec emax f2 = false ->
    Fleb f1 f2 = true \/
    Fleb (next_ulp _ _ prec_gt_0_ Hmax f2) f1 = true.
Proof.
  intros.
  destruct f1 as [|[]| |[]], f2 as [|[]| |[]]; try discriminate H0; try discriminate H1;
    try (left; reflexivity); try (right; reflexivity).
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ true m e e0))).
    unfold next_ulp_ff, B2FF.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end;
      try (intro; reflexivity).
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ true m e e0))).
    unfold next_ulp_ff, B2FF.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end;
      try (intro; reflexivity).
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ false m e e0))).
    unfold next_ulp_ff, B2FF.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end;
      try (intro; reflexivity).
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_infinity _ _ true))).
    unfold B2FF, next_ulp_ff, binary_overflow, overflow_to_inf, negb, FF2B, valid_binary, Fleb, Bcompare.
    intros _.
    clear H0 H1. apply -> andb_true_iff in e0. destruct e0.
    rewrite Z.leb_compare, Z.compare_antisym in H1.
    destruct Z.compare; inversion H1; clear H1; try reflexivity.
    unfold canonic_mantissa, FLT_exp in H0. apply Zeq_bool_eq in H0.
    assert (Z.pos (digits2_pos m) <= prec) by (zify; omega).
    destruct (pow2m1_pos prec prec_gt_0_) as [? []]. rewrite H2.
    fold (Pos.compare x m). fold (Z.compare (Z.pos x) (Z.pos m)).
    rewrite <- H2. clear x H2 H3.
    apply Zpower_le with (r:=radix2) in H1.
    rewrite Zpos_digits2_pos in *.
    destruct (Zdigits_correct radix2 (Z.pos m)).
    simpl radix_val in *.
    assert (Z.pos m < 2^prec)%Z by (zify; omega).
    destruct (Z.compare_spec (2^prec-1) (Z.pos m)); try reflexivity; omega.
  - clear H0 H1. unfold Fleb, next_ulp, Bcompare.
    generalize (next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ true m0 e1 e2)).
    unfold B2FF, next_ulp_ff, FF2B.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end; intros _.
    + unfold bounded, canonic_mantissa, FLT_exp in e0.
      rewrite andb_true_iff in e0. destruct e0. apply Zeq_bool_eq in H0.
      apply Z.eqb_eq in Heqb.
      left. destruct (Z.compare_spec e e1); auto; try (zify; omega).
      apply Pos.eqb_eq in Heqb0. subst m0. destruct m; auto.
    + rewrite Z.compare_antisym.
      destruct (Z.compare_spec e1 e); auto.
      simpl. fold (Pos.compare m m0). fold (Z.compare (Z.pos m) (Z.pos m0)).
      fold (Pos.compare (Pos.pred m0) m). fold (Z.compare (Z.pos (Pos.pred m0)) (Z.pos m)).
      apply Pos.eqb_neq in Heqb0. rewrite Pos2Z.inj_pred by auto.
      destruct (Z.compare_spec (Z.pos m) (Z.pos m0)); auto.
      right.
      destruct (Z.compare_spec (Z.pred (Z.pos m0)) (Z.pos m)); auto.
      zify; omega.
    + rewrite Z.compare_antisym.
      destruct (Z.compare_spec e1 e); auto.
      simpl. fold (Pos.compare m m0). fold (Z.compare (Z.pos m) (Z.pos m0)).
      fold (Pos.compare (Pos.pred m0) m). fold (Z.compare (Z.pos (Pos.pred m0)) (Z.pos m)).
      rewrite Pos2Z.inj_pred.
      destruct (Z.compare_spec (Z.pos m) (Z.pos m0)); auto.
      right.
      destruct (Z.compare_spec (Z.pred (Z.pos m0)) (Z.pos m)); auto.
      zify; omega.
      intro. subst. simpl in Heqb0. destruct prec as [|[]|]; discriminate.
    + apply Z.eqb_neq in Heqb0. apply Z.eqb_neq in Heqb.
      unfold bounded, canonic_mantissa, FLT_exp in e2, e0.
      rewrite andb_true_iff in e2, e0. destruct e2, e0.
      apply Zeq_bool_eq in H0. apply Zeq_bool_eq in H2.
      assert (Z.pos (digits2_pos m0) = prec) by (zify; omega).
      destruct (Z.compare_spec e e1); destruct (Z.compare_spec (Z.pred e1) e); auto; try (zify; omega).
      * left. fold (Pos.compare m m0). destruct (Pos.compare_spec m m0); auto.
        exfalso. subst e.
        rewrite Zpos_digits2_pos in *.
        assert (Zdigits radix2 (Z.pos m) = prec) by (zify; omega).
        assert (Z.pos m <= Z.pos (Pos.pred m0)) by (rewrite Pos2Z.inj_pred; zify; omega).
        apply Zdigits_le with (beta:=radix2) in H8; [|zify; omega].
        assert (Z.pos (Pos.pred m0) <= Z.pos m0) by (rewrite Pos2Z.inj_pred; zify; omega).
        apply Zdigits_le with (beta:=radix2) in H9; [|zify; omega].
        omega.
      * right. destruct (pow2m1_pos prec) as [? []]. auto.
        rewrite H7.
        fold (Pos.compare x m). fold (Z.compare (Z.pos x) (Z.pos m)).
        rewrite <- H7. clear x H7 H8.
        assert (Z.pos (digits2_pos m) <= prec) by (zify; omega).
        destruct (Z.compare_spec (2^prec-1) (Z.pos m)); try reflexivity.
        apply Zpower_le with (r:=radix2) in H7. simpl radix_val in H7.
        rewrite Zpos_digits2_pos in H7.
        destruct (Zdigits_correct radix2 (Z.pos m)).
        simpl radix_val in *.
        zify; omega.
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_zero _ _ b))).
    unfold next_ulp_ff, B2FF, FF2B, Fleb, Bcompare.
    intros _.
    destruct (Z.compare_spec (3-emax-prec) e).
    destruct m; reflexivity.
    auto.
    clear H0 H1.
    unfold bounded, canonic_mantissa, FLT_exp in e0.
    rewrite andb_true_iff in e0. destruct e0. apply Zeq_bool_eq in H0.
    zify; omega.
  - right. unfold next_ulp.
    generalize ((next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ true m0 e1 e2))).
    unfold next_ulp_ff, B2FF.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end;
      try (intro; reflexivity).
  - clear H H0. unfold Fleb, next_ulp, Bcompare.
    generalize (next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                (B754_finite _ _ false m0 e1 e2)).
    unfold B2FF, next_ulp_ff, FF2B.
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end; intros _.
    + rewrite Z.compare_antisym.
      destruct (Z.compare_spec e1 e); auto.
      simpl. fold (Pos.compare m m0). fold (Z.compare (Z.pos m) (Z.pos m0)).
      fold (Pos.compare (Pos.succ m0) m). fold (Z.compare (Z.pos (Pos.succ m0)) (Z.pos m)).
      destruct (Z.compare_spec (Z.pos m) (Z.pos m0)); auto.
      right.
      destruct (Z.compare_spec (Z.pos (Pos.succ m0)) (Z.pos m)); auto.
      zify; omega.
    + left. rewrite Z.leb_gt in Heqb. apply Z.eqb_eq in Heqb0.
      destruct (Z.compare_spec e e1); auto.
      * fold (Pos.compare m m0). fold (Z.compare (Z.pos m) (Z.pos m0)).
        destruct (Z.compare_spec (Z.pos m) (Z.pos m0)); auto.
        assert (Z.pos (Pos.succ m0) <= Z.pos m) by (zify; omega).
        apply Zdigits_le with (beta:=radix2) in H2; [|zify; omega].
        rewrite Zpos_digits2_pos in *.
        assert (Zdigits radix2 (Z.pos m) > prec) by omega.
        unfold bounded, canonic_mantissa, FLT_exp in e0.
        rewrite andb_true_iff in e0. destruct e0. apply Zeq_bool_eq in H4.
        rewrite Zpos_digits2_pos in *.
        zify; omega.
      * unfold bounded, canonic_mantissa, FLT_exp in e0.
        rewrite andb_true_iff in e0. destruct e0. apply Z.leb_le in H2.
        zify; omega.
    + rewrite Z.leb_gt in Heqb. apply Z.eqb_neq in Heqb0.
      destruct (Z.compare_spec e e1); auto.
      * fold (Pos.compare m m0). destruct (Pos.compare_spec m m0); auto.
        right.
        assert (Z.pos (Pos.succ m0) <= Z.pos m) by (zify; omega).
        apply Zdigits_le with (beta:=radix2) in H2; [|zify; omega].
        rewrite Zpos_digits2_pos in *.
        assert (Zdigits radix2 (Z.pos m) > prec) by omega.
        unfold bounded, canonic_mantissa, FLT_exp in e0.
        rewrite andb_true_iff in e0. destruct e0. apply Zeq_bool_eq in H4.
        rewrite Zpos_digits2_pos in *.
        zify; omega.
      * right. clear H1.
        destruct (Z.compare_spec (Z.succ e1) e); auto.
        unfold bounded, canonic_mantissa, FLT_exp in e0, e2.
        rewrite andb_true_iff in e0. destruct e0. apply Zeq_bool_eq in H1.
        rewrite andb_true_iff in e2. destruct e2. apply Zeq_bool_eq in H3.
        assert (Z.pos (digits2_pos m) = prec) by (zify; omega).
        rewrite Zpos_digits2_pos in *.
        destruct (Zdigits_correct radix2 (Z.pos m)).
        rewrite H5 in H6.
        destruct (pow2_pos _ prec_gt_0_) as [? []]. rewrite H8.
        fold (Pos.compare x m). destruct (Pos.compare_spec x m); auto.
        simpl radix_val in *.
        zify; omega. zify; omega.
Qed.

Lemma next_ulp_le:
  forall prec emax prec_gt_0_ Hmax f,
    is_nan _ _ f = false ->
    Fleb f (next_ulp prec emax prec_gt_0_ Hmax f) = true.
Proof.
  intros.
  destruct f as [|[]| |]; try reflexivity. discriminate.
  unfold next_ulp.
  generalize (next_ulp_ff_valid_binary prec emax prec_gt_0_ Hmax
                                       (B754_finite _ _ b m e e0)).
  unfold next_ulp_ff, B2FF, Fleb, Bcompare, FF2B.
  repeat (match goal with |- context [if ?b then _ else _] => destruct b eqn:? end); intros _;
    try rewrite Z.compare_refl.
  - reflexivity.
  - fold (Pos.compare m (Pos.pred m)). destruct (Pos.compare_spec m (Pos.pred m)); try reflexivity.
    destruct (Pos.succ_pred_or m). subst; discriminate.
    zify; omega.
  - fold (Pos.compare m (Pos.pred m)). destruct (Pos.compare_spec m (Pos.pred m)); try reflexivity.
    destruct (Pos.succ_pred_or m). subst; discriminate.
    zify; omega.
  - destruct (Z.compare_spec e (Z.pred e)); try (zify; omega). reflexivity.
  - fold (Pos.compare m (Pos.succ m)). destruct (Pos.compare_spec m (Pos.succ m)); try (zify; omega). reflexivity.
  - reflexivity.
  - destruct (Z.compare_spec e (Z.succ e)); try (zify; omega). reflexivity.
Qed.

Lemma next_ulp_incr:
  forall prec emax prec_gt_0_ Hmax f1 f2, 1 < prec ->
    Fleb f1 f2 = true ->
    Fleb (next_ulp prec emax prec_gt_0_ Hmax f1)
         (next_ulp prec emax prec_gt_0_ Hmax f2) = true.
Proof.
  intros.
  destruct (next_ulp_lub _ _ prec_gt_0_ Hmax f2 f1). auto.
  destruct f1 as [|[]| |[]], f2; try discriminate; reflexivity.
  destruct f1; try discriminate; reflexivity.
  - destruct (is_finite _ _ f1) eqn:?. destruct (is_finite _ _ f2) eqn:?.
    rewrite Fleb_Rle in H0, H1; auto.
    Rle_bool_case H0. 2:discriminate.
    Rle_bool_case H1. 2:discriminate.
    pose proof (Rle_antisym _ _ H2 H3).
    destruct (is_finite_strict _ _ f1) eqn:?. destruct (is_finite_strict _ _ f2) eqn:?.
    apply B2R_inj in H4; auto. subst. apply Fleb_refl.
    unfold next_ulp, next_ulp_ff, B2FF, Bopp. rewrite is_nan_FF2B.
    destruct f2 as [|[]|? []|]; try reflexivity. discriminate.
    repeat (match goal with |- is_nan_FF (if ?b then _ else _) = false => destruct b end);
      reflexivity.
    destruct f1, f2; try discriminate.
    apply F2R_eq_0_reg in H4. destruct b; discriminate.
    destruct f1, f2; try discriminate.
    unfold next_ulp, Fleb. simpl. rewrite Zcompare_refl. auto.
    symmetry in H4. apply F2R_eq_0_reg in H4. destruct b0; discriminate.
    destruct f2 as [|[]| |], f1 as [| | |[]]; discriminate.
    destruct f1 as [|[]| |], f2 as [|[]| |[]]; try discriminate; try reflexivity.
    auto.
  - eapply Fleb_trans. eauto. apply next_ulp_le.
    destruct f1 as [|[]| |[]], f2; try discriminate; reflexivity.
Qed.

Definition half_backward_zoffloat (a:Z)
           prec emax Hprec Hmax : binary_float prec emax :=
  if 0 <? a then
    binary_normalize prec emax Hprec Hmax mode_UP a 0 false
  else
    next_ulp prec emax Hprec Hmax
      (binary_normalize prec emax Hprec Hmax mode_DN (a-1) 0 false).

Lemma half_backward_zoffloat_correct :
  forall prec emax Hprec Hmax f i a, 1 < prec ->
    ZofB prec emax f = Some i ->
    a <= i ->
    Fleb (half_backward_zoffloat a _ _ Hprec Hmax) f = true.
Proof.
  intros * Hprec1. intros.
  pose proof ZofB_correct _ _ f. rewrite H in H1.
  destruct (is_finite _ _ f) eqn:?; inv H1.
  unfold half_backward_zoffloat.
  destruct (Z.ltb_spec 0 a).
  - destruct f; try (inv H; omega).
    simpl in H0. pose proof Ztrunc_Z2R 0. simpl in H2. omega.
    destruct b.
    pose proof zoffloat_incr _ _ (B754_finite _ _ true m e e0) (B754_zero _ _ false) _ _ eq_refl H eq_refl. omega.
    assert (round radix2 (FLT_exp (3 - emax - prec) prec)
                  (round_mode mode_UP) (F2R (beta:=radix2) {| Fnum := a; Fexp := 0 |}) <=
            B2R _ _ (B754_finite _ _ false m e e0))%R.
    { eapply round_le_generic; auto.
      apply generic_format_B2R.
      unfold F2R, Fnum, Fexp, bpow. rewrite Rmult_1_r.
      eapply Rle_trans.
      apply Z2R_le, H0.
      rewrite Ztrunc_floor. apply Zfloor_lb.
      apply F2R_ge_0_compat. compute. discriminate. }
    use_float_specs.
    rewrite Rlt_bool_true in H3.
    + destruct H3 as [? []].
      rewrite Fleb_Rle; auto. apply Rle_bool_true.
      rewrite H3. auto.
    + rewrite Rabs_pos_eq.
      eapply Rle_lt_trans. eauto. eapply bounded_lt_emax; eauto.
      eapply round_ge_generic; auto. apply generic_format_0.
      apply Rmult_le_pos. apply Z2R_le with (m:=Z0). simpl; omega. apply bpow_ge_0.
  - use_float_specs.
    Rlt_bool_case H2.
    + destruct H2 as [? []].
      destruct (next_ulp_lub _ _ Hprec Hmax f (binary_normalize _ _ Hprec Hmax mode_DN (a - 1) 0 false) Hprec1).
      destruct f; try discriminate; reflexivity.
      destruct binary_normalize; try discriminate; reflexivity.
      2:auto.
      exfalso.
      rewrite Fleb_Rle, H2 in H6 by auto.
      Rle_bool_case H6; [|discriminate].
      apply round_le with (beta:=radix2) (fexp:=Fcore_FIX.FIX_exp 0) (rnd:=round_mode mode_ZR) in H7; eauto with typeclass_instances.
      apply Z2R_le in H0. simpl in H7. unfold round at 1 in H7.
      unfold F2R, scaled_mantissa in H7. simpl in H7.
      rewrite !Rmult_1_r in H7.
      pose proof (Rle_trans _ _ _ H0 H7).
      clear - H8 H1.
      assert (round radix2 (FLT_exp (3 - emax - prec) prec)
                    (round_mode mode_DN) (Z2R (a-1)) <= Z2R (a-1))%R.
      { unfold round, F2R, Fnum, Fexp, scaled_mantissa. simpl.
        eapply Rle_trans.
        apply Rmult_le_compat_r. apply bpow_ge_0. apply Zfloor_lb.
        rewrite Rmult_assoc, <- bpow_plus, Z.add_opp_diag_l. simpl. ring_simplify. apply Rle_refl. }
      apply round_le with (beta:=radix2) (fexp:=Fcore_FIX.FIX_exp 0) (rnd:=round_mode mode_ZR) in H;
        eauto with typeclass_instances.
      pose proof (Rle_trans _ _ _ H8 H). clear H1 H8 H.
      rewrite round_generic in H0; eauto.
      apply le_Z2R in H0. omega.
      unfold generic_format, F2R, Fnum, Fexp, scaled_mantissa, Fcore_FIX.FIX_exp, canonic_exp.
      simpl. rewrite !Rmult_1_r, Ztrunc_Z2R. auto.
    + rewrite Rlt_bool_true in H2.
      destruct binary_normalize as [|[]|? []|]; try discriminate.
      destruct f; try discriminate; try reflexivity.
      destruct (next_ulp_lub _ _ Hprec Hmax (B754_finite _ _ b m e e0) (B754_infinity _ _ true) Hprec1); auto.
      destruct b; discriminate.
      unfold F2R, Fnum, Fexp, bpow. rewrite Rmult_1_r. apply Z2R_lt with (n:=0%Z).
      omega.
Qed.

Lemma half_backward_zoffloat_complete :
  forall prec emax Hprec Hmax a i, 1 < prec ->
    ZofB prec emax (half_backward_zoffloat a _ _ Hprec Hmax) = Some i ->
    a <= i.
Proof.
  intros.
  unfold half_backward_zoffloat in H0.
  destruct (Zlt_bool_spec 0 a).
  - rewrite ZofB_correct in H0. use_float_specs.
    Rlt_bool_case H2.
    + destruct H2 as (? & ? & ?).
      rewrite H4 in H0. inv H0. rewrite H2, <- (Ztrunc_Z2R a) at 1.
      apply Ztrunc_le.
      edestruct round_UP_pt as (_ & A & _). 2:eapply Rle_trans, A. eauto.
      unfold F2R, Fnum, Fexp. simpl. fourier.
    + rewrite Rlt_bool_false in H2.
      destruct binary_normalize, Rlt_bool; discriminate.
      apply F2R_ge_0_compat. simpl. omega.
  - rewrite ZofB_correct in H0. use_float_specs.
    pose proof next_ulp_round_overflow _ _ Hprec Hmax
         (F2R (beta:=radix2) {| Fnum := a - 1; Fexp := 0 |})
         mode_DN
         (binary_normalize prec emax Hprec Hmax mode_DN (a - 1) 0 false) H.
    refine (let A := H3 _ _ (Fleb_refl _ _ _ (binary_normalize_is_nan _ _ _ _ _ _ _ _)) in _).
    { destruct Rlt_bool; intuition. }
    clear H3.
    unfold RF_lt, RFF_lt in A. rewrite is_finite_FF_B2FF, FF2R_B2FF in A.
    destruct
      (is_finite prec emax
         (next_ulp prec emax Hprec Hmax
            (binary_normalize prec emax Hprec Hmax mode_DN (a - 1) 0 false))) eqn:?;
      inv H0.
    unfold F2R in A. simpl in A. rewrite Rmult_1_r in A.
    rewrite (Rlt_bool_true (F2R (beta:=radix2) {| Fnum := a - 1; Fexp := 0 |}) 0) in H2
      by (apply F2R_lt_0_compat; simpl; omega).
    rewrite Ztrunc_ceil.
    + apply Z.nlt_ge. intro.
      assert (∀ x y, x < y -> x <= y-1) by (intros; omega). apply H3 in H0. clear H3.
      eapply Z2R_le, Rle_lt_trans in H0. 2:eauto.
      eapply Rlt_le_trans in H0. 2:apply Zceil_ub. fourier.
    + edestruct next_ulp_lub with (f1:=B754_zero prec emax false) as [B|B].
      auto. auto. apply binary_normalize_is_nan.
      2:rewrite Fleb_Rle in B; [Rle_bool_case B; [now eauto|discriminate]|now eauto|now eauto].
      exfalso.
      match goal with H : context [Rlt_bool ?a ?b] |- _ =>
        Rlt_bool_case H; [destruct H as [H []]|
                          unfold binary_overflow in H; simpl in H;
                          apply B2FF_inj with (y:=B754_infinity _ _ _) in H]
      end.
      * rewrite Fleb_Rle, H2 in B by auto. Rle_bool_case B. 2:discriminate.
        unfold F2R, Fnum, Fexp, bpow in H5. rewrite Rmult_1_r in H5.
        assert (a-1 <= -1) by omega.
        eapply Z2R_le, round_le, Rle_trans in H6; eauto.
        rewrite round_generic in H6; auto. simpl in H6. fourier.
        apply generic_format_opp, generic_format_bpow with (e:=0).
        unfold FLT_exp. zify; omega.
      * rewrite H2 in B. discriminate.
Qed.

Lemma zofb_neg :
  ∀ prec emax f,
    ZofB prec emax (Bopp _ _ (fun a b => (a,b)) f) =
    option_map Zopp (ZofB prec emax f).
Proof.
  intros. rewrite !ZofB_correct, is_finite_Bopp, B2R_Bopp.
  rewrite Ztrunc_opp. destruct is_finite; auto.
Qed.

Lemma Bconv_Bopp:
  ∀ prec emax
    prec' emax' Hprec' Hemax'
    conv_pl opp_pl opp_pl'
    x,
    (∀ s pl, (let '(s, pl) := opp_pl s pl in conv_pl s pl) =
             (let '(s, pl) := conv_pl s pl in opp_pl' s pl)) ->
    Bopp prec' emax' opp_pl' (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl mode_NE x) =
    Bconv prec emax prec' emax' Hprec' Hemax' conv_pl mode_NE (Bopp prec emax opp_pl x).
Proof.
  intros.
  use_float_specs.
  rewrite B2R_Bopp, round_NE_opp, Rabs_Ropp in H1.
  destruct x; try reflexivity; intros.
  unfold Bopp, Bconv.
  specialize (H b n). destruct (conv_pl b n), (opp_pl b n). rewrite H. auto.
  assert (∀ prec emax x, is_finite prec emax x = true -> B2R _ _ x <> 0%R ->
                         is_finite_strict _ _ x = true).
  { destruct x; try discriminate; try reflexivity.
    destruct 2. auto. }
  specialize (H0 eq_refl). specialize (H1 eq_refl).
  change ZnearestE with (round_mode mode_NE) in H1.
  Rlt_bool_case H1.
  - destruct H0 as (? & ? & ?). destruct H1 as (? & ? & ?).
    apply B2R_Bsign_inj.
    rewrite is_finite_Bopp. auto.
    auto.
    rewrite B2R_Bopp, H0, H1. auto.
    rewrite H7. clear - H5 H4. destruct Bconv; simpl in H5; subst b; auto. discriminate.
  - apply B2FF_inj with (y:=B754_infinity _ _ _) in H0.
    apply B2FF_inj with (y:=B754_infinity _ _ _) in H1.
    rewrite H0, H1. auto.
Qed.

Lemma Bconv_Babs:
  ∀ prec emax
    prec' emax' Hprec' Hemax'
    conv_pl abs_pl abs_pl'
    x,
    (∀ s pl, (let '(s, pl) := abs_pl s pl in conv_pl s pl) =
             (let '(s, pl) := conv_pl s pl in abs_pl' s pl)) ->
    Babs prec' emax' abs_pl' (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl mode_NE x) =
    Bconv prec emax prec' emax' Hprec' Hemax' conv_pl mode_NE (Babs prec emax abs_pl x).
Proof.
  intros.
  use_float_specs.
  rewrite B2R_Babs, round_NE_abs, Rabs_Rabsolu in H1 by auto.
  destruct x; try reflexivity; intros.
  unfold Babs, Bconv.
  specialize (H b n). destruct (conv_pl b n), (abs_pl b n). rewrite H. auto.
  assert (∀ prec emax x, is_finite prec emax x = true -> B2R _ _ x <> 0%R ->
                         is_finite_strict _ _ x = true).
  { destruct x; try discriminate; try reflexivity.
    destruct 2. auto. }
  specialize (H0 eq_refl). specialize (H1 eq_refl).
  change ZnearestE with (round_mode mode_NE) in H1.
  Rlt_bool_case H1.
  - destruct H0 as (? & ? & ?). destruct H1 as (? & ? & ?).
    apply B2R_Bsign_inj.
    rewrite is_finite_Babs. auto.
    auto.
    rewrite B2R_Babs, H0, H1. auto.
    rewrite H7, !Bsign_Babs. auto. auto. destruct Bconv; try reflexivity; discriminate.
  - apply B2FF_inj with (y:=B754_infinity _ _ _) in H0.
    apply B2FF_inj with (y:=B754_infinity _ _ _) in H1.
    rewrite H0, H1. auto.
Qed.

Lemma Bconv_Bplus:
  ∀ prec emax Hprec Hemax
    prec' emax' Hprec' Hemax'
    conv_pl conv_pl' plus_pl plus_pl'
    x y,
    2*prec + 1 <= prec' ->
    emax < emax' ->

    is_nan _ _ x = false ->
    is_nan _ _ y = false ->

    (∀ x y, is_nan _ _ x = false -> is_nan _ _ y = false ->
     ∀ x' y', is_nan _ _ x' = false -> is_nan _ _ y' = false ->
     (let '(b, pl) := plus_pl' x' y'  in conv_pl b pl) = plus_pl x y) ->

    Bplus prec emax Hprec Hemax plus_pl mode_NE x y =
    Bconv prec' emax' prec emax Hprec Hemax conv_pl mode_NE
      (Bplus prec' emax' Hprec' Hemax' plus_pl' mode_NE
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE x)
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE y)).
Proof.
  intros * prec_le emax_le NANx NANy NAN. unfold Fcore_FLX.Prec_gt_0 in *.
  destruct (is_finite _ _ x) eqn:FINx;
    [assert (Hx:=FINx);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hx; eauto; try omega|];
  (destruct (is_finite _ _ y) eqn:FINy;
    [assert (Hy:=FINy);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hy; eauto; try omega|]).
  - use_float_specs. clear H H0.
    destruct Hx as (? & ? & ?), Hy as (? & ? & ?).
    rewrite H, H4, H5, H7, Rlt_bool_true in H2.
    + destruct H2 as (? & ? & ?); auto.
      assert (round radix2 (FLT_exp (3-emax-prec) prec) (round_mode mode_NE)
                    (round radix2 (FLT_exp (3-emax'-prec') prec')
                           (round_mode mode_NE) (B2R _ _ x + B2R _ _ y)) =
              round radix2 (FLT_exp (3-emax-prec) prec)
                    (round_mode mode_NE) (B2R _ _ x + B2R _ _ y)).
      { apply double_round_plus_FLT; auto. omega.
        apply FLT_format_B2R. auto. apply FLT_format_B2R. auto. }
      rewrite H2, H10, H9 in H1.
      Rlt_bool_case H1.
      * destruct H1 as (? & ? & ?), H3 as (? & ? & ?); auto.
        apply B2R_Bsign_inj; auto; congruence.
      * apply B2FF_inj. destruct H3; auto. rewrite H1, H3; auto. f_equal.
        replace 0%R with (0+0)%R by ring.
        destruct x, y; try discriminate; simpl in H12; subst b.
        simpl. rewrite Rcompare_Eq by auto. destruct b0; auto.
        destruct b0; [rewrite Rcompare_Lt|rewrite Rcompare_Gt]; auto; apply Rplus_lt_compat;
        (apply F2R_lt_0_compat || apply F2R_gt_0_compat); reflexivity.
    + assert (∀ x, Rabs (B2R prec emax x) <=
                   Fcore_ulp.pred radix2 (FLT_exp (3 - emax - prec) prec) (bpow radix2 emax))%R.
      { intros.
        apply Fcore_ulp.le_pred_lt. auto.
        apply generic_format_abs, generic_format_B2R.
        apply generic_format_bpow. unfold FLT_exp. zify; omega.
        apply bpow_gt_0.
        apply abs_B2R_lt_emax. }
      pose proof (H8 x). specialize (H8 y).
      eapply Rplus_le_compat in H8. 2:apply H9. clear H9.
      eapply Rle_trans in H8. 2:apply Rabs_triang.
      rewrite <- round_NE_abs by auto.
      eapply Rle_lt_trans. apply round_le, H8. auto. apply (valid_rnd_round_mode mode_NE).
      rewrite <- double.
      unfold Fcore_ulp.pred.
      rewrite ln_beta_bpow, Z.add_simpl_r, Req_bool_true, Rmult_minus_distr_l by auto.
      replace (FLT_exp (3 - emax - prec) prec emax) with (emax - prec)
        by (unfold FLT_exp; zify; omega).
      rewrite round_generic. 2:apply (valid_rnd_round_mode mode_NE).
      * assert (1+emax <= emax') by omega.
        apply bpow_le with (r:=radix2) in H9.
        rewrite bpow_plus in H9. simpl in H9.
        pose proof (bpow_gt_0 radix2 (emax - prec)).
        fourier.
      * apply generic_format_FLT.
        exists {| Fnum := radix2^prec - 1; Fexp := emax - prec + 1 |}.
        { split; [|split].
          - unfold F2R, Fnum, Fexp.
            rewrite Z2R_minus, Z2R_Zpower, bpow_plus,
                    Rmult_minus_distr_r, <- Rmult_assoc, <- bpow_plus,
                    Zplus_minus by omega.
            simpl. ring.
          - unfold Fnum.
            assert (0 < radix2 ^ prec) by (apply Zpower_gt_0; omega).
            assert (prec <= prec') by omega. apply Zpower_le with (r:=radix2) in H10.
            zify; omega.
          - unfold Fexp. omega. }
  - destruct Hx as (? & ? & ?).
    destruct x, y; try discriminate.
    + auto.
    + destruct Bconv; try discriminate; auto.
  - destruct Hy as (? & ? & ?).
    destruct x, y; try discriminate.
    + auto.
    + destruct Bconv; try discriminate; auto.
  - destruct x, y; try discriminate.
    simpl. destruct (eqb b b0). auto. simpl.
    specialize (NAN (B754_infinity prec emax b) (B754_infinity prec emax b0) eq_refl eq_refl
                    (B754_infinity prec' emax' b) (B754_infinity prec' emax' b0) eq_refl eq_refl).
    rewrite <- NAN.
    destruct plus_pl'. simpl. destruct conv_pl. auto.
Qed.

Lemma Bconv_Bmult:
  ∀ prec emax Hprec Hemax
    prec' emax' Hprec' Hemax'
    conv_pl conv_pl' mult_pl mult_pl'
    x y,
    2*prec <= prec' ->
    2*emax <= emax' ->

    is_nan _ _ x = false ->
    is_nan _ _ y = false ->

    (∀ x y, is_nan _ _ x = false -> is_nan _ _ y = false ->
     ∀ x' y', is_nan _ _ x' = false -> is_nan _ _ y' = false ->
     (let '(b, pl) := mult_pl' x' y'  in conv_pl b pl) = mult_pl x y) ->

    Bmult prec emax Hprec Hemax mult_pl mode_NE x y =
    Bconv prec' emax' prec emax Hprec Hemax conv_pl mode_NE
      (Bmult prec' emax' Hprec' Hemax' mult_pl' mode_NE
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE x)
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE y)).
Proof.
  intros * prec_le emax_le NANx NANy NAN. unfold Fcore_FLX.Prec_gt_0 in *.
  destruct (is_finite _ _ x) eqn:FINx;
    [assert (Hx:=FINx);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hx; eauto; try omega|];
  (destruct (is_finite _ _ y) eqn:FINy;
    [assert (Hy:=FINy);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hy; eauto; try omega|]).
  - use_float_specs. clear H H0.
    destruct Hx as (? & ? & ?), Hy as (? & ? & ?).
    rewrite H, H4, H5, H7, Rlt_bool_true in H2.
    + destruct H2 as (? & ? & ?); auto.
      assert (round radix2 (FLT_exp (3-emax-prec) prec) (round_mode mode_NE)
                    (round radix2 (FLT_exp (3-emax'-prec') prec')
                           (round_mode mode_NE) (B2R _ _ x * B2R _ _ y)) =
              round radix2 (FLT_exp (3-emax-prec) prec)
                    (round_mode mode_NE) (B2R _ _ x * B2R _ _ y)).
      { apply double_round_mult_FLT; auto. omega.
        apply FLT_format_B2R. auto. apply FLT_format_B2R. auto. }
      rewrite H2, H10, H9 in H1. 2:apply is_finite_not_is_nan; rewrite H8, H0, H6; auto.
      Rlt_bool_case H1.
      * destruct H1 as (? & ? & ?), H3 as (? & ? & ?). rewrite H8, H0, H6. auto.
        apply B2R_Bsign_inj. rewrite H14, FINx, FINy. auto. apply H12.
        congruence.
        rewrite H15, H13. auto.
        apply is_finite_not_is_nan. rewrite H14, FINx, FINy. auto.
      * apply B2FF_inj. rewrite H1, H3. auto.
        rewrite H8, H0, H6. auto.
    + rewrite round_generic, Rabs_mult.
      eapply Rlt_le_trans.
      apply Rmult_le_0_lt_compat. apply Rabs_pos. apply Rabs_pos.
      apply abs_B2R_lt_emax. apply abs_B2R_lt_emax.
      rewrite <- bpow_plus. apply bpow_le. omega.
      apply (valid_rnd_round_mode mode_NE).
      eapply double_round_mult_aux; try apply generic_format_B2R.
      split; intros; unfold FLT_exp; zify; omega.
  - destruct Hx as (? & ? & ?).
    destruct x, y; try discriminate.
    + specialize (NAN (B754_zero _ _ b) (B754_infinity _ _ b0) eq_refl eq_refl
                      (B754_zero _ _ b) (B754_infinity _ _ b0) eq_refl eq_refl).
      simpl.
      destruct mult_pl', mult_pl. simpl. rewrite NAN. auto.
    + destruct Bconv; try discriminate; auto.
      symmetry in H. apply F2R_eq_0_reg in H. destruct b; discriminate.
      simpl in H1. subst b1. auto.
  - destruct Hy as (? & ? & ?).
    destruct x, y; try discriminate.
    + specialize (NAN (B754_infinity _ _ b) (B754_zero _ _ b0) eq_refl eq_refl
                      (B754_infinity _ _ b) (B754_zero _ _ b0) eq_refl eq_refl).
      simpl.
      destruct mult_pl', mult_pl. simpl. rewrite NAN. auto.
    + destruct Bconv; try discriminate; auto.
      symmetry in H. apply F2R_eq_0_reg in H. destruct b0; discriminate.
      simpl in H1. subst b1. auto.
  - destruct x, y; try discriminate. auto.
Qed.

Lemma Bconv_Bdiv:
  ∀ prec emax Hprec Hemax
    prec' emax' Hprec' Hemax'
    conv_pl conv_pl' div_pl div_pl'
    x y,
    2 * prec <= prec' ->
    2 + emax + 2 * prec  <= emax' + prec'->
    -3 + 2 * emax + prec <= emax' ->

    is_nan _ _ x = false ->
    is_nan _ _ y = false ->

    (∀ x y, is_nan _ _ x = false -> is_nan _ _ y = false ->
     ∀ x' y', is_nan _ _ x' = false -> is_nan _ _ y' = false ->
     (let '(b, pl) := div_pl' x' y'  in conv_pl b pl) = div_pl x y) ->

    Bdiv prec emax Hprec Hemax div_pl mode_NE x y =
    Bconv prec' emax' prec emax Hprec Hemax conv_pl mode_NE
      (Bdiv prec' emax' Hprec' Hemax' div_pl' mode_NE
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE x)
        (Bconv prec emax prec' emax' Hprec' Hemax' conv_pl' mode_NE y)).
Proof.
  intros * prec_le le1 le2 NANx NANy NAN. unfold Fcore_FLX.Prec_gt_0 in *.
  destruct (is_finite _ _ x) eqn:FINx;
    [assert (Hx:=FINx);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hx; eauto; try omega|];
  (destruct (is_finite _ _ y) eqn:FINy;
    [assert (Hy:=FINy);
     eapply (Bconv_widen_exact prec emax prec' emax' Hprec Hprec' Hemax')
     with (conv_nan:=conv_pl') (m:=mode_NE) in Hy; eauto; try omega|]).
  destruct (Req_EM_T (B2R _ _ y) 0).
  - destruct y; try discriminate.
    + destruct x; try discriminate.
      specialize (NAN (B754_zero _ _ b0) (B754_zero _ _ b) eq_refl eq_refl
                      (B754_zero _ _ b0) (B754_zero _ _ b) eq_refl eq_refl).
      simpl. rewrite <- NAN. destruct div_pl'. simpl. destruct conv_pl. auto.
      destruct Hx as (? & ? & ?). destruct Bconv; try discriminate.
      symmetry in H. apply F2R_eq_0_reg in H. destruct b0; discriminate.
      simpl in H1. subst b1. auto.
    + apply F2R_eq_0_reg in e. destruct b; discriminate.
  - use_float_specs. clear H H0.
    destruct Hx as (? & ? & ?), Hy as (? & ? & ?).
    rewrite H, H4, H5, H7, Rlt_bool_true in H2.
    + destruct H2 as (? & ? & ?); auto.
      assert (round radix2 (FLT_exp (3-emax-prec) prec) (round_mode mode_NE)
                    (round radix2 (FLT_exp (3-emax'-prec') prec')
                           (round_mode mode_NE) (B2R _ _ x / B2R _ _ y)) =
              round radix2 (FLT_exp (3-emax-prec) prec)
                    (round_mode mode_NE) (B2R _ _ x / B2R _ _ y)).
      { apply double_round_div_FLT; auto. exists 1. auto. omega.
        apply FLT_format_B2R. auto. apply FLT_format_B2R. auto. }
      rewrite H2, H10, H9 in H1. 2:apply is_finite_not_is_nan; rewrite H8; auto.
      Rlt_bool_case H1.
      * destruct H1 as (? & ? & ?), H3 as (? & ? & ?); auto. rewrite H8; auto.
        apply B2R_Bsign_inj. rewrite H14; auto. apply H12.
        congruence.
        rewrite H15, H13. auto.
        apply is_finite_not_is_nan. rewrite H14; auto.
      * apply B2FF_inj. rewrite H1, H3. auto. auto.
        rewrite H8; auto.
    + rewrite <- round_NE_abs by auto.
      unfold Rdiv. rewrite Rabs_mult, Rabs_Rinv by auto.
      assert (bpow radix2 (3-emax-prec) <= Rabs (B2R _ _ y))%R.
      { revert n.
        edestruct FLT_format_B2R with (x:=y) as ([] & -> & ? & ?). auto.
        rewrite <- F2R_Zabs. unfold F2R, Fcore_defs.Fnum, Fcore_defs.Fexp in *.
        intro.
        assert (Fnum <> 0).
        { contradict n. subst. simpl. ring. }
        rewrite <- Rmult_1_l at 1.
        apply Rmult_le_compat. fourier. apply bpow_ge_0.
        apply Z2R_le with (m:=1). zify; omega.
        apply bpow_le. auto. }
      apply Rinv_le in H8. 2:apply bpow_gt_0. rewrite <- bpow_opp in H8.
      assert (Rabs (B2R prec emax x) <=
              Fcore_ulp.pred radix2 (FLT_exp (3 - emax - prec) prec) (bpow radix2 emax))%R.
      { intros.
        apply Fcore_ulp.le_pred_lt. auto.
        apply generic_format_abs, generic_format_B2R.
        apply generic_format_bpow. unfold FLT_exp. zify; omega.
        apply bpow_gt_0.
        apply abs_B2R_lt_emax. }
      eapply Rle_lt_trans. apply round_le. auto. apply (valid_rnd_round_mode mode_NE).
      apply Rmult_le_compat.
      apply Rabs_pos.
      apply Rlt_le, Rinv_0_lt_compat.
      destruct (Rabs_pos (B2R _ _ y)). auto. apply Rabs_no_R0 in n. destruct n. auto.
      apply H9. apply H8.
      unfold Fcore_ulp.pred.
      rewrite ln_beta_bpow, Z.add_simpl_r, Req_bool_true, Rmult_minus_distr_r,
              <- !bpow_plus by auto.
      replace (FLT_exp (3 - emax - prec) prec emax)
      with (emax-prec)
        by (unfold FLT_exp; zify; omega).
      rewrite round_generic.
      assert (0 < bpow radix2 (emax - prec + - (3 - emax - prec)))%R
        by apply bpow_gt_0.
      assert (emax + - (3 - emax - prec) <= emax') by omega.
      apply bpow_le with (r:=radix2) in H11.
      fourier.
      apply (valid_rnd_round_mode mode_NE).
      apply generic_format_FLT.
      exists {| Fnum := radix2^prec - 1; Fexp := emax - prec + - (3 - emax - prec) |}.
      { split; [|split].
        - unfold F2R, Fnum, Fexp.
          rewrite Z2R_minus, Z2R_Zpower, Rmult_minus_distr_r.
          f_equal. rewrite <- bpow_plus. f_equal. ring.
          unfold Z2R, P2R. ring. omega.
        - unfold Fnum.
          assert (0 < radix2 ^ prec) by (apply Zpower_gt_0; omega).
          assert (prec <= prec') by omega. apply Zpower_le with (r:=radix2) in H11.
          zify; omega.
        - unfold Fexp. omega. }
  - destruct Hx as (? & ? & ?).
    destruct x, y; try discriminate.
    + auto.
    + destruct Bconv; try discriminate; simpl in H1; subst b1; auto.
  - destruct Hy as (? & ? & ?).
    destruct x, y; try discriminate.
    + auto.
    + destruct Bconv; try discriminate; simpl in H1; subst b1; auto.
  - destruct x, y; try discriminate. simpl.
    specialize (NAN (B754_infinity _ _ b) (B754_infinity _ _ b0) eq_refl eq_refl
                    (B754_infinity _ _ b) (B754_infinity _ _ b0) eq_refl eq_refl).
    rewrite <- NAN. destruct div_pl'. simpl. destruct conv_pl. auto.
Qed.

Transparent Float32.add Float.add Float32.sub Float.sub Float32.mul Float.mul
            Float32.div Float.div Float.to_single Float.of_single.
Transparent Float.transform_quiet_pl Float.reduce_pl Float.to_single_pl.
Transparent Archi.choose_binop_pl_32 Archi.choose_binop_pl_64
            Archi.default_pl_32 Archi.default_pl_64
            Archi.float_of_single_preserves_sNaN.

Lemma transform_quiet_32_64:
  ∀ n,
    Float32.transform_quiet_pl n =
    Float.reduce_pl (Float.transform_quiet_pl (Float.expand_pl n)).
Proof.
  Transparent Float32.transform_quiet_pl Float.transform_quiet_pl Float.reduce_pl Float.expand_pl.
  intros.
  unfold Float32.transform_quiet_pl, Float.transform_quiet_pl, Float.reduce_pl, Float.expand_pl.
  f_equal.
  assert (∀ (b:bool) (e:b=b), e = eq_refl).
  { destruct b; intros;
    exact (match e with eq_refl => eq_refl end). }
  assert (∀ (a b:bool) (e1 e2:a=b), e1 = e2).
  { intros. destruct a, b; try discriminate; rewrite H; apply H. }
  apply H0.
Qed.

Lemma plus_double_round:
  ∀ (x y:float32),
    Float32.add x y = Float.to_single (Float.add (Float.of_single x) (Float.of_single y)).
Proof.
  intros.
  destruct (is_nan _ _ x) eqn:NANx, (is_nan _ _ y) eqn:NANy.
  - destruct x, y; try discriminate. simpl.
    rewrite 2!Float.transform_quiet_pl_idempotent.
    f_equal. apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANy).
    destruct x; try discriminate; simpl;
    destruct (Float.of_single y), y; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANx).
    destruct y; try discriminate; simpl;
    destruct (Float.of_single x), x; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - apply Bconv_Bplus. omega. omega. auto. auto.
    intros.
    assert ((let '(s, pl) := Archi.default_pl_64 in Float.to_single_pl s pl) =
            Archi.default_pl_32).
    { unfold Float.transform_quiet_pl, Float.to_single_pl, Archi.default_pl_32, Float.reduce_pl. simpl.
      f_equal. f_equal.
      assert (∀ (b:bool) (e:b=b), e = eq_refl).
      { destruct b; intros;
        exact (match e with eq_refl => eq_refl end). }
      apply H3. }
    destruct x0, y0, x', y'; try discriminate; apply H3.
Qed.

Lemma sub_double_round:
  ∀ (x y:float32),
    Float32.sub x y = Float.to_single (Float.sub (Float.of_single x) (Float.of_single y)).
Proof.
  intros.
  destruct (is_nan _ _ x) eqn:NANx, (is_nan _ _ y) eqn:NANy.
  - destruct x, y; try discriminate. simpl.
    rewrite 2!Float.transform_quiet_pl_idempotent.
    f_equal. apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANy).
    destruct x; try discriminate; simpl;
    destruct (Float.of_single y), y; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANx).
    destruct y; try discriminate; simpl;
    destruct (Float.of_single x), x; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - unfold Float32.sub, Float.sub, Bminus, Float.of_single.
    erewrite Bconv_Bopp.
    apply Bconv_Bplus. omega. omega. auto.
    destruct y; try discriminate; reflexivity.
    intros.
    assert ((let '(s, pl) := Archi.default_pl_64 in Float.to_single_pl s pl) =
            Archi.default_pl_32).
    { unfold Float.transform_quiet_pl, Float.to_single_pl, Archi.default_pl_32, Float.reduce_pl. simpl.
      f_equal. f_equal.
      assert (∀ (b:bool) (e:b=b), e = eq_refl).
      { destruct b; intros;
        exact (match e with eq_refl => eq_refl end). }
      apply H3. }
    destruct x0, y0, x', y'; try discriminate; apply H3.
    auto.
Qed.

Lemma mult_double_round:
  ∀ (x y:float32),
    Float32.mul x y = Float.to_single (Float.mul (Float.of_single x) (Float.of_single y)).
Proof.
  intros.
  destruct (is_nan _ _ x) eqn:NANx, (is_nan _ _ y) eqn:NANy.
  - destruct x, y; try discriminate. simpl.
    rewrite 2!Float.transform_quiet_pl_idempotent.
    f_equal. apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANy).
    destruct x; try discriminate; simpl;
    destruct (Float.of_single y), y; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANx).
    destruct y; try discriminate; simpl;
    destruct (Float.of_single x), x; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - apply Bconv_Bmult. omega. omega. auto. auto.
    intros.
    assert ((let '(s, pl) := Archi.default_pl_64 in Float.to_single_pl s pl) =
            Archi.default_pl_32).
    { unfold Float.transform_quiet_pl, Float.to_single_pl, Archi.default_pl_32, Float.reduce_pl. simpl.
      f_equal. f_equal.
      assert (∀ (b:bool) (e:b=b), e = eq_refl).
      { destruct b; intros;
        exact (match e with eq_refl => eq_refl end). }
      apply H3. }
    destruct x0, y0, x', y'; try discriminate; apply H3.
Qed.

Lemma div_double_round:
  ∀ (x y:float32),
    Float32.div x y = Float.to_single (Float.div (Float.of_single x) (Float.of_single y)).
Proof.
  intros.
  destruct (is_nan _ _ x) eqn:NANx, (is_nan _ _ y) eqn:NANy.
  - destruct x, y; try discriminate. simpl.
    rewrite 2!Float.transform_quiet_pl_idempotent.
    f_equal. apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANy).
    destruct x; try discriminate; simpl;
    destruct (Float.of_single y), y; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - pose proof (floatofsingle_nan _ NANx).
    destruct y; try discriminate; simpl;
    destruct (Float.of_single x), x; try discriminate; simpl;
    rewrite !Float.transform_quiet_pl_idempotent; f_equal;
    apply transform_quiet_32_64.
  - apply Bconv_Bdiv. omega. omega. omega. auto. auto.
    intros.
    assert ((let '(s, pl) := Archi.default_pl_64 in Float.to_single_pl s pl) =
            Archi.default_pl_32).
    { unfold Float.transform_quiet_pl, Float.to_single_pl, Archi.default_pl_32, Float.reduce_pl. simpl.
      f_equal. f_equal.
      assert (∀ (b:bool) (e:b=b), e = eq_refl).
      { destruct b; intros;
        exact (match e with eq_refl => eq_refl end). }
      apply H3. }
    destruct x0, y0, x', y'; try discriminate; apply H3.
Qed.
